#ifndef __TCAMERA_HH
#define __TCAMERA_HH
//
// Project : lib3D
// File    : TCamera.hh
// Purpose : class representing a camera (viewing plane etc)
// Author  : Ronald Kriemann
//
// arch-tag: bae150b4-7944-4c71-a0f2-98e757e8d1bc
//

#include <iostream>

#include "TMatrix.hh"
#include "TArray.hh"
#include "TVector.hh"
#include "TRay.hh"

namespace RGL
{

//
// projection is defined by origin and direction of camera
// viewing plane is calculated from this (always look up)
//
class TCamera
{
protected:
    // dimension of the actual film
    // (used to compute view plane)
    real             _film_width, _film_height;
    
    // distance of the film (view plane) from the origin
    real             _film_dist;
    
    // origin and point to look at
    TVector          _origin, _lookat;

    // orthonormal system of the camera 
    TVector          _u, _v, _w;

    // upper left corner of the view plane
    TVector          _ulc;

    // viewplane in uvw-system
    // (down, since in image y-coord. also goes down)
    TVector          _vp_right, _vp_down;
    
    // dimension of view plane
    real             _width, _height;

    // front- and backplane of view-volume
    real             _front, _back;

public:
    //////////////////////////////////////////////
    //
    // constructor and destructor
    //

    TCamera ();

    virtual ~TCamera ();

    //////////////////////////////////////////////
    //
    // define view plane
    //

    // construct plane by a and b such that u || a
    // and v lies in the plane of a and b
    // the length of a and b defines the aspect ratio of
    // the resulting view plane
    void construct_by_plane ( const TVector & a, const TVector & b );

    // construct view plane by a given direction and
    // orthogonal but otherwise arbitrary plane vectors
    void construct_by_dir ( const TVector & a );

    // define camera by given data
    void set_origin_lookat ( const TVector & orig, const TVector & lookat );
    void set_origin_dir    ( const TVector & orig, const TVector & dir );
    void set               ( const TCamera & cam );

    // set dimensions of film (and hence of view plane)
    void set_film_size ( real width, real height );

    // set film distance from origin
    void set_film_dist ( real dist );
    
    //////////////////////////////////////////////
    //
    // setup ray corresponding to pixel
    // coordinates (i,j), i,j \in [0,1]
    //

    virtual void set_rays ( TArray< TRay > & rays, real i, real j ) const = 0;
        
    //////////////////////////////////////////////
    //
    // transform camera
    //

    void move       ( const TVector & dist );
    void move_horiz ( real f );
    void move_vert  ( real f );
    void move_dir   ( real f );
    
    void rotate ( const TVector & dir, real angle );
    void rotate ( const TVector & orig,
                  const TVector & dir,
                  real angle );

    void rotate_horiz ( real angle );
    void rotate_vert  ( real angle );
    
    ////////////////////////////////////////////////////////
    //
    // access local variables
    //

    const TVector & origin  () const { return _origin; }
    const TVector & look_at () const { return _lookat; }

    const TVector & up    () const { return _v; }
    const TVector & right () const { return _u; }

    real vp_width  () const { return _width; }
    real vp_height () const { return _height; }

    real front () const { return _front; }
    real back  () const { return _back; }

    void set_view_vol ( real f, real b ) { _front = f; _back = b; }
    
    ///////////////////////////////////////////////////////
    //
    // misc.
    //

    // copy operator
    TCamera & operator = ( const TCamera & cam ) { set( cam ); return *this; }

    // stream output
    friend std::ostream & operator << ( std::ostream & os, const TCamera & c );
    
protected:

    //
    // update internal data
    //
    
    virtual void update ();
};

} // namespace

#endif  // __TCAMERA_HH
