#ifndef __TFILE_HH
#define __TFILE_HH
//
// Project : lib3D
// File    : TFile.hh
// Purpose : classes for encapsulating IO and file access
// Author  : Ronald Kriemann
//
// arch-tag: b3872791-a7bb-45c3-a1d5-263916da273d
//

#include <stdio.h>

#ifdef USE_ZLIB
#include <zlib.h>
#endif

#ifdef USE_BZ2
#include <bzlib.h>
#endif

#include "TString.hh"

namespace RGL
{

////////////////////////////////////////////////////
//
// base class for all IO handlers
//
class TIO
{
public:
    //
    // constructor and destructor
    //

    TIO () {}

    virtual ~TIO () {}
    
    //
    // stream management
    //

    virtual bool eof () const = 0;
    
    //
    // basic input/output
    //

    // write string 
    virtual int puts ( const TString & string ) = 0;

    // read string
    virtual int gets ( TString & string ) = 0;
};

////////////////////////////////////////////////////
//
// class for handling IO via readline
//
class TReadline : public TIO
{
protected:
    // prompt to display (default = "> ")
    TString  _prompt;

    // true if end-of-file reached
    bool     _eof;

    // length of history
    int      _hist_len;
    
public:
    //
    // constructor and destructor
    //

    TReadline ();
    TReadline ( const TString & prompt );

    virtual ~TReadline ();
    
    //
    // set local data
    //

    void set_prompt   ( const TString & str ) { _prompt   = str; }
    void set_hist_len ( int             n   ) { _hist_len = n; }

    //
    // stream management
    //

    virtual bool eof () const { return _eof; }
    
    //
    // IO
    //

    // write string 
    virtual int puts ( const TString & string );

    // read string
    virtual int gets ( TString & string );
};

////////////////////////////////////////////////////
//
// baseclass defining interface
//
class TFile : public TIO
{
protected:
    // true if file is open
    bool  _is_open;

public:
    //
    // constructor and destructor
    //

    TFile ();

    TFile ( const TString & fname, const TString & mode );

    virtual ~TFile ();
    
    //
    // file management
    //

    virtual int open  ( const TString & fname, const TString & mode );
    virtual int close ();

    virtual bool is_open () const { return _is_open; }
    
protected:
    //
    // internal functions to open and close files
    //

    virtual int _open  ( const TString & fname, const TString & mode ) = 0;
    virtual int _close () = 0;
};

////////////////////////////////////////////////////
//
// class for standard, uncompressed IO
//
class TStdFile : public TFile
{
protected:
    // file handle
    FILE * _file;

public:
    //
    // constructor and destructor
    //

    TStdFile ();

    TStdFile ( const TString & fname, const TString & mode );

    virtual ~TStdFile ();
    
    //
    // file management
    //

    virtual bool eof () const;
    
    //
    // file IO
    //

    // write string 
    virtual int puts ( const TString & buf );

    // read string
    virtual int gets ( TString & buf );

protected:
    //
    // internal functions to open and close files
    //

    virtual int _open  ( const TString & fname, const TString & mode );
    virtual int _close ();
};

#ifdef USE_ZLIB
////////////////////////////////////////////////////
//
// class for gzip compressed files
//
class TGZFile : public TFile
{
protected:
    // file handle
    gzFile  _file;

public:
    //
    // constructor and destructor
    //

    TGZFile ();

    TGZFile ( const TString & fname, const TString & mode );

    virtual ~TGZFile ();
    
    //
    // file management
    //

    virtual bool eof () const;
    
    //
    // file IO
    //

    // write string 
    virtual int puts ( const TString & buf );

    // read string
    virtual int gets ( TString & buf );

protected:
    //
    // internal functions to open and close files
    //

    virtual int _open  ( const TString & fname, const TString & mode );
    virtual int _close ();
};
#endif


#ifdef USE_BZ2
////////////////////////////////////////////////////
//
// class for bzip2 compressed files
//
class TBZFile : public TFile
{
protected:
    // file handle
    FILE    * _file;
    BZFILE  * _bzf;

    // true of reading, false if writing to file
    bool      _reading;
    
    // indicates eof
    bool      _eof;
    
    // internal buffer for reading
    char      _buffer[10240];
    int       _pos, _len;

    // data for bzip2
    int       _err;
    uchar     _unused[ BZ_MAX_UNUSED ];
    int       _nunused;
    
public:
    //
    // constructor and destructor
    //

    TBZFile ();

    TBZFile ( const TString & fname, const TString & mode );

    virtual ~TBZFile ();
    
    //
    // file management
    //

    virtual bool eof () const;
    
    //
    // file IO
    //

    // write string 
    virtual int puts ( const TString & buf );

    // read string
    virtual int gets ( TString & buf );

protected:
    //
    // internal functions to open and close files
    //

    virtual int _open  ( const TString & fname, const TString & mode );
    virtual int _close ();
};
#endif

////////////////////////////////////////////////////
//
// File handler factory
// - returns correct file handle for given file
//
class TFileFactory
{
public:
    //
    // constructor and destructor
    //

    virtual ~TFileFactory () {}
    
    //
    // analyses filename and return appropriate file-handler
    //

    virtual TFile * open ( const TString & fname, const TString & mode ) const;
};

// default file factory
extern TFileFactory  file_fac;

}// namespace

#endif // __TFILE_HH
