#ifndef __TGRAPHOBJECT_HH
#define __TGRAPHOBJECT_HH
//
// Project : lib3D
// File    : TGraphObject.hh
// Purpose : baseclass for all graphical objects
// Author  : Ronald Kriemann
//
// arch-tag: b0b2f7be-4a45-4065-b27e-753545ded359
//

#include "TSLL.hh"

#include "TVector.hh"
#include "TTexture.hh"
#include "TRay.hh"

namespace RGL
{

// forward decl.
class TPolygon;

//
// define basic interface for alle graphical objects
//
class TGraphObject
{
public:
    //
    // datastructure for information about
    // last intersecting ray
    //
    typedef struct
    {
        ulong  id;
        real   distance;
        real   intersected;
    } lr_info_t;
    
protected:
    // transformation matrix: global to local coordinates
    // (transformation of normal is computed by
    //  using transposed upper 3x3 sub-matrix)
    TMatrix     _trans;

    // the texture of this object
    TTexture  * _texture;

    // ID of last tested ray (avoid multiple tests)
    lr_info_t   _lr_info;
    
public:
    //////////////////////////////////////////////////
    //
    // constructor and destructor
    //

    TGraphObject ()               { _lr_info.id = (ulong) -1; _texture = NULL; }
    TGraphObject ( TTexture * t ) { _lr_info.id = (ulong) -1; _texture = NULL; set_texture( t ); }

    virtual ~TGraphObject () { if ( _texture != NULL ) _texture->free_ref(); }

    //////////////////////////////////////////////////
    //
    // access local methods
    //

    TTexture *       texture ()       { check_texture(); return _texture; }
    const TTexture * texture () const { return _texture; }

    void set_texture ( TTexture * t )
    {
        if ( _texture != NULL ) _texture->free_ref();
        _texture = t;
        if ( _texture != NULL ) _texture->add_ref();
    }
    
    //////////////////////////////////////////////////
    //
    // methods for transforming
    //

    // move object
    virtual void move ( const TVector & v )
    {
        for ( uint i = 0; i < 3; i++ )
            _trans(i,3) -= _trans(i,0)*v[0] + _trans(i,1)*v[1] + _trans(i,2)*v[2];
    }

    // rotate object
    virtual void rotate ( const TVector & axis, real angle );

    // scale object (each axis individually)
    virtual void scale ( const TVector & s )
    {
        for ( uint i = 0; i < 3; i++ )
        {
            real c = 1.0 / s[i];
            
            _trans(0,i) *= c;
            _trans(1,i) *= c;
            _trans(2,i) *= c;
        }// for
    }
    // each axis the same
    virtual void scale ( real s ) { scale( TVector( s, s, s ) ); }

    // transform by given matrix
    virtual void transform ( const TMatrix & m )
    {
        TMatrix T(m);

        T.invert();
        _trans.mul_right( T );
    }
        
    //////////////////////////////////////////////////
    //
    // raytracing methods
    //

    // compute intersection with ray
    // - return true if intersected, else false
    virtual bool intersect ( const TRay & ray, real    & dist ) const;
    //virtual bool intersect ( const TRay & ray, TIStack & stack ) const;

    // compute normal at point
    virtual bool calc_normal ( const TVector & point,
                               TVector       & normal ) const;

    //////////////////////////////////////////////////
    //
    // misc. methods
    //

    // store bounding box defined by xy, xz and yz planes
    // in given vectors
    // - return false of bb is not defined (e.g. infinite object)
    virtual bool bounding_box ( TVector & min, TVector & max ) const;

    // print object
    virtual void print () const;
    
    //////////////////////////////////////////////////
    //
    // last ray management
    //

    // set info about last ray
    void set_last_ray ( const TRay & ray, real dist, bool inter )
    {
        _lr_info.id          = ray.id();
        _lr_info.distance    = dist;
        _lr_info.intersected = inter;
    }

    // check id of last ray
    bool is_last_ray  ( const TRay & ray ) const { return (_lr_info.id == ray.id()); }

    // get distance of last intersection point
    real last_ray_distance () const { return _lr_info.distance; }

    // get intersection status of last ray
    bool last_ray_intersected () const { return _lr_info.intersected; }
    
    //////////////////////////////////////////////////
    //
    // polygon renderer methods
    //

    // tesselate object (approx. with piecewise polygonal surface)
    virtual void tesselate ( TSLL< TPolygon * > & list ) const;

    //////////////////////////////////////////////////
    //
    // virtual constructors
    //

    virtual TGraphObject * clone () const { return new TGraphObject; }
    virtual TGraphObject * copy  () const;
    
protected:
    
    //////////////////////////////////////////////////
    //
    // check/build texture
    //

    void check_texture ()
    {
        if ( _texture == NULL )
        {
            _texture = new TTexture;
            _texture->add_ref();
        }// if
    }
    
    //////////////////////////////////////////////////
    //
    // matrix transformation
    //

    void trans_ray_to_obj ( const TRay & source, TRay & dest ) const
    { source.transform( _trans, dest ); }
    
    void trans_to_world ( const TVector & v, TVector & u ) const
    {
        TMatrix  T( _trans );
        T.invert().mul_vec( v, u );
    }
    
    void trans_to_obj   ( const TVector & v, TVector & u ) const
    { _trans.mul_vec( v, u ); }

    void trans_normal   ( const TVector & v, TVector & u ) const
    { _trans.mul_vec3( v, u, true ); u.normalise(); }
};

} // namespace

#endif  // __TGRAPHOBJECT_HH
