#ifndef __TOCTTREEINTERSECT_HH
#define __TOCTTREEINTERSECT_HH
//
// Project : lib3D
// File    : TOctTreeIntersect.hh
// Purpose : class for intersection tests based on octtrees
// Author  : Ronald Kriemann
//
// arch-tag: 610125b0-e75c-4fae-824f-cab5c937ebcf
//

#include "TArray.hh"
#include "TIntersect.hh"

namespace RGL
{

// forward declaration
class TNode;

//
// basic intersection test (test all objects)
//
class TOctTreeIntersect : public TIntersect
{
public:
    //
    // represents node in octtree
    // - sons are ordered like:
    //   lower part:  0 1
    //                2 3
    //   upper part:  4 5
    //                6 7
    // 
    //
    
    class TNode
    {
    protected:
        // ID of this node
        ulong                   _id;
    
        // bounding box
        TVector                 _min, _max, _middle, _diam;
    
        // sons
        TNode                 * _sons[8];
    
        // list of objects (empty in non-leafs)
        TSLL< TGraphObject * >  _objects;
    
    public:
        //////////////////////////////////////
        //
        // constructor and destructor
        //
    
        TNode ( const TVector & min_bb,
                const TVector & max_bb,
                uint            id = 0 )
                : _id(id), _min(min_bb), _max(max_bb)
        {
            for ( uint i = 0; i < 8; i++ ) _sons[i] = NULL;
            _middle.set( 0.5, min_bb, 0.5, max_bb );
            _diam.set( 1.0, _middle, -1.0, min_bb );
        }
    
        ~TNode ()
        { for ( uint i = 0; i < 8; i++ ) delete _sons[i]; }
    
        //////////////////////////////////////
        //
        // access data
        //
    
        ulong   id () const { return _id; }
        ulong & id ()       { return _id; }
    
        const TVector & min    () const { return _min; }
        const TVector & max    () const { return _max; }
        const TVector & middle () const { return _middle; }
        const TVector & diam   () const { return _diam; }
    
        TNode       * son ( uint i )       { assert( i < 8 ); return _sons[i]; }
        const TNode * son ( uint i ) const { assert( i < 8 ); return _sons[i]; }
    
        void remove_son ( uint i ) 
        { assert( i < 8 ); delete _sons[i]; _sons[i] = NULL; }
    
        TSLL< TGraphObject * >       & objects ()       { return _objects; }
        const TSLL< TGraphObject * > & objects () const { return _objects; }
    
        //////////////////////////////////////
        //
        // octtree-functionality
        //
    
        // return parent-ID
        ulong parent_id () const { return _id >> 3; }
    
        // return neighbour-ID
        // - n selects neighbour : left=0, right, down, up, front, back=5
        ulong neighbour_id ( uint n ) const;
    
        // build all sons
        void build_sons ();
    
        // check if given object is inside boundingbox
        bool is_inside ( const TGraphObject * obj ) const;
    
        // check if given point is inside boundingbox
        bool is_inside ( const TVector & v ) const;
    
        // return if leaf (no sons)
        bool is_leaf () const;
    
        // compute point of intersection with bounding-box
        bool bb_intersect ( const TRay & ray,
                            TVector    & hit_point ) const;
    
        //////////////////////////////////////////////
        //
        // output
        //
    
        // stream output
        void print ( std::ostream & os, uint ofs ) const;
    };

protected:
    // root of octtree
    TNode                 * _root;

    // maximal depth in tree
    uint                    _max_depth;
    
    // list of infinite objects
    TSLL< TGraphObject * >  _inf_objects;

    // maximal number of nodes in a leaf
    uint                    _max_leaf_obj;

    // array for ID -> node conversion
    TArray< TNode * >       _nodes;
    
public:
    //////////////////////////////////////////////
    //
    // constructor and destructor
    //

    TOctTreeIntersect ( const TWorld * w = NULL );

    virtual ~TOctTreeIntersect ();
    
    //////////////////////////////////////////////
    //
    // access internal data
    //

    virtual void set_world ( const TWorld * w );
    
    //////////////////////////////////////////////
    //
    // intersection test
    //

    // return intersected object and store distance
    // - return NULL of no such object can be found
    // - define viewing-frustrum by front and back
    //   (< 0 means not defined [infinity]) 
    virtual TGraphObject * intersect ( const TRay & ray,
                                       real       & distance,
                                       real         front,
                                       real         back ) const;

    //////////////////////////////////////////////
    //
    // output
    //

    // print octtree (inclusiv given objects) into vrml-file
    virtual void print_vrml ( const TSLL< TGraphObject * > * objects,
                              ulong                          pid = 0 ) const;

    // stream output
    virtual void print ( std::ostream & os ) const;
    
protected:
    // build octtree out of world
    void build_tree ( const TWorld * world );

    // divide node/object-list and build tree
    void divide ( TNode                  * node,
                  TSLL< TGraphObject * > & objects,
                  uint                     level ) const;

    // get start-node for search
    TNode * get_start_node ( const TRay & ray, const TNode * root, TVector & point ) const;

    // proceed along ray and return next node in octtree
    TNode * next_node ( const TRay & ray, const TNode * node, TVector & point ) const;

    // do intersection test with ray and node-objects
    TGraphObject * intersect ( const TNode * node,
                               const TRay  & ray,
                               real        & hit_distance,
                               real          front,
                               real          back ) const
    {
        assert( node != NULL );
        return intersect_list( node->objects(), ray, hit_distance, front, back );
    }

    // do intersection test with ray and infinite objects
    TGraphObject * intersect_inf ( const TRay & ray,
                                   real       & hit_distance,
                                   real         front,
                                   real         back ) const
    {
        return intersect_list( _inf_objects, ray, hit_distance, front, back );
    }

    // do intersection test with ray and objects in given list
    TGraphObject * intersect_list ( const TSLL< TGraphObject * > & objects,
                                    const TRay                   & ray,
                                    real                         & hit_distance,
                                    real                           front,
                                    real                           back ) const;
};

}// namespace

#endif // __TINTERSECT_HH
