#ifndef __TPARSER_HH
#define __TPARSER_HH
//
// Project : lib3D
// File    : TParser.hh
// Purpose : class for reading world-descriptions
// Author  : Ronald Kriemann
//
// arch-tag: 213c1bcf-7027-4e5a-8433-1db7f51258c1
//

#include <stdio.h>

#include "defines.hh"
#include "TString.hh"
#include "TWorld.hh"
#include "TCamera.hh"
#include "TSLL.hh"
#include "TFile.hh"

// external definition of text-position
extern int text_row, text_col;

namespace RGL
{

//
// different return types
//
enum { TYPE_UNDEF, TYPE_NUMBER, TYPE_BOOL,
       TYPE_STRING, TYPE_VECTOR,
       TYPE_GRAPH_OBJECT, TYPE_SPHERE,
       TYPE_TRANSFORM,
       TYPE_COLOUR, TYPE_COL_COEFF, TYPE_TEXTURE,
       TYPE_LIGHT
};

enum { OP_UNDEF,
       OP_ASSIGN, OP_ADDASSIGN, OP_SUBASSIGN,
       OP_NEG,
       OP_ADD, OP_SUB, OP_PREINC, OP_POSTINC, OP_PREDEC, OP_POSTDEC,
       OP_MUL, OP_DIV,
       OP_LT, OP_GT, OP_LE, OP_GE, OP_NE, OP_EQ,
       OP_AND, OP_OR, OP_NOT,
       OP_SQRT, OP_SIN, OP_COS, OP_TAN, OP_ASIN, OP_ACOS, OP_ATAN,
       OP_NORM2,
       OP_STMT,
       OP_WHILE,
       OP_ADD_WORLD
};

//
//
//
class TParser
{
public:

    //////////////////////////////////////////////////
    //
    // local datatypes
    //

    //
    // type for the text-position
    //
    class TTextPos
    {
    protected:
        int _row, _col;

    public:
        TTextPos ( int r, int c ) : _row(r), _col(c) {}

        void set_pos ( int r, int c ) { _row = r; _col = c; }
        
        friend std::ostream & operator << ( std::ostream & os,
                                            const TTextPos & pos )
        { return (os << "(" << pos._row << ", " << pos._col << ")"); }
    };
    
    //
    // basic type for storing and exchanging data
    //
    
    class TNodeData
    {
    public:
        TNodeData () {}

        virtual ~TNodeData () { dealloc(); }
        
        virtual uint type () const { return TYPE_UNDEF; }
        virtual bool is_type  ( uint t ) const { return t == TYPE_UNDEF; }

        // create copy of data
        virtual TNodeData * copy () const { return new TNodeData; }

        // remove data from memory
        virtual void dealloc () {}
    };

    //
    // class for a variable
    //
    
    class TVariable
    {
    protected:
        // data associated with variable
        TNodeData   * _data;

        // name of this variable
        TString       _name;

        // is variable constant
        bool          _constant;
        
    public:
        // constructor and destructor
        TVariable ( const TString & name )
                : _data(NULL), _name(name), _constant(false)
        {}

        // destructor
        virtual ~TVariable () { delete _data; }
        
        // access data
        const TString & name     () const { return _name; }
        bool            is_const () const { return _constant; }

        void set_constant () { _constant = true; }
        
        // get/set data
        TNodeData       * data ()       { return _data; }
        const TNodeData * data () const { return _data; }

        void set_data ( TNodeData * data, const TTextPos & pos );
        
        // return type of node
        uint type () const { return (_data != NULL ? _data->type() : TYPE_UNDEF); }
        virtual bool is_type ( uint t ) const
        { return (_data != NULL ? _data->is_type(t) : (t == TYPE_UNDEF)); }
        
        // print node
        void print ( const TTextPos & pos ) const;
    };
    
    //
    // baseclass for nodes in the parsetree
    //
    class TNode
    {
    protected:
        // text-position, this node occurs
        TTextPos  _pos;
        
    public:
        // constructor
        TNode () : _pos( text_row, text_col ) {}

        // destructor
        virtual ~TNode () {}

        // access position
        const TTextPos & pos () const { return _pos; }
        void update_pos ( int r, int c ) { _pos.set_pos( r, c ); }

        // return/check type of node
        virtual uint type () const { return TYPE_UNDEF; }
        virtual bool is_type ( uint t ) const { return t == type(); }
        
        // eval node and return result
        virtual TNodeData * eval () const { return NULL; }
    };
    
    //
    // a node associated with a variable
    //

    class TVarNode : public TNode
    {
    protected:
        // associated variable
        TVariable   * _var;
        
    public:
        // constructor and destructor
        TVarNode ( TVariable * var ) : _var(var) {}

        // access variable
        TVariable * variable () { return _var; }
        
        // return type of node
        virtual uint type () const { return (_var != NULL ? _var->type() : TYPE_UNDEF); }
        virtual bool is_type ( uint t ) const
        { return (_var != NULL ? _var->is_type(t) : (t == TYPE_UNDEF)); }
        
        // eval node and return result
        virtual TNodeData * eval () const;
    };

    //
    // unary operator
    //

    class TUnaryNode : public TNode
    {
    protected:
        // only son of node
        TNode * _son;

    public:
        // constructor
        TUnaryNode ( TNode * son ) : _son(son) { assert( _son != NULL ); }

        // destructor
        virtual ~TUnaryNode () { delete _son; }
    };

    //
    // binary operator
    //
    
    class TBinaryNode : public TNode
    {
    protected:
        // two sons of node
        TNode * _son0, * _son1;

    public:
        // constructor
        TBinaryNode ( TNode * son0, TNode * son1 )
                : _son0(son0), _son1(son1)
        { assert(( _son0 != NULL ) && ( _son1 != NULL )); }

        // destructor
        virtual ~TBinaryNode () { delete _son0; delete _son1; }
    };

protected:

    // list of variables
    TSLL< TVariable * >  _var_list;

    // world to build
    TWorld             * _world;

    // camera definition
    TVector              _cam_orig, _cam_dir, _cam_up;

    // file input
    TIO                * _input;
    
public:

    /////////////////////////////////////////////
    //
    // constructor and destructor
    //

    TParser ( TWorld * w = NULL );

    ~TParser ();

    //
    // access data
    //

    void apply_camera ( TCamera * camera ) const;
    
    void set_camera    ( const TVector & orig, const TVector & dir );
    void set_camera_up ( const TVector & up );
    
    /////////////////////////////////////////////
    //
    // read/parse file
    //

    void parse ( TIO * input );

    // read from input (for scanner)
    int readbuf ( char * buf, int max_size );
    
    /////////////////////////////////////////////
    //
    // build parse-tree
    //

    // convert name-string into variable
    TNode * name_to_var ( const char * name );
    TNode * name_to_var ( const TString & name ) { return name_to_var( name.c_str() ); }

    // build new node in parse-tree
    TNode * new_node ( uint node_type, TNode * n1 );
    TNode * new_node ( uint node_type, TNode * n1, TNode * n2 );

    // node creation
    TNode * new_datanode ( uint dtype, void * data );
    TNode * new_vector   ( TNode * n0, TNode * n1, TNode * n2 ) const;
    TNode * new_funccall ( const char * name, TNode * args ) const;
    TNode * new_stmt     ( TNode * n ) const;
    TNode * new_block    ( TNode * n ) const;
    TNode * new_while    ( TNode * stop, TNode * block ) const;
    TNode * new_if       ( TNode * cond, TNode * block1, TNode * block2 ) const;

    // set variable as constant
    void set_constant ( TNode * var ) const;
    
    // append node2 to statement-list
    TNode * append_stmt ( TNode * node1, TNode * node2 ) const;

    // append node2 to argument-list
    TNode * append_arg ( TNode * n1, TNode * n2 ) const;
    
    // eval given parse-tree and remove it
    void eval_tree ( TNode * n ) const;
};

////////////////////////////////////////////////
//
// special nodedata-types
//

class TNDNumber : public TParser::TNodeData
{
protected:
    real  _number;

public:
    TNDNumber ( real n ) : _number(n) {}

    real & number ()       { return _number; }
    real   number () const { return _number; }
    
    virtual uint type () const;
    virtual bool is_type  ( uint t ) const;
    
    virtual TNodeData * copy () const;
};

class TNDBool : public TParser::TNodeData
{
protected:
    bool  _boolean;

public:
    TNDBool ( bool b ) : _boolean(b) {}

    bool & boolean ()       { return _boolean; }
    bool   boolean () const { return _boolean; }
    
    virtual uint type () const;
    virtual bool is_type  ( uint t ) const;
    
    virtual TNodeData * copy () const;
};

class TNDString : public TParser::TNodeData
{
protected:
    TString  _string;

public:
    TNDString ( const TString & s ) : _string(s) {}

    TString &       string ()       { return _string; }
    const TString & string () const { return _string; }
    
    virtual uint type () const;
    virtual bool is_type  ( uint t ) const;
    
    virtual TNodeData * copy () const;
};

class TNDVector : public TParser::TNodeData
{
protected:
    TVector  _vec;

public:
    TNDVector ( const TVector & v ) : _vec(v) {}

    TVector &       vector ()       { return _vec; }
    const TVector & vector () const { return _vec; }
    
    virtual uint type () const;
    virtual bool is_type  ( uint t ) const;
    
    virtual TNodeData * copy () const;
};

class TNDGraphObject : public TParser::TNodeData
{
protected:
    TGraphObject  * _obj;
    
public:
    TNDGraphObject ( TGraphObject * obj ) : _obj(obj) {}

    ~TNDGraphObject () { delete _obj; }
    
    virtual uint type () const;
    virtual bool is_type  ( uint t ) const;
    
    virtual TGraphObject * graph_object () { return _obj; }

    virtual void reset_object ();

    virtual TNodeData * copy () const;
};

////////////////////////////////////////////////
//
// global error handling
//
void parse_warn ( const TString           & msg,
                  const TParser::TTextPos & pos );

void parse_error ( const TString           & msg,
                   const TParser::TTextPos & pos );

void parse_error ( const TString & msg );

}// namespace

#endif // __TPARSER_HH
