#ifndef __TRAYTRACER_HH
#define __TRAYTRACER_HH
//
// Project : lib3D
// File    : TRaytracer.hh
// Purpose : baseclass for all raytracers
// Author  : Ronald Kriemann
//
// arch-tag: 3ca00c3e-e7ec-4ce4-b1bf-3e19f629820d
//

#include "TWorld.hh"
#include "TImage.hh"
#include "TRay.hh"
#include "TCamera.hh"
#include "TIntersect.hh"
#include "THPDesc.hh"
#include "TProgressBar.hh"

namespace RGL
{

//
// the name says it all: rendering via raytracing
//
class TRaytracer
{
public:
    //
    // describes a tile (part of an image)
    //
    typedef struct
    {
        uint  x_ofs, y_ofs;    // start pos
        uint  width, height;   // dimension of tile
        uint  iwidth, iheight; // dimension of global image
    } TTileDesc;

    //
    // statistics of raytracing
    //
    typedef struct
    {
        uint    no_of_rays;
        uint    reflected_rays, refracted_rays;
        uint    deepest_level;
    } TRTStat;
    
protected:
    // camera defining view
    TCamera       * _camera;

    // world of objects to raytrace
    TWorld        * _world;

    // intersection object
    TIntersect    * _intersect;
    
    // background colour
    TColour         _background;
    
    // maximal depth of trace-tree
    uint            _max_depth;

    // minimal ray-intensity
    real            _min_intensity;

    // compute shadows
    bool            _with_shadow;
    
    // some statistics
    mutable TRTStat _stat;
    
public:
    ///////////////////////////////////////////////
    //
    // constructor and destructor
    //

    TRaytracer ();

    virtual ~TRaytracer ();

    ///////////////////////////////////////////////
    //
    // access local data
    //

    TWorld  * world  () { return _world;  }
    
    const TWorld  * world  () const { return _world;  }
    const TCamera * camera () const { return _camera; }

    virtual void set_camera ( TCamera * c ) { _camera = c; }
    virtual void set_world  ( TWorld  * w ) { _world  = w; _intersect->set_world( w ); }

    virtual void set_background ( const TColour & c ) { _background = c; }

    uint no_of_rays     () const { return _stat.no_of_rays; }
    uint reflected_rays () const { return _stat.reflected_rays; }
    uint refracted_rays () const { return _stat.refracted_rays; }
    uint deepest_level  () const { return _stat.deepest_level; }

    virtual void set_max_depth     ( uint n ) { _max_depth     = n; }
    virtual void set_min_intensity ( real f ) { _min_intensity = f; }
    
    ///////////////////////////////////////////////
    //
    // rendering
    //

    // render world to printer
    virtual void render_world ( TImage & img, TProgressBar * bar = NULL ) const;

    // render part of image
    virtual void render_tile ( const TTileDesc & tile,
                               TImage          & img,
                               TProgressBar    * bar = NULL ) const = 0;
    
protected:
    ///////////////////////////////////////////////
    //
    // raytracing methods
    //

    // recursive raytracing
    void trace_ray ( const TRay & ray,
                     uint         depth,
                     real         intensity,
                     TColour    & colour ) const;

    // calculate intersection of ray with objects in world
    TGraphObject * intersect ( const TRay & ray, real & distance ) const;

    // shade object at given point
    void shade ( const THPDesc & hp,
                 const TRay    & ray,
                 uint            depth,
                 real            intensity,
                 TColour       & colour ) const;

    // compute reflected/refracted ray
    void comp_reflected ( const THPDesc & hp,
                          const TVector & incident,
                          TRay          & ray ) const;

    bool comp_refracted ( const TGraphObject * medium0,
                          const TGraphObject * medium1,
                          const THPDesc      & hp,
                          const TVector      & incident,
                          TRay               & ray ) const;

    // return true if minimal intensity or maximal depth was reached
    bool reached_stop ( real weight, uint depth ) const
    {
        if ((depth > _max_depth) || (weight < _min_intensity))
            return true;
        else
            return false;
    }

    ///////////////////////////////////////////////
    //
    // misc. methods
    //

    // reset statistical information
    void reset_stat () const;
};

} // namespace

#endif  // __TRAYTRACER_HH
