#ifndef __TIMER_HH
#define __TIMER_HH
//
// Project : misc
// File    : timer.hh
// Purpose : class for a timer (for speed tests)
// Author  : Ronald Kriemann
//
// arch-tag: c5fb5292-c55b-46b7-8f04-c29e9f09bb5a
//

#include <sys/time.h>
#include <sys/resource.h>

#ifdef SUNOS
#include <sys/lwp.h>
#endif
    
#include <unistd.h>

#include <iostream>

namespace RGL
{

// type for different timer-measurment
typedef enum { REAL_TIME, CPU_TIME, USER_TIME, SYSTEM_TIME }  timetype_t;

//
// the timer class
//
class TTimer 
{
protected:
    // start and overall time
    double          _start, _sum;

    // temp structs for measuring different times
    mutable struct timeval  _timeval_data;
#ifdef SUNOS
    mutable struct lwpinfo  _lwpinfo_data;
#else
    mutable struct rusage   _rusage_data;
#endif
    
    // what kind of time we should stop
    timetype_t      _type;
    
public:
    /////////////////////////////////////////////////
    //
    // constructor
    //

    TTimer ( timetype_t type = CPU_TIME ) : _start(0), _sum(0), _type(type) {}
    
    /////////////////////////////////////////////////
    //
    // compute times
    //
    
    // start timer, resets time-sum
    TTimer & start    () { _start = system_time(); _sum = 0.0; return *this; }

    // pause timer, add elapsed time to sum
    TTimer & pause    () { _sum  += system_time() - _start; _start = 0.0; return *this; }

    // continue timer, set start time to current time
    TTimer & cont     () { _start = system_time(); return *this; }

    // returns elapsed time without stopping timer
    float elapsed () const { return (_start > 0 ? _sum + system_time() - _start : _sum); }

    /////////////////////////////////////////////////
    //
    // output
    //

    // set type of time to measure and return *this (for output)
    TTimer & real () { _type = REAL_TIME;   return *this; }
    TTimer & cpu  () { _type = CPU_TIME;    return *this; }
    TTimer & user () { _type = USER_TIME;   return *this; }
    TTimer & sys  () { _type = SYSTEM_TIME; return *this; }
    
    // stream output (only the time)
    friend std::ostream & operator << ( std::ostream & os, const TTimer & timer );

protected:
    /////////////////////////////////////////////////
    //
    // internal functions
    //

    // get time of system (usertime or real)
    double system_time () const;
};

} // namespace

#endif  // __TIMER_HH
