#ifndef __TVRML_HH
#define __TVRML_HH
//
// Project : lib3D
// File    : TVRML.hh
// Purpose : class for a vrml-printer
// Author  : Ronald Kriemann
//
// arch-tag: b30d31eb-e720-4c2c-b938-677c6800485a
//

#include <stdio.h>

#include "TString.hh"
#include "TTexture.hh"

namespace RGL
{

//
// produces VRML1.0 for now
//
class TVRML
{
public:
    // different versions of vrml
    enum TVersion { VRML1, VRML2, VRML97 };
    
protected:
    // the file we save in
    FILE        * _file;
    TString       _filename;

    // version of vrml to write in (default is vrml1)
    TVersion      _version;

    // actual type of drawing
    TTexture      _texture;

    // is printer active (aka did we started printing)
    bool          _active;

    // for lineTo (old pos)
    TVector       _old_pos;

    // for indentation
    uint          _ofs;
    
public:
    //////////////////////////////////////////////////
    //
    // constructor and destructor
    //

    TVRML ();

    virtual ~TVRML ();
    
    //////////////////////////////////////////////////
    //
    // access local data
    //

    bool active () const { return _active; }

    TTexture &       texture ()       { return _texture; }
    const TTexture & texture () const { return _texture; }
    
    //////////////////////////////////////////////////
    //
    // methods for drawing
    //

    // begin and end drawing
    virtual void begin ();
    virtual void begin ( const char * name );

    virtual void end ();

    //
    // 3D - drawing
    //

    virtual void draw_point ( const TVector & v );
    virtual void draw_line  ( const TVector & orig, const TVector & dest );

    // special methods for drawing polygons
    virtual void draw_triangle ( const TVector & x1, const TVector & x2,
                                 const TVector & x3 );
    virtual void fill_triangle ( const TVector & x1, const TVector & x2,
                                 const TVector & x3 );
    
    virtual void draw_rect ( const TVector & x1, const TVector & x2,
                             const TVector & x3, const TVector & x4 );
    virtual void fill_rect ( const TVector & x1, const TVector & x2,
                             const TVector & x3, const TVector & x4 );

    // draw polygon by giving an array of points
    virtual void draw_poly ( TVector * points, int count );
    virtual void fill_poly ( TVector * points, int count );

    // draw a sphere
    virtual void draw_sphere ( const TVector & center, real radius );
    
    // don't draw, just set old_pos
    virtual void move_to ( const TVector & v ) { _old_pos = v; }

    // draw line (old_pos, v) and set old_pos
    virtual void line_to ( const TVector & v ) { draw_line( _old_pos, v ); _old_pos = v; }

    //////////////////////////////////////////////////
    //
    // methods for setting up style
    //

    // set up texture for all following objects
    virtual void set_texture ( const TTexture & t );

    // for simplicity: set color-part of texture
    virtual void set_colour ( const TColour & c ) { _texture.set_colour( c ); }
    
private:
    //////////////////////////////////////////////////
    //
    // misc helping-methods
    //

    void begin_section ( const char * name = NULL );
    void end_section ();

    void write ( const char * text );

    void write_colour ();
};

} // namespace

#endif  // __TVRML_HH
