#ifndef __TVECTOR_HH
#define __TVECTOR_HH
//
// Project : lib3D
// File    : TVector.hh
// Purpose : class for a 3D vector
// Author  : Ronald Kriemann
//
// arch-tag: fdad9b23-226e-4f24-8c63-f31ce7a88bd4
//

#include <assert.h>

#include <cmath>
#include <iostream>

#include "defines.hh"

namespace RGL
{

// forward declaration
class TMatrix;

//
// vector in R^3
//
class TVector
{
    // some friends
    friend class TMatrix;

protected:
    // the coeff.
    real  _data[3];

public:
    /////////////////////////////////////////////////////
    //
    // constructor and destructor
    //

    TVector () 
    { _data[0] = 0.0; _data[1] = 0.0; _data[2] = 0.0; }
    
    TVector ( const TVector & v ) { *this = v; }

    TVector ( real x, real y, real z )
    { _data[0] = x; _data[1] = y; _data[2] = z; }


    ////////////////////////////////////////////////////
    //
    // access coefficients
    //

    real & operator [] ( uint i )       { assert(i<3); return _data[i]; }
    real   operator [] ( uint i ) const { assert(i<3); return _data[i]; }

    real & x ()       { return _data[0]; }
    real & y ()       { return _data[1]; }
    real & z ()       { return _data[2]; }

    real   y () const { return _data[1]; }
    real   x () const { return _data[0]; }
    real   z () const { return _data[2]; }
    
    ////////////////////////////////////////////////////
    //
    // usual vector methods/operators
    //

    TVector & operator = ( const TVector & v )
    {
        _data[0] = v._data[0]; _data[1] = v._data[1]; _data[2] = v._data[2];
        return *this;
    }

    TVector & scale ( real a )
    {
        _data[0] *= a; _data[1] *= a; _data[2] *= a;
        return *this;
    }

    TVector & set ( real a )
    {
        _data[0] = a; _data[1] = a; _data[2] = a;
        return *this;
    }

    TVector & set ( real x, real y, real z )
    {
        _data[0] = x; _data[1] = y; _data[2] = z;
         return *this;
    }

    TVector & set ( real a, const TVector & v )
    {
        _data[0] = a * v._data[0];
        _data[1] = a * v._data[1];
        _data[2] = a * v._data[2];
        return *this;
    }
    
    TVector & set ( real a, const TVector & v,
                    real b, const TVector & w )
    {
        _data[0] = a * v._data[0] + b * w._data[0];
        _data[1] = a * v._data[1] + b * w._data[1];
        _data[2] = a * v._data[2] + b * w._data[2];
        return *this;
    }

    TVector & add ( real a, const TVector & v )
    {
        _data[0] += a * v._data[0];
        _data[1] += a * v._data[1];
        _data[2] += a * v._data[2];
        return *this;
    }

    TVector & add ( real a, const TVector & v,
                    real b, const TVector & w )
    {
        _data[0] += a * v._data[0] + b * w._data[0];
        _data[1] += a * v._data[1] + b * w._data[1];
        _data[2] += a * v._data[2] + b * w._data[2];
        return *this;
    }

    // dot product
    friend real dot ( const TVector & v, const TVector & w )
    {
        return (v._data[0] * w._data[0] +
                v._data[1] * w._data[1] +
                v._data[2] * w._data[2] );        
    }

    // euclidean norm
    real norm2 () const
    { return sqrt(_data[0]*_data[0] + _data[1]*_data[1] + _data[2]*_data[2] ); }
    
    friend real norm2 ( const TVector & v ) { return v.norm2(); }
        
    // cross product : *this = x \times y
    TVector & set_cross ( const TVector & x, const TVector & y )
    {
        _data[0] = (x._data[1] * y._data[2]) - (x._data[2] * y._data[1]);
        _data[1] = (x._data[2] * y._data[0]) - (x._data[0] * y._data[2]);
        _data[2] = (x._data[0] * y._data[1]) - (x._data[1] * y._data[0]);
        return *this;
    }

    // normalise wrt. euclidean-norm (return norm)
    real normalise ()
    {
        real s = norm2();
        
        if ( s != 0.0 )
            scale( 1.0 / s );
        
        return s;
    }

    ///////////////////////////////////////////////////
    //
    // misc methods and operators
    //

    // stream output
    friend std::ostream & operator << ( std::ostream & os, const TVector & v );
};

///////////////////////////////////////////////////////
//
// some standard vectors
//

extern const TVector VEC_X;
extern const TVector VEC_Y;
extern const TVector VEC_Z;
extern const TVector VEC_0;

} // namespace

#endif  // __TVECTOR_HH
