//
// Project : librender
// File    : TOctTreeIntersect.cc
// Purpose : class for intersection tests based on octtrees
// Author  : Ronald Kriemann
//
// arch-tag: 7fa2ce64-467d-43bc-94e7-203bd8523ba9
//

#include "TArray.hh"
#include "TVRML.hh"
#include "TTimer.hh"
#include "TSphere.hh"
#include "defines.hh"

#include "TOctTreeIntersect.hh"

namespace RGL
{

////////////////////////////////////////////////
//
// local defines
//
////////////////////////////////////////////////

// neighbour definition
#define LEFT_NB   0
#define RIGHT_NB  1
#define DOWN_NB   2
#define UP_NB     3
#define FRONT_NB  4
#define BACK_NB   5

#define INVALID_ID  ((ulong) -1)

////////////////////////////////////////////////
//
// local data
//
////////////////////////////////////////////////

// conversion of bit-position to son-position
static unsigned char bit_to_son[8] = { 2, 3, 6, 7, 0, 1, 4, 5 };

////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////
//
// TOctTreeIntersect (implementation)
//
////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////

//
// constructor and destructor
//

TOctTreeIntersect::TOctTreeIntersect ( const TWorld * w )
        : TIntersect(w), _root(NULL)
{
    _max_leaf_obj = 2;
    _max_depth    = 4;
}

TOctTreeIntersect::~TOctTreeIntersect ()
{
    delete _root;
}
    
//
// access internal data
//

void
TOctTreeIntersect::set_world ( const TWorld * w )
{
    _world = w;

    if ( verbose(3) )
    {
        TTimer  timer( REAL_TIME );

        timer.start();
        build_tree( w );
        timer.pause();

        std::cout << "(TOctTreeIntersect) set_world : time to build tree = "
                  << timer << std::endl;
    }// if
    else
        build_tree( w );
    
    if ( verbose(5) )
        print_vrml( & _world->objects(), 0 );
}

//
// intersection test
//   return intersected object and store distance
//   - return NULL of no such object can be found
//   - define viewing-frustrum by front and back
//     (< 0 means not defined [infinity]) 
//
TGraphObject *
TOctTreeIntersect::intersect ( const TRay & ray,
                               real       & distance,
                               real         front,
                               real         back ) const
{
    //
    // go through octtree and test intersection with
    // all objects along the ray
    //

    TVector        point;
    TNode        * node       = get_start_node( ray, _root, point );
    TGraphObject * hit_object = NULL;
    TVector        hp;

    while ( node != NULL )
    {
        //
        // determine intersecting object in local node
        //

        hit_object = intersect( node, ray, distance, front, back );

        if ( hit_object != NULL )
        {
            hp.set( 1.0, ray.origin(), distance, ray.direction() );

            // check whether hitpoint is really in the node, otherwise
            // a different object might be in front in another node
            // along the ray
            if ( node->is_inside( hp ) )
                break;
        }// if

        // proceed to next node in ray-direction
        node = next_node( ray, node, point );
    }// while
    
    //
    // test against infinite objects
    //

    real           inf_dist;
    TGraphObject * inf_obj;
    
    inf_obj = intersect_inf( ray, inf_dist, front, back );

    if (( inf_obj != NULL ) &&
        (( hit_object == NULL ) || ( inf_dist < distance )))
    {
        hit_object = inf_obj;
        distance   = inf_dist;
    }// if

    //
    // intersection test complete, return hit-object
    //
    
    return hit_object;
}

//
// build octtree out of world
//
void
TOctTreeIntersect::build_tree ( const TWorld * world )
{
    assert( world != NULL );
    
    // remove old tree
    if ( _root )
        delete _root;

    _inf_objects.remove_all();
    
    //
    // compute bounding box of world-objects
    //

    TSLL< TGraphObject * >  fin_objects;
    TWorld::TObjIterator    iter;
    TVector                 min_bb, max_bb;
    TVector                 omin_bb, omax_bb;
    bool                    first = true;
    
    for ( iter = world->objects().first(); ! iter.eol(); ++iter )
    {
        if ( iter()->bounding_box( omin_bb, omax_bb ) )
        {
            fin_objects.append( iter() );

            min_bb.x() = ( first ? omin_bb.x() : min( min_bb.x(), omin_bb.x() ) );
            min_bb.y() = ( first ? omin_bb.y() : min( min_bb.y(), omin_bb.y() ) );
            min_bb.z() = ( first ? omin_bb.z() : min( min_bb.z(), omin_bb.z() ) );

            max_bb.x() = ( first ? omax_bb.x() : max( max_bb.x(), omax_bb.x() ) );
            max_bb.y() = ( first ? omax_bb.y() : max( max_bb.y(), omax_bb.y() ) );
            max_bb.z() = ( first ? omax_bb.z() : max( max_bb.z(), omax_bb.z() ) );

            first = false;
        }// if
        else
            _inf_objects.append( iter() );
    }// for

    if ( verbose(3) )
        std::cout << "(TOctTreeIntersect) build_tree : "
                  << fin_objects.size() << " finite object(s), "
                  << _inf_objects.size() << " infinite object(s)" << std::endl;
    
    //
    // build root
    //
    
    _root = new TNode( min_bb, max_bb, 1 );

    if ( verbose(3) )
        std::cout << "(TOctTreeIntersect) build_tree : bounding-box = "
                  << min_bb << " - " << max_bb << std::endl;
    
    //
    // create tree recursivly
    //

    divide( _root, fin_objects, 0 );

    //
    // assign nodes to array according to ID
    //

    uint             max_n  = (1 << (3 * (_max_depth+1))) + 1;
    uint             node_n = 0;
    TSLL< TNode * >  list;

    _nodes.set_size( max_n );

    if ( verbose(3) )
        std::cout << "(TOctTreeIntersect) build_tree : maximal number of nodes = "
                  << max_n << std::endl;
    
    for ( uint i = 0; i < max_n; i++ )
        _nodes[i] = NULL;

    list.append( _root );
    while ( list.size() > 0 )
    {
        TNode * node = list.behead();

        if ( node->id() >= max_n )
            std::cout << "(TOctTreeIntersect) build_tree : node-ID bigger than max-ID : "
                      << node->id() << std::endl;
        
        _nodes[ node->id() ] = node;
        node_n++;

        if ( ! node->is_leaf() )
        {
            for ( uint i = 0; i < 8; i++ )
            {
                if ( node->son(i) != NULL )
                    list.append( node->son(i) );
            }// for
        }// if
    }// while

    if ( verbose(3) )
        std::cout << "(TOctTreeIntersect) build_tree : actual number of nodes  = "
                  << node_n << std::endl;
}

//
// divide node/object-list and build tree
//
void
TOctTreeIntersect::divide ( TNode                  * node,
                            TSLL< TGraphObject * > & objects,
                            uint                     level ) const
{
    //
    // first check if max-level or minimal object-number reached
    //
    
    if (( level >= _max_depth ) || ( objects.size() <= _max_leaf_obj ))
    {
        //
        // assign objects to node
        //

        if ( objects.size() > 0 )
            node->objects().copy( objects );
        
        return;
    }// if
    
    //
    // create all sons
    //

    node->build_sons();

    //
    // check for each object in which son it is
    //

    TSLL< TGraphObject * >::TIterator  iter;
    TArray< TSLL< TGraphObject * > >   son_objects( 8 );

    iter = objects.first();
    while ( ! iter.eol() )
    {
        TGraphObject * obj = iter();

        for ( uint i = 0; i < 8; i++ )
        {
            if ( node->son(i)->is_inside( obj ) )
                son_objects[i].append( obj );
        }// for

        ++iter;
    }// while

    //
    // proceed recursion
    //

    for ( uint i = 0; i < 8; i++ )
        divide( node->son(i), son_objects[i], level+1 );
}

//
// get start-node for search
//
TOctTreeIntersect::TNode *
TOctTreeIntersect::get_start_node ( const TRay                     & ray,
                                    const TOctTreeIntersect::TNode * root,
                                    TVector                        & point ) const
{
    //
    // first check where the origin of the ray lies,
    // outside or inside octtree
    //

    // set point to origin of ray, iff inside octtree
    if ( _root->is_inside( ray.origin() ) )
        point = ray.origin();
    else
    {
        //
        // origin lies outside of octree => determine intersection
        // of ray with octtree
        //

        if ( ! _root->bb_intersect( ray, point ) )
            return NULL;
    }// else
    
    //
    // now determine leaf with inner-point inside it
    //
    
    TNode * node = _root;

    while ( ! node->is_leaf() )
    {
#if 1
        //
        // go through each coordinate and compare whether it is left
        // or right and choose son accordingly
        //

        uint  son_no = 0;
        
        for ( uint i = 0; i < 3; i++ )
        {
            if ( point[i] > node->middle()[i] )
                son_no |= (1 << i);
        }// for

        // map bits to real son number
        node = node->son( bit_to_son[son_no] );

        assert( node->is_inside( point ) );
#else
        //
        // check against all sons, if point is inside
        // and replace node by this son
        //

        bool found = false;
            
        for ( uint i = 0; i < 8; i++ )
        {
            TNode * son = node->son(i);
            
            if ( son->is_inside( point ) )
            {
                found = true;
                node  = son;
                break;
            }// if
        }// for

        if ( ! found )
        {
            std::cout << "(TOctTreeIntersect) get_start_node : no son contains given point"
                      << std::endl;
            break;
        }// if
#endif        
    }// while

    //
    // last node is leaf and contains the inner point,
    // this is our start-node
    //
    
    return node;
}

//
// proceed along ray and return next node in octtree
//
TOctTreeIntersect::TNode *
TOctTreeIntersect::next_node ( const TRay                     & ray,
                               const TOctTreeIntersect::TNode * node,
                               TVector                        & point ) const
{
    TVector  w, d( ray.direction() );

    //
    // set up w-vector according to ray-direction
    //

    // 0 : left, down, front, 1 : right, up, back
    uint  x_nb = 0, y_nb = 0, z_nb = 0;
    bool  use_x = true, use_y = true, use_z = true;

    if      ( d.x() == 0.0 ) use_x = false;
    else if ( d.x() <  0.0 ) { w.x() = node->min().x() - point.x(); }
    else                     { w.x() = node->max().x() - point.x(); x_nb = 1; }
    
    if      ( d.y() == 0.0 ) use_y = false;
    else if ( d.y() <  0.0 ) { w.y() = node->min().y() - point.y(); }
    else                     { w.y() = node->max().y() - point.y(); y_nb = 1; }
    
    if      ( d.z() == 0.0 ) use_z = false;
    else if ( d.z() <  0.0 ) { w.z() = node->min().z() - point.z(); }
    else                     { w.z() = node->max().z() - point.z(); z_nb = 1; }

    if ( ! ( use_x || use_y || use_z ) )
    {
        std::cout << "(TOctTreeIntersect) next_node : ray has zero direction"
                  << std::endl;
        return NULL;
    }// if
        
    //
    // determine neighbour
    //

    real  t_x = 0.0, t_y = 0.0, t_z = 0.0;
    uint  ndir = (ulong) -1;

    if ( use_x ) t_x = w.x() / d.x();
    if ( use_y ) t_y = w.y() / d.y();
    if ( use_z ) t_z = w.z() / d.z();

    // look for smallest t
    if ( use_x )
    {
        if ( use_y )
        {
            if ( t_x < t_y )
            {
                if ( use_z && ( t_z < t_x ) ) ndir = (z_nb == 0 ? FRONT_NB : BACK_NB);
                else                          ndir = (x_nb == 0 ? LEFT_NB  : RIGHT_NB);
            }// if
            else
            {
                if ( use_z && ( t_z < t_y ) ) ndir = (z_nb == 0 ? FRONT_NB : BACK_NB);
                else                          ndir = (y_nb == 0 ? DOWN_NB  : UP_NB);
            }// else
        }// if
        else
        {
            if ( use_z && ( t_z < t_x ) ) ndir = (z_nb == 0 ? FRONT_NB : BACK_NB);
            else                          ndir = (x_nb == 0 ? LEFT_NB  : RIGHT_NB);
        }// else
    }// if
    else
    {
        if ( use_y )
        {
            if ( use_z && ( t_z < t_y ) ) ndir = (z_nb == 0 ? FRONT_NB : BACK_NB);
            else                          ndir = (y_nb == 0 ? DOWN_NB  : UP_NB);
        }// if
        else
        {
            if ( use_z ) ndir = (z_nb == 0 ? FRONT_NB : BACK_NB);
            else
                assert( false );          // cannot reach this
        }// else
    }// else

    //
    // ID -> node conversion
    //

    uint    nid       = node->neighbour_id( ndir );
    TNode * neighbour = NULL;

    if ( nid == INVALID_ID )
        return NULL;
    
    if ( nid < _nodes.size() )
        neighbour = _nodes[ nid ];

    // if no neighbour, check against father-IDs
    while ((neighbour == NULL) && (nid != 0))
    {
        nid >>= 3;
        
        if ( nid < _nodes.size() )
            neighbour = _nodes[nid];
    }// while

    // if still no neighbour, return
    if ( neighbour == NULL )
        return NULL;
    
    //
    // compute intersecting point with neighbour
    //

    real  t = 0.0;

    switch ( ndir )
    {
        case LEFT_NB  : t = t_x; break;
        case RIGHT_NB : t = t_x; break;
        case UP_NB    : t = t_y; break;
        case DOWN_NB  : t = t_y; break;
        case FRONT_NB : t = t_z; break;
        case BACK_NB  : t = t_z; break;
        default :
            std::cout << "(TOctTreeIntersect) next_node : no such neighbour"
                      << std::endl;
    }// switch

    point.add( t, d );

    //
    // if neighbour is not a leaf, search for node which
    // contains computed node
    //

    while ( ! neighbour->is_leaf() )
    {
#if 1
        uint  son_no = 0;
        
        for ( uint i = 0; i < 3; i++ )
        {
            if ( point[i] > neighbour->middle()[i] )
                son_no |= (1 << i);
        }// for

        // map bits to real son number
        neighbour = neighbour->son( bit_to_son[son_no] );

        assert( neighbour->is_inside( point ) );
#else        
        bool found = false;
        
        for ( uint i = 0; i < 8; i++ )
        {
            if ( neighbour->son(i)->is_inside( point ) )
            {
                found     = true;
                neighbour = neighbour->son(i);
                break;
            }// if
        }// for

        if ( ! found )
        {
            std::cout << "(TOctTreeIntersect) next_node : no son contains point ("
                      << point << ", "
                      << neighbour->min() << " - "
                      << neighbour->max() << ")" << std::endl
                      << "    ray = " << ray << std::endl;
            return NULL;
        }// if
#endif
    }// if

    //
    // now the neighbour is the node we are looking for
    //
    
    return neighbour;
}

//
// do intersection test with ray and objects in given list
//
TGraphObject *
TOctTreeIntersect::intersect_list ( const TSLL< TGraphObject * > & objects,
                                    const TRay                   & ray,
                                    real                         & hit_distance,
                                    real                           front,
                                    real                           back ) const
{
    // no objects, no intersection
    if ( objects.size() == 0 )
        return NULL;
    
    //
    // go over each object in the list and calculate
    // intersection
    //

    TSLL< TGraphObject * >::TIterator   iter;
    TGraphObject                      * min_object = NULL;
    real                                min_distance = -1e100, distance;
    bool                                intersected;
    
    for ( iter = objects.first(); ! iter.eol(); ++iter )
    {
        if ( iter()->is_last_ray( ray ) )
        {
            //
            // object was already tested against this ray
            // just extract data
            //
        
            intersected = iter()->last_ray_intersected();
            distance    = iter()->last_ray_distance();
        }// if
        else
        {
            //
            // intersect with object and compute data
            //
        
            intersected = iter()->intersect( ray, distance );
            iter()->set_last_ray( ray, distance, intersected );
        }// if

        //
        // check if intersection point is inside viewing frustrum
        //
        
        if ( intersected &&
             inside_view( distance, front, back ) &&
             (( min_object == NULL ) || ( distance < min_distance )))
        {
            min_object   = iter();
            min_distance = distance;
        }// if
    }// while

    //
    // set hit-distance and return hit-object
    //
    
    hit_distance = min_distance;

    return min_object;
}

//////////////////////////////////////////////
//
// output
//

//
// print octtree (inclusiv given objects) into vrml-file
//
void
TOctTreeIntersect::print_vrml ( const TSLL< TGraphObject * > * objects,
                                ulong                          pid ) const
{
    TVRML  vrml;

    vrml.begin( "octtree.wrl" );
    
    //
    // print objects first
    //

    if ( objects != NULL )
    {
        TSLL< TGraphObject * >::TIterator  iter = objects->first();

        vrml.set_colour( TColour( "green" ) );
        
        while ( ! iter.eol() )
        {
            TSphere * sphere = dynamic_cast< TSphere * >( iter() );
            
            if ( sphere != NULL )
                vrml.draw_sphere( sphere->center(), sphere->radius() );
            else
                std::cout << "(TOctTreeIntersect) print_vrml : unsupported objecttype"
                          << std::endl;
            
            iter++;
        }// while
    }// if

    //
    // now print octtree
    //

    TSLL< TNode * >  nodes;
    uint             count = 0;

    vrml.set_colour( TColour( "black" ) );

    nodes.append( _root );
    while ( nodes.size() > 0 )
    {
        TNode * node = nodes.behead();

        count++;
        
        //
        // print node
        //

        if ( node->is_leaf() )
        {
            bool draw = true;
            
            if ( node->id() == pid )
            {
                if ( node->objects().size() > 0 )
                    vrml.set_colour( TColour( "red" ) );
                else
                    vrml.set_colour( TColour( "darkred" ) );
            }// if
            else if ( node->objects().size() > 0 )
                vrml.set_colour( TColour( "black" ) );
            else
                draw = false; //vrml.set_colour( TColour( "gray" ) );

            if ( draw )
            {
                vrml.move_to( node->min() );
                vrml.line_to( TVector( node->max().x(), node->min().y(), node->min().z() ) );
                vrml.line_to( TVector( node->max().x(), node->max().y(), node->min().z() ) );
                vrml.line_to( TVector( node->min().x(), node->max().y(), node->min().z() ) );
                vrml.line_to( TVector( node->min().x(), node->min().y(), node->min().z() ) );
                vrml.line_to( TVector( node->min().x(), node->min().y(), node->max().z() ) );
                vrml.line_to( TVector( node->max().x(), node->min().y(), node->max().z() ) );
                vrml.line_to( TVector( node->max().x(), node->min().y(), node->min().z() ) );
                vrml.move_to( TVector( node->min().x(), node->max().y(), node->min().z() ) );
                vrml.line_to( TVector( node->min().x(), node->max().y(), node->max().z() ) );
                vrml.line_to( TVector( node->max().x(), node->max().y(), node->max().z() ) );
                vrml.line_to( TVector( node->max().x(), node->max().y(), node->min().z() ) );
                vrml.draw_line( TVector( node->min().x(), node->min().y(), node->max().z() ),
                                TVector( node->min().x(), node->max().y(), node->max().z() ) );
                vrml.draw_line( TVector( node->max().x(), node->min().y(), node->max().z() ),
                                TVector( node->max().x(), node->max().y(), node->max().z() ) );
            }// if
        }// if
        else
        {
            //
            // go on with all sons of this node
            //
            
            for ( uint i = 0; i < 8; i++ )
            {
                if ( node->son(i) != NULL )
                    nodes.append( node->son(i) );
            }// for
        }// else
    }// while

    vrml.set_colour( TColour( "red" ) );
    vrml.draw_line(  TVector( 0, 0, 0 ), TVector( 1, 0, 0 ) );
    vrml.set_colour( TColour( "green" ) );
    vrml.draw_line(  TVector( 0, 0, 0 ), TVector( 0, 1, 0 ) );
    vrml.set_colour( TColour( "blue" ) );
    vrml.draw_line(  TVector( 0, 0, 0 ), TVector( 0, 0, 1 ) );

    //
    // finish
    //

    vrml.end();

    if ( verbose(5) )
        std::cout << "(TOctTreeIntersect) print_vrml : printed "
                  << count << " nodes" << std::endl;
}

//
// stream output
//
void
TOctTreeIntersect::print ( std::ostream & os ) const
{
    if ( _root != NULL )
        _root->print( os, 0 );
    else
        os << "(empty)" << std::endl;
}
    

////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////
//
// TNode (implementation)
//
////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////

//////////////////////////////////////
//
// octtree-functionality
//

//
// lookup-tables for neighbour-selection
//
static const uint go_neighbour[6][8] =
{
    { 1, 0, 3, 2, 5, 4, 7, 6 },  // to left
    { 1, 0, 3, 2, 5, 4, 7, 6 },  // to right
    { 4, 5, 6, 7, 0, 1, 2, 3 },  // to down
    { 4, 5, 6, 7, 0, 1, 2, 3 },  // to up
    { 2, 3, 0, 1, 6, 7, 4, 5 },  // to front
    { 2, 3, 0, 1, 6, 7, 4, 5 }   // to back
};

static const bool must_decrement[6][8] =
{
    { 1, 0, 1, 0, 1, 0, 1, 0 },  // to left
    { 0, 0, 0, 0, 0, 0, 0, 0 },  // to right
    { 1, 1, 1, 1, 0, 0, 0, 0 },  // to down
    { 0, 0, 0, 0, 0, 0, 0, 0 },  // to up 
    { 0, 0, 0, 0, 0, 0, 0, 0 },  // to front
    { 1, 1, 0, 0, 1, 1, 0, 0 }   // to back
};

static const bool must_increment[6][8] =
{
    { 0, 0, 0, 0, 0, 0, 0, 0 },  // to left
    { 0, 1, 0, 1, 0, 1, 0, 1 },  // to right
    { 0, 0, 0, 0, 0, 0, 0, 0 },  // to down
    { 0, 0, 0, 0, 1, 1, 1, 1 },  // to up 
    { 0, 0, 1, 1, 0, 0, 1, 1 },  // to front
    { 0, 0, 0, 0, 0, 0, 0, 0 }   // to back
};

//
// return neighbour-ID
// - n selects neighbour : left=0, right, up, down, front, back=5
//
ulong
TOctTreeIntersect::TNode::neighbour_id ( uint n ) const
{
    assert( n < 6 );

    // if root => no neighbour
    if ( _id == 1 )
        return INVALID_ID;

    // extract last number from id
    uint  shift = 3;
    uint  mask  = 0x7;
    uint  sid   = _id & mask;
    uint  nid   = (_id & ~7) | go_neighbour[n][sid];
    bool  inc   = must_increment[n][sid];
    bool  dec   = must_decrement[n][sid];
    
    while ((dec || inc) && ((_id >> shift) != 1) )
    {
        sid    = (_id >> shift) & mask;
        nid   &= ~(mask << shift);
        nid   |= go_neighbour[n][sid] << shift;
        inc    = must_increment[n][sid];
        dec    = must_decrement[n][sid];
        shift += 3;
    }// while

    // check if out of octtree
    if ((dec || inc) && ((_id >> shift) == 1))
        return INVALID_ID;
    
    return nid;
}

//
// build all sons
//
void
TOctTreeIntersect::TNode::build_sons ()
{
    for ( uint i = 0; i < 8; i++ )
    {
        if ( _sons[i] != NULL )
            delete _sons[i];
    }// for

    TVector  middle;

    middle.set( 1.0, _max, 1.0, _min ).scale( 0.5 );

    //
    // lower part
    //
    
    _sons[0] = new TNode( TVector( _min.x(), _min.y(), middle.z() ),
                          TVector( middle.x(), middle.y(), _max.z() ),
                          (_id << 3) | 0 );
    _sons[1] = new TNode( TVector( middle.x(), _min.y(), middle.z() ),
                          TVector( _max.x(), middle.y(), _max.z() ),
                          (_id << 3) | 1 );
    _sons[2] = new TNode( TVector( _min.x(), _min.y(), _min.z() ),
                          TVector( middle.x(), middle.y(), middle.z() ),
                          (_id << 3) | 2 );
    _sons[3] = new TNode( TVector( middle.x(), _min.y(), _min.z() ),
                          TVector( _max.x(), middle.y(), middle.z() ),
                          (_id << 3) | 3 );
    
    //
    // upper part
    //
    
    _sons[4] = new TNode( TVector( _min.x(), middle.y(), middle.z() ),
                          TVector( middle.x(), _max.y(), _max.z() ),
                          (_id << 3) | 4 );
    _sons[5] = new TNode( TVector( middle.x(), middle.y(), middle.z() ),
                          TVector( _max.x(), _max.y(), _max.z() ),
                          (_id << 3) | 5 );
    _sons[6] = new TNode( TVector( _min.x(), middle.y(), _min.z() ),
                          TVector( middle.x(), _max.y(), middle.z() ),
                          (_id << 3) | 6 );
    _sons[7] = new TNode( TVector( middle.x(), middle.y(), _min.z() ),
                          TVector( _max.x(), _max.y(), middle.z() ),
                          (_id << 3) | 7 );
}

//
// check if given object is inside boundingbox
//
bool
TOctTreeIntersect::TNode::is_inside ( const TGraphObject * obj ) const
{
    TVector  omin, omax;

    if ( obj->bounding_box( omin, omax ) )
    {
        //
        // test against node-bounding-box
        //

        if ( ( omin.x() > _max.x() + EPSILON ) ||
             ( omax.x() < _min.x() - EPSILON ) ||
             ( omin.y() > _max.y() + EPSILON ) ||
             ( omax.y() < _min.y() - EPSILON ) ||
             ( omin.z() > _max.z() + EPSILON ) ||
             ( omax.z() < _min.z() - EPSILON ) )
            return false;
        else
            return true;
    }// if
    else
        return false;
}

//
// check if given point is inside boundingbox
//
bool
TOctTreeIntersect::TNode::is_inside ( const TVector & v ) const
{
    //
    // test against node-bounding-box
    //

    if ( ( v.x() > _max.x() + EPSILON ) ||
         ( v.x() < _min.x() - EPSILON ) ||
         ( v.y() > _max.y() + EPSILON ) ||
         ( v.y() < _min.y() - EPSILON ) ||
         ( v.z() > _max.z() + EPSILON ) ||
         ( v.z() < _min.z() - EPSILON ) )
        return false;
    else
        return true;
}

//
// return if leaf (no sons)
//
bool
TOctTreeIntersect::TNode::is_leaf () const
{
    for ( uint i = 0; i < 8; i++ )
    {
        if ( son(i) != NULL )
            return false;
    }// for

    return true;
}

//
// compute point of intersection with bounding-box
//
#define INSIDE_PLANE( d1, d2, v ) ( ((v).d1() >= _min.d1() - EPSILON) && \
                                    ((v).d1() <= _max.d1() + EPSILON) && \
                                    ((v).d2() >= _min.d2() - EPSILON) && \
                                    ((v).d2() <= _max.d2() + EPSILON) )
                         
bool
TOctTreeIntersect::TNode::bb_intersect ( const TRay & ray,
                                         TVector    & hit_point ) const
{
#if 1
    //
    // use Slabs method to check intersection by computing the
    // minimal/maximal intersection distances for all planes
    // and comparing the minimal/maximal value
    // (see "Real-time Rendering" by Akenine-Mller and Haines, 13.6.1)
    //

    TVector  p;
    real     tmin = -1e100;
    real     tmax =  1e100;

    p.set( 1.0, middle(), -1.0, ray.origin() );

    for ( uint i = 0; i < 3; i++ )
    {
        real  e = p[i];
        real  f = ray.direction()[i];

        if ( std::abs( f ) > EPSILON )
        {
            real  t1, t2;

            f = 1.0 / f;
            
            t1 = ( e + diam()[i] ) * f;
            t2 = ( e - diam()[i] ) * f;

            if ( t1 > t2 ) swap( t1, t2 );

            tmin = ::max( tmin, t1 );
            tmax = ::min( tmax, t2 );

            if (( tmin > tmax ) || ( tmax < 0.0 ))
                return false;
        }// if
        else if (( -e-diam()[i] > 0.0 ) || ( -e+diam()[i] < 0.0 ))
            return false;
    }// for

    if ( tmin > 0.0 )
        hit_point.set( 1.0, ray.origin(), tmin, ray.direction() );
    else
        hit_point.set( 1.0, ray.origin(), tmax, ray.direction() );

    return true;
#else
    //
    // test against all planes of bounding box
    // and choose the nearest
    //

    bool     intersected = false;
    real     min_t = 1e100;
    TVector  hp;
    
    // XY planes
    if ( ray.direction().z() != 0.0 )
    {
        // front/back plane
        real t[2] = { (_min.z() - ray.origin().z()) / ray.direction().z(),
                      (_max.z() - ray.origin().z()) / ray.direction().z() };

        for ( uint i = 0; i < 2; i++ )
        {
            if ((t[i] > 0) && (t[i] < min_t))
            {
                hp.set( 1.0, ray.origin(), t[i], ray.direction() );
                
                if ( INSIDE_PLANE( x, y, hp ) )
                {
                    min_t       = t[i];
                    hit_point   = hp;
                    intersected = true;
                }// if
            }// if
        }// for
    }// if

    // XZ planes
    if ( ray.direction().y() != 0.0 )
    {
        // down/up plane
        real t[2] = { (_min.y() - ray.origin().y()) / ray.direction().y(),
                      (_max.y() - ray.origin().y()) / ray.direction().y() };

        for ( uint i = 0; i < 2; i++ )
        {
            if ((t[i] > 0) && (t[i] < min_t))
            {
                hp.set( 1.0, ray.origin(), t[i], ray.direction() );
                
                if ( INSIDE_PLANE( x, z, hp ) )
                {
                    min_t       = t[i];
                    hit_point   = hp;
                    intersected = true;
                }// if
            }// if
        }// for
    }// if

    // YZ planes
    if ( ray.direction().x() != 0.0 )
    {
        // front/back plane
        real t[2] = { (_min.x() - ray.origin().x()) / ray.direction().x(),
                      (_max.x() - ray.origin().x()) / ray.direction().x() };

        for ( uint i = 0; i < 2; i++ )
        {
            if ((t[i] > 0) && (t[i] < min_t))
            {
                hp.set( 1.0, ray.origin(), t[i], ray.direction() );
                
                if ( INSIDE_PLANE( y, z, hp ) )
                {
                    min_t       = t[i];
                    hit_point   = hp;
                    intersected = true;
                }// if
            }// if
        }// for
    }// if

    return intersected;
#endif
}

//////////////////////////////////////////////
//
// output
//

//
// stream output
//
void
TOctTreeIntersect::TNode::print ( std::ostream & os, uint ofs ) const
{
    // indent
    for ( uint i = 0; i < ofs; i++ )
        os << " ";

    // write node
    os << _min << " -> " << _max << std::endl;

    for ( uint i = 0; i < 8; i++ )
    {
        if ( son(i) != NULL )
        {
            for ( uint j = 0; j < ofs+2; j++ )
                os << " ";

            os << "son " << i << std::endl;
            son(i)->print( os, ofs+4 );
        }// if
    }// for
}

}// namespace
