//
// Project : lib3D
// File    : TSolidNoise.cc
// Purpose : class for computing solid noise
// Author  : Ronald Kriemann
//
// arch-tag: 4856cd5b-5370-4c1f-bfcf-439826927db1
//

#include "TRNG.hh"

#include "TSolidNoise.hh"

namespace RGL
{

//////////////////////////////////////////////
//
// constructor and destructor
//

TSolidNoise::TSolidNoise ( uint seed )
{
    //
    // setup vectors
    //

    _grid[0].set(  1,  1,  0 );
    _grid[1].set( -1,  1,  0 );
    _grid[2].set(  1, -1,  0 );
    _grid[3].set( -1, -1,  0 );
    
    _grid[4].set(  1,  0,  1 );
    _grid[5].set( -1,  0,  1 );
    _grid[6].set(  1,  0, -1 );
    _grid[7].set( -1,  0, -1 );
    
    _grid[8].set(  0,  1,  1 );
    _grid[9].set(  0, -1,  1 );
    _grid[10].set(  0,  1, -1 );
    _grid[11].set(  0, -1, -1 );
    
    _grid[12].set(  1,  1,  0 );
    _grid[13].set( -1,  1,  0 );
    _grid[14].set(  0, -1,  1 );
    _grid[15].set(  0, -1, -1 );

    //
    // create random lookup
    //
 
    TRNG  rng( seed );
    
    for ( uint i = 0; i < 16; i++ )
        _phi[i] = i;
    
    for ( uint i = 0; i < 16; i++ )
    {
        uint  pos = uint( rng( 16.0 ) );
        int   tmp = _phi[pos];
        
        _phi[pos] = _phi[i];
        _phi[i]   = tmp; 
    }// for
}
   
//////////////////////////////////////////////
//
// supply noise and turbulence
//

#define OMEGA( i, j, k, v ) \
            ( omega( v.x() ) * omega( v.y() ) * omega( v.z() ) * gamma( i, j, k, v ) )

real
TSolidNoise::noise ( const TVector & v ) const
{
    long     i  = long(std::floor( v.x() ));
    long     j  = long(std::floor( v.y() ));
    long     k  = long(std::floor( v.z() ));
    real     xi = v.x() - real(i);
    real     yj = v.y() - real(j);
    real     zk = v.z() - real(k);
    real     sum = 0.0;
    TVector  w;
    
    sum += OMEGA(   i,   j,   k, w.set(     xi,     yj,     zk ) );
    sum += OMEGA(   i,   j, k+1, w.set(     xi,     yj, zk-1.0 ) );
    sum += OMEGA(   i, j+1,   k, w.set(     xi, yj-1.0,     zk ) );
    sum += OMEGA(   i, j+1, k+1, w.set(     xi, yj-1.0, zk-1.0 ) );
    sum += OMEGA( i+1,   j,   k, w.set( xi-1.0,     yj,     zk ) );
    sum += OMEGA( i+1,   j, k+1, w.set( xi-1.0,     yj, zk-1.0 ) );
    sum += OMEGA( i+1, j+1,   k, w.set( xi-1.0, yj-1.0,     zk ) );
    sum += OMEGA( i+1, j+1, k+1, w.set( xi-1.0, yj-1.0, zk-1.0 ) );

    return sum;
}

real
TSolidNoise::turbulence ( const TVector & v, uint depth ) const
{
    TVector  w;
    real     weight = 1.0;
    real     turb;

    turb = std::abs( noise(v) );

    for ( uint i = 1; i < depth; i++ )
    {
        weight *= 2.0;
        w.set( weight, v );

        turb += std::abs( noise(w) ) / weight;
    }// for

    return turb;
}

}// namespace
