//
// Project : lib3d
// File    : TColour.cc
// Purpose : Definition of a class for colour-management
// Author  : Ronald Kriemann
//
// arch-tag: bfc72521-85f9-4364-bc47-97523ea24f7f
//

#include <cmath>

#include "TColour.hh"

namespace RGL
{

//
// get/set colour in HSV-space ( h = -1 or 0 <= h <= 360, 0 <= s,v <= 1 )
//
void
TColour::hsv ( real & h, real & s, real & v ) const
{
    real  max = _red;
    real  min = _red;

    // locate maximal / minimal value of rgb-components
    if ( _green > max )  max = _green;
    if ( _blue  > max )  max = _blue;
    if ( _green < min )  min = _green;
    if ( _blue  < min )  min = _blue;

    // max value is v
    v = max;

    // calc saturation
    if ( max != 0.0)
        s = (max -min) / max;
    else
        s = 0.0;

    if ( s == 0.0 )
        h = -1.0;             // undefined colour
    else
    {
        real  delta = max - min;

        if ( _red == max )
            h = (_green - _blue) / delta;
        else if ( _green == max )
            h = 2.0 + ((_blue - _red) / delta);
        else if (_blue == max)
            h = 4.0 + ((_red - _green) / delta);

        h *= 60.0;

        if ( h < 0.0 )
            h += 360.0;
    }// else
}

TColour & 
TColour::set_hsv ( real h, real s, real v )
{
    if ( s == 0 )
    {
        //
        // achromatic case (gray)
        //
        
        if ( h == -1 )
            _red = _green = _blue = v;
        else
            std::cout << "(TColour) setHSV : undefined colour" << std::endl;
    }// if
    else
    {
        //
        // chromatic case
        //
        
        real  f, p, q, t;
        int   i;
        
        // check interval
        if (( h < 0.0 ) || ( h > 360.0 ))
        {
            while ( h > 360.0 ) h -= 360.0;
            while ( h <   0.0 ) h += 360.0;
        }// if

        // sector 0 to 5
        h /= 60.0;
        i  = int( std::floor( h ) );

        // factorial part of h
        f  = h - real( i );
        
        p  = v * (1.0 - s);
        q  = v * (1.0 - (s * f));
        t  = v * (1.0 - (s * (1.0 - f)));

        switch( i )
        {
            case 0 : _red = v; _green = t; _blue = p; break;
            case 1 : _red = q; _green = v; _blue = p; break;
            case 2 : _red = p; _green = v; _blue = t; break;
            case 3 : _red = p; _green = q; _blue = v; break;
            case 4 : _red = t; _green = p; _blue = v; break;
            case 5 : _red = v; _green = p; _blue = q; break;
        }// switch
    }// else

    return *this;
}

//
// set colour as named-value
//
TColour &
TColour::set_named ( const TString & name )
{
    return set_rgb32( name_to_rgb( name ) );
}

//
// set colour from 32 bit integer (from lsb to msb: red,green,blue,alpha)
//

#define RGB_RED( v )    (( (v) >> 16 ) & 0xFF)
#define RGB_GREEN( v )  (( (v) >>  8 ) & 0xFF)
#define RGB_BLUE( v )   (( (v)       ) & 0xFF)

TColour &
TColour::set_rgb32 ( uint rgb )
{
    return set_rgb( real( RGB_RED(   rgb ) ) / 255.0,
                    real( RGB_GREEN( rgb ) ) / 255.0,
                    real( RGB_BLUE(  rgb ) ) / 255.0,
                    false );
}

//
// makes colour lighter or darker
//

// void
// TColour::light ( int f )
// {
//     if ( f <= 0 )                          // invalid lightness factor
//         return;
//     else if ( f < 100 )                    // makes colour darker
//     {
//         dark( 10000 / f );
//         return;
//     }// else
    
//     real  h, s, v;
    
//     hsv( h, s, v );
//     v = (f * v) / 100;
    
//     if ( v > 1.0 )                         // overflow
//     {
//         s -= (f * 106) / 100;              // adjust saturation
//         if ( s < 0 )
//             s = 0;
//         v = 1.0;
//     }// if

//     set_hsv( h, s, v );
// }

// void
// TColour::dark ( int f )
// {
//     if ( f <= 0 )                          // invalid darkness factor
//         return;
//     else if ( f < 100 )                    // makes colour lighter
//     {
//         light( 10000 / f );
//         return;
//     }// else
    
//     real  h, s, v;

//     hsv( h, s, v );
//     v = (v * 100) / f;

//     set_hsv( h, s, v );
// }

//
// stream output
//
std::ostream &
operator << ( std::ostream & os, const TColour & c )
{
    os << "( " << c.red() << ", " << c.green() << ", " << c.blue() << " )";
    return os;
}

} // namespace

