//
// Project : lib3D
// File    : TFile.cc
// Purpose : classes for encapsulating file access
// Author  : Ronald Kriemann
//
// arch-tag: 5ee837ed-6ca2-4314-9fd0-e946711009c2
//

#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <errno.h>

#ifdef USE_RL
#include <readline/readline.h>
#include <readline/history.h>
#endif

#include <iostream>

#include "TFile.hh"

namespace RGL
{

/////////////////////////////////////////////
/////////////////////////////////////////////
//
// TReadline
//
/////////////////////////////////////////////
/////////////////////////////////////////////

//
// constructor and destructor
//

TReadline::TReadline ()
{
    _prompt   = "> ";
    _eof      = false;
    _hist_len = 100;
}

TReadline::TReadline ( const TString & prompt )
{
    _prompt   = prompt;
    _eof      = false;
    _hist_len = 100;
}

TReadline::~TReadline ()
{
}

//
// IO
//

//
// write string
//
int
TReadline::puts ( const TString & buf )
{
    // simply write to stdio
    fprintf( stdout, buf.c_str() );
    return 0;
}

//
// read string
//
int
TReadline::gets ( TString & buf )
{
#ifdef USE_RL
    using_history();

    char  * str = readline( _prompt.c_str() );
    
    if ( str != NULL )
    {
        add_history( str );
        buf = str;
        
        free( str );   // important to call "free" since
                       // it was allocated by malloc
        
        stifle_history( _hist_len );
    }// if
    else
        _eof = true;
#else
    std::cout << "(TReadline) gets : no readline supported" << std::endl;
#endif

    return 0;
}

/////////////////////////////////////////////
/////////////////////////////////////////////
//
// TFile
//
/////////////////////////////////////////////
/////////////////////////////////////////////

//
// constructor and destructor
//

TFile::TFile ()
        : _is_open(false)
{}

TFile::TFile ( const TString & fname, const TString & mode )
        : _is_open(false)
{
    open( fname, mode );
}

TFile::~TFile ()
{
    if ( is_open() )
        close();
}
    
//
// file management
//

int
TFile::open ( const TString & fname, const TString & mode )
{
    if ( is_open() )
        close();

    int  ret = _open( fname, mode );
    
    if ( ret == 0 )
        _is_open = true;

    return ret;
}

int
TFile::close ()
{
    if ( is_open() )
    {
        _is_open = false;

        return _close();
    }// if

    return 0;
}

/////////////////////////////////////////////
/////////////////////////////////////////////
//
// TStdFile
//
/////////////////////////////////////////////
/////////////////////////////////////////////

//
// constructor and destructor
//

TStdFile::TStdFile ()
        : _file(NULL)
{}

TStdFile::TStdFile ( const TString & fname, const TString & mode )
        : TFile( fname, mode )
{
}

TStdFile::~TStdFile ()
{
    if ( is_open() )
        close();
}

//
// file management
//

int
TStdFile::_open ( const TString & fname, const TString & mode )
{
    if ( is_open() )
        close();

    _file = fopen( fname.c_str(), mode.c_str() );

    if ( _file == NULL )
    {
        std::cout << "(TStdFile) _open : could not open file \"" << fname
                  << "\" (" << strerror( errno ) << ")" << std::endl;
        return -1;
    }// if

    return 0;
}

int
TStdFile::_close ()
{
    if ( is_open() )
    {
        if ( fclose( _file ) != 0 )
        {
            std::cout << "(TStdFile) _close : error while closing file"
                      << " (" << strerror( errno ) << ")" << std::endl;
            return -1;
        }// if
    }// if

    return 0;
}

bool
TStdFile::eof () const
{
    if ( ! is_open() )
    {
        std::cout << "(TStdFile) eof : file not open" << std::endl;
        return true;
    }// if

    return feof( _file );
}
    
//
// write string 
//
int
TStdFile::puts ( const TString & buf )
{
    if ( ! is_open() )
    {
        std::cout << "(TStdFile) puts : file not open" << std::endl;
        return -1;
    }// if
    
    if ( fputs( buf.c_str(), _file ) < 0 )
    {
        std::cout << "(TStdFile) puts : error while writing to file "
                  << " (" << strerror( errno ) << ")" << std::endl;
        return -1;
    }// if

    return 0;
}

//
// read string
//
int
TStdFile::gets ( TString & buf )
{
    if ( ! is_open() )
    {
        std::cout << "(TStdFile) gets : file not open" << std::endl;
        return -1;
    }// if

    int  len = buf.size();
    
    if ( buf.size() < 256 )
    {
        buf.set_size( 256 );
        len = 256;
    }// if
        
    if ( fgets( buf.c_str(), len, _file ) == NULL )
    {
        if ( ! feof( _file ) )
        {
            std::cout << "(TStdFile) gets : error while reading from file "
                      << " (" << strerror( errno ) << ")" << std::endl;
            return -1;
        }// if
        else
            return 0;
    }// if

    return 0;
}

/////////////////////////////////////////////
/////////////////////////////////////////////
//
// TGZFile
//
/////////////////////////////////////////////
/////////////////////////////////////////////

#ifdef USE_ZLIB

//
// constructor and destructor
//

TGZFile::TGZFile ()
{}

TGZFile::TGZFile ( const TString & fname, const TString & mode )
        : TFile( fname, mode )
{
}

TGZFile::~TGZFile ()
{
    if ( is_open() )
        close();
}

//
// file management
//

int
TGZFile::_open ( const TString & fname, const TString & mode )
{
    if ( is_open() )
        close();

    _file = gzopen( fname.c_str(), mode.c_str() );

    if ( _file == NULL )
    {
        std::cout << "(TGZFile) _open : could not open file \"" << fname
                  << "\" (" << strerror( errno ) << ")" << std::endl;
        return -1;
    }// if

    return 0;
}

int
TGZFile::_close ()
{
    int ret = 0;
    
    if ( is_open() )
    {
        if ( gzclose( _file ) != 0 )
        {
            std::cout << "(TGZFile) _close : error while closing file"
                      << " (" << strerror( errno ) << ")" << std::endl;
            ret = -1;
        }// if
    }// if

    return ret;
}

bool
TGZFile::eof () const
{
    if ( ! is_open() )
    {
        std::cout << "(TGZFile) eof : file not open" << std::endl;
        return true;
    }// if

    return gzeof( _file );
}
    
//
// write string 
//
int
TGZFile::puts ( const TString & buf )
{
    if ( ! is_open() )
    {
        std::cout << "(TGZFile) puts : file not open" << std::endl;
        return -1;
    }// if
    
    if ( gzputs( _file, buf.c_str() ) < 0 )
    {
        std::cout << "(TGZFile) puts : error while writing to file "
                  << " (" << strerror( errno ) << ")" << std::endl;
        return -1;
    }// if

    return 0;
}

//
// read string
//
int
TGZFile::gets ( TString & buf )
{
    if ( ! is_open() )
    {
        std::cout << "(TGZFile) gets : file not open" << std::endl;
        return -1;
    }// if
    
    int  len = buf.size();
    
    if ( buf.size() < 256 )
    {
        buf.set_size( 256 );
        len = 256;
    }// if
        
    if ( gzgets( _file, buf.c_str(), len ) == NULL )
    {
        if ( ! gzeof( _file ) )
        {
            std::cout << "(TGZFile) gets : error while reading from file "
                      << " (" << strerror( errno ) << ")" << std::endl;
            return -1;
        }// if
        else
            return 0;
    }// if

    return 0;
}

#endif

/////////////////////////////////////////////
/////////////////////////////////////////////
//
// TBZFile
//
/////////////////////////////////////////////
/////////////////////////////////////////////

#ifdef USE_BZ2

//
// constructor and destructor
//

TBZFile::TBZFile ()
        : _eof(false), _pos(0), _len(0)
{}

TBZFile::TBZFile ( const TString & fname, const TString & mode )
        : TFile( fname, mode ), _eof(false), _pos(0), _len(0)
{
}

TBZFile::~TBZFile ()
{
    if ( is_open() )
        close();
}

//
// file management
//

int
TBZFile::_open ( const TString & fname, const TString & mode )
{
    if ( is_open() )
        close();

    // analyse mode and reject invalid ones
    uint  ml   = mode.length();
    
    if ( ml > 0 )
    {
        if      ( mode[0] == 'r' ) _reading = true;
        else if ( mode[0] == 'w' ) _reading = false;
        else if ( mode[0] == 'a' ) _reading = false;

        for ( uint i = 0; i < ml; i++ )
        {
            if ( mode[i] == '+' )
            {
                std::cout << "(TBZFile) open : invalid mode (" << mode << "), "
                          << "allowed is 'r', 'w' or 'a'" << std::endl;
                return -1;
            }// if
        }// for
    }// for
         
    _file = fopen( fname.c_str(), mode.c_str() );
    
    if ( _file == NULL )
    {
        std::cout << "(TBZFile) _open : could not open file \"" << fname
                  << "\" (" << strerror( errno ) << ")" << std::endl;
        return -1;
    }// if

    if ( _reading )
    {
        _nunused = 0;
        _bzf = BZ2_bzReadOpen ( & _err, _file, 0, 0, _unused, 0 );
    }// if
    else
        _bzf = BZ2_bzWriteOpen ( & _err, _file, 9, 0, 30 );

    if (( _bzf == NULL ) || ( _err != BZ_OK ))
    {
        std::cout << "(TBZFile) _open : error while opening bzip2-stream" << std::endl;
        fclose( _file );
        return -1;
    }// if
    
    _eof = false;
    
    return 0;
}

int
TBZFile::_close ()
{
    if ( is_open() )
    {
        uint  n1, n2, n3, n4;
        
        _eof = false;

        if ( _reading )
            BZ2_bzReadClose( & _err, _bzf );
        else
            BZ2_bzWriteClose64( & _err, _bzf, 0, & n1, & n2, & n3, & n4 );

        if ( _err != BZ_OK )
        {
            std::cout << "(TBZFile) _close : error while closing bzip2-stream" << std::endl;
            fclose( _file );
            return -1;
        }// if

        if ( fclose( _file ) != 0 )
        {
            std::cout << "(TBZFile) _close : error while closing file"
                      << " (" << strerror( errno ) << ")" << std::endl;
            return -1;
        }// if
    }// if

    return 0;
}

bool
TBZFile::eof () const
{
    if ( is_open() ) return _eof;
    else             return true;
}
    
//
// write string 
//
int
TBZFile::puts ( const TString & buf )
{
    if ( ! is_open() )
    {
        std::cout << "(TBZFile) puts : file not open" << std::endl;
        return -1;
    }// if

    if ( _reading )
    {
        std::cout << "(TBZFile) puts : file opened for reading" << std::endl;
        return -1;
    }// if
    
    BZ2_bzWrite( & _err, _bzf, (void*) buf.c_str(), buf.length() );

    if ( _err != BZ_OK )
    {
        std::cout << "(TBZFile) puts : error while writing to file "
                  << " (" << strerror( errno ) << ")" << std::endl;
        return -1;
    }// if

    return 0;
}

//
// read string
//
int
TBZFile::gets ( TString & buf )
{
    if ( ! is_open() )
    {
        std::cout << "(TBZFile) gets : file not open" << std::endl;
        return -1;
    }// if

    if ( ! _reading )
    {
        std::cout << "(TBZFile) gets : file opened for writing" << std::endl;
        return -1;
    }// if
    
    //
    // if there is still something in the buffer, copy
    // everything until \0 or len is reached
    //

    if ( _len > 0 )
    {
        while (( _buffer[_pos] != '\n' ) && (_pos < _len))
            buf += _buffer[_pos++];

        if ( _pos < _len )
        {
            // data still left
            buf += _buffer[_pos++];
            buf += '\0';
            return 0;
        }// if

        // no data left
    }// if

    //
    // read until eof, len or \0 reached
    //

    while ( true )
    {
        _pos = 0;
        _len = 0;

        _len = BZ2_bzRead( & _err, _bzf, _buffer, sizeof(_buffer) );

        if ( _err == BZ_STREAM_END )
            _eof = true;
        else if ( _err != BZ_OK )
        {
            std::cout << "(TBZFile) gets : error while reading from file "
                      << " (" << strerror( errno ) << ")" << std::endl;
            return -1;
        }// if
        
        while (( _buffer[_pos] != '\n' ) && (_pos < _len))
            buf += _buffer[_pos++];
        
        if ( _pos < _len )
        {
            // still data left
            buf += _buffer[_pos++];
            buf += '\0';
            return 0;
        }// if
    }// while
    
    return 0;
}

#endif

/////////////////////////////////////////////
/////////////////////////////////////////////
//
// TFileFactory
//
/////////////////////////////////////////////
/////////////////////////////////////////////

//
// analyses filename and return appropriate file-handler
//

TFile *
TFileFactory::open ( const TString & fname, const TString & mode ) const
{
    //
    // determine file suffix
    //

    TString  suffix( 256 );
    int      len   = fname.length();
    int      pos   = len-1;
    bool     found = false;
    
    while ( pos >= 0 )
    {
        if ( fname[pos] == '.' )
        {
            suffix = fname.index( pos + 1 );
            found = true;
            break;
        }// if

        --pos;
    }// while

    suffix.to_lower();
    
    TFile  * file = NULL;
    
    //
    // no suffix found: return standard file handler
    //
    
    if ( ! found )
        file = new TStdFile;

    if ( file == NULL )
    {
        //
        // compare suffix with known extensions
        //

#ifdef USE_ZLIB
        if (( suffix == "gz" ) || ( suffix == "z" ))
            file = new TGZFile;
#endif

#ifdef USE_BZ2
        if (( suffix == "bz"  ) || ( suffix == "bz2" ))
            file = new TBZFile;
#endif
    }// if

    // no handler found: use standard handler
    if ( file == NULL )
        file = new TStdFile;

    //
    // check if opening file is error-free
    //
    
    if ( file->open( fname, mode ) == 0 )
        return file;
    
    // error: return NULL
    delete file;
    
    return NULL;
}

//
// default file factory
//

TFileFactory  file_fac;

}// namespace
