//
// Project : misc
// File    : TNiederreiter.cc
// Purpose : class for a Niederreiter sequence generator
// Author  : Ronald Kriemann
//
// arch-tag: 2548c682-cbb9-4570-adcf-07e9b6a28f5f
//

#include "TNiederreiter.hh"

////
////
//// Implementation stems from the GNU Scientific Library
//// (just restricted to 6 dimensions)
////
////

namespace RGL
{

///////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////
//
// local defines
//
///////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////

// Z_2 field operations
#define NIED2_ADD(x,y) (((x) + (y)) % 2)
#define NIED2_MUL(x,y) (((x) * (y)) % 2)
#define NIED2_SUB(x,y) NIED2_ADD((x),(y))

#define MAXV (NIED2_NBITS + NIED2_MAX_DEGREE)

///////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////
//
// local variables
//
///////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////

// primitive polynomials in binary encoding */
static const int primitive_poly[ NIED2_MAX_DIMENSION+1 ][ NIED2_MAX_PRIM_DEGREE+1 ] =
{
    { 1, 0, 0, 0, 0, 0 },  //  1            
    { 0, 1, 0, 0, 0, 0 },  //  x            
    { 1, 1, 0, 0, 0, 0 },  //  1 + x        
    { 1, 1, 1, 0, 0, 0 },  //  1 + x + x^2  
    { 1, 1, 0, 1, 0, 0 },  //  1 + x + x^3  
    { 1, 0, 1, 1, 0, 0 },  //  1 + x^2 + x^3
    { 1, 1, 0, 0, 1, 0 }   //  1 + x + x^4  
};

/* degrees of primitive polynomials */
static const int poly_degree[ NIED2_MAX_DIMENSION + 1 ] =
{
    0, 1, 1, 2, 3, 3, 4
};

///////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////
//
// local functions
//
///////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////

static void calculate_cj ( int  cj[ NIED2_NBITS ][ NIED2_MAX_DIMENSION ],
                           uint dim );

///////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////
//
// TNiederreiter (implementation)
//
///////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////

////////////////////////////////////////////
//
// access Niederreiter sequence
//

//
// return next number in sequence
//
void
TNiederreiter::next ( TArray< double > & val, double max )
{
    const double recip = 1.0 / (double) (1U << NIED2_NBITS); // 2^(-nbits)

    val.set_size( _dim );
    
    // Load the result from the saved state
    for ( uint i = 0; i < _dim; i++ )
        val[ i ] = max * _nextq[ i ] * recip;

    //
    // Find the position of the least-significant zero in sequence_count.
    // This is the bit that changes in the Gray-code representation as
    // the count is advanced.
    //
    
    int r = 0;
    int c = _sequence_count;
    
    while ( true )
    {
        if (( c % 2 ) == 1 )
        {
            ++r;
            c /= 2;
        }// if
        else
            break;
    }// while

    if ( r >= NIED2_NBITS )
    {
        std::cout << "(TNiederreiter) next : failed" << std::endl;
        return;
    }// if

    // calculate the next state
    for ( uint i = 0; i < _dim; i++ )
        _nextq[ i ] ^= _cj[ r ][ i ];

    _sequence_count++;
}

//
// return next number in sequence (special version for dim = 3)
//
void
TNiederreiter::next ( TVector & v, double max )
{
    if ( _dim != 3 )
    {
        std::cout << "(TNiederreiter) next : dimension of Niederreiter seq. not 3" << std::endl;
        return;
    }// if
    
    const double recip = 1.0 / (double) (1U << NIED2_NBITS); // 2^(-nbits)

    // Load the result from the saved state
    for ( uint i = 0; i < 3; i++ )
        v[ i ] = max * _nextq[ i ] * recip;

    //
    // Find the position of the least-significant zero in sequence_count.
    // This is the bit that changes in the Gray-code representation as
    // the count is advanced.
    //
    
    int r = 0;
    int c = _sequence_count;
    
    while ( true )
    {
        if (( c % 2 ) == 1 )
        {
            ++r;
            c /= 2;
        }// if
        else
            break;
    }// while

    if ( r >= NIED2_NBITS )
    {
        std::cout << "(TNiederreiter) next : failed" << std::endl;
        return;
    }// if

    // calculate the next state
    for ( uint i = 0; i < 3; i++ )
        _nextq[ i ] ^= _cj[ r ][ i ];

    _sequence_count++;
}
    
//
// initialise Niederreiter sequence
//
void
TNiederreiter::init ( uint dim )
{
    std::cout << "niederreiter" << std::endl;
    if (( dim < 1 ) || ( dim > 6 ))
    {
        std::cout << "(TNiederreiter) init : invalid dimension; valid = 1 ... 6" << std::endl;
        return;
    }// if

    _dim = dim;
    
    calculate_cj( _cj, _dim );

    for ( uint i = 0; i < _dim; i++ )
        _nextq[ i ] = 0;
    
    _sequence_count = 0;
}

///////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////
//
// local functions (implementation)
//
///////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////


//
// Multiply polynomials over Z_2.
// Notice use of a temporary vector,
// side-stepping aliasing issues when
// one of inputs is the same as the output
// [especially important in the original fortran version, I guess].
//
static void
poly_multiply ( const int pa[], int   pa_degree,
                const int pb[], int   pb_degree,
                int       pc[], int * pc_degree )
{
    int pt[ NIED2_MAX_DEGREE + 1 ];
    int pt_degree = pa_degree + pb_degree;
    
    for ( int k = 0; k <= pt_degree; k++ )
    {
        int term = 0;
        
        for ( int j = 0; j <= k; j++ )
        {
            const int conv_term = NIED2_MUL(pa[k-j], pb[j]);
            
            term = NIED2_ADD(term, conv_term);
        }// for
        
        pt[k] = term;
    }// for
    
    for ( int k = 0; k <= pt_degree; k++ )
        pc[k] = pt[k];
    
    for ( int k = pt_degree+1; k <= NIED2_MAX_DEGREE; k++ )
        pc[k] = 0;
    
    *pc_degree = pt_degree;
}

// Calculate the values of the constants V(J,R) as
// described in BFN section 3.3.
// 
//  px = appropriate irreducible polynomial for current dimension
//  pb = polynomial defined in section 2.3 of BFN.
//  pb is modified
//
static void
calculate_v ( const int px[], int   px_degree,
              int       pb[], int * pb_degree,
              int       v[],  int   maxv )
{
    const int nonzero_element   = 1;  // nonzero element of Z_2
    const int arbitrary_element = 1;  // arbitray element of Z_2

    //
    // The polynomial ph is px**(J-1), which is the value of B on arrival.
    // In section 3.3, the values of Hi are defined with a minus sign:
    // don't forget this if you use them later !
    //
    
    int ph[ NIED2_MAX_DEGREE + 1 ];
    int bigm = *pb_degree;      // m from section 3.3

    for ( int k = 0; k <= NIED2_MAX_DEGREE; k++ )
        ph[k] = pb[k];

    //
    // Now multiply B by PX so B becomes PX**J.
    // In section 2.3, the values of Bi are defined with a minus sign :
    // don't forget this if you use them later !
    // ( m from section 2.3 )
    //
    
    poly_multiply( px, px_degree, pb, *pb_degree, pb, pb_degree );

    int m = *pb_degree;
    
    // Now choose a value of Kj as defined in section 3.3.
    // We must have 0 <= Kj < E*J = M.
    // The limit condition on Kj does not seem very relevant
    // in this program.
    //
    
    //
    // Quoting from BFN: "Our program currently sets each K_q
    // equal to eq. This has the effect of setting all unrestricted
    // values of v to 1."
    // Actually, it sets them to the arbitrary chosen value.
    // Whatever.
    //
    
    int kj = bigm;

    //
    // Now choose values of V in accordance with
    // the conditions in section 3.3.
    //
    
    for ( int r = 0; r < kj; r++ )
        v[r] = 0;

    v[kj] = 1;


    if ( kj >= bigm )
    {
        for ( int r = kj+1; r < m; r++ )
            v[r] = arbitrary_element;
    }// if
    else
    {
        // This block is never reached.

        int term = NIED2_SUB( 0, ph[kj] );

        for ( int r = kj+1; r < bigm; r++ )
        {
            v[r] = arbitrary_element;

            // Check the condition of section 3.3,
            // remembering that the H's have the opposite sign.  [????????]
            //
            
            term = NIED2_SUB( term, NIED2_MUL( ph[r], v[r] ) );
        }// for

        // Now v[bigm] != term.
        v[ bigm ] = NIED2_ADD( nonzero_element, term );

        for ( int r = bigm+1; r < m; r++ )
            v[r] = arbitrary_element;
    }// else

    //
    // Calculate the remaining V's using the recursion of section 2.3,
    // remembering that the B's have the opposite sign.
    //
    
    for ( int r = 0; r <= maxv-m; r++ )
    {
        int term = 0;
        
        for ( int k = 0; k < m; k++ )
            term = NIED2_SUB( term, NIED2_MUL( pb[k], v[r+k] ) );
        
        v[r+m] = term;
    }// for
}

static void
calculate_cj ( int  cj[ NIED2_NBITS ][ NIED2_MAX_DIMENSION ],
               uint dim )
{
    int  ci[ NIED2_NBITS ][ NIED2_NBITS ];
    int  v[ MAXV + 1 ];

    for ( uint i = 0; i < dim; i++ )
    {
        const int poly_index = i + 1;

        //
        // Niederreiter (page 56, after equation (7), defines two
        // variables Q and U.  We do not need Q explicitly, but we
        // do need U.
        //
        
        int u = 0;

        // For each dimension, we need to calculate powers of an
        // appropriate irreducible polynomial, see Niederreiter
        // page 65, just below equation (19).
        // Copy the appropriate irreducible polynomial into PX,
        // and its degree into E.  Set polynomial B = PX ** 0 = 1.
        // M is the degree of B.  Subsequently B will hold higher
        // powers of PX.
        //
        
        int pb[ NIED2_MAX_DEGREE + 1 ];
        int px[ NIED2_MAX_DEGREE + 1 ];
        int px_degree = poly_degree[poly_index];
        int pb_degree = 0;
        int k;

        for ( k = 0; k <= px_degree; k++ )
        {
            px[k] = primitive_poly[poly_index][k];
            pb[k] = 0;
        }// for

        for ( ; k < NIED2_MAX_DEGREE + 1; k++ )
        {
            px[k] = 0;
            pb[k] = 0;
        }// for

        pb[0] = 1;

        for ( int j = 0; j < NIED2_NBITS; j++ )
        {
            //
            // If U = 0, we need to set B to the next power of PX
            // and recalculate V.
            //
            
            if ( u == 0)
                calculate_v( px, px_degree, pb, & pb_degree, v, MAXV );

            //
            // Now C is obtained from V.  Niederreiter
            // obtains A from V (page 65, near the bottom), and then gets
            // C from A (page 56, equation (7)).  However this can be done
            // in one step.  Here CI(J,R) corresponds to
            // Niederreiter's C(I,J,R).
            //

            for ( int r = 0; r < NIED2_NBITS; r++)
                ci[r][j] = v[r+u];

            // Advance Niederreiter's state variables.
            if ( ++u == px_degree )
                u = 0;
        }// for

        //
        // The array CI now holds the values of C(I,J,R) for this value
        // of I.  We pack them into array CJ so that CJ(I,R) holds all
        // the values of C(I,J,R) for J from 1 to NBITS.
        //

        for ( int r = 0; r < NIED2_NBITS; r++ )
        {
            int term = 0;
            
            for ( int j = 0; j < NIED2_NBITS; j++ )
                term = 2*term + ci[r][j];

            cj[r][i] = term;
        }// for
    }// for
}

}// namespace
