//
// Project : 
// File    : TString.cc
// Purpose : class for a string
// Author  : Ronald Kriemann
//
// arch-tag: dd9fb2fd-94ed-41c4-8793-020d9949034d
//

#include <ctype.h>

#include <stdio.h>
#include <string.h>

#include "TString.hh"

namespace RGL
{

//
// constructor
//
TString::TString ( uint size )
        : TArray< char >( size )
{
    // reset data
    for ( uint i = 0; i < size; i++ )
        (*this)[i] = '\0';
}

TString::TString ( const TString & string )
{
    *this = string;
}

TString::TString ( const char * string, int size )
{
    set( string, size );
}

//
// access string status
//
uint
TString::length () const
{
    if ( size() == 0 )
        return 0;
    else
        return strlen( c_str() );
}

//
// import (export) from (to) normal strings
//
void
TString::copy ( char * string, uint len ) const
{
    uint  n = min( size(), len );
    
    for ( uint i = 0; i < n; i++ )
        string[i] = (*this)[i];

    // if there is some space, set an ending 0
    if ( n < len )
        string[ n ] = '\0';
}

void
TString::set ( const char * string, int len )
{
    uint  n = (len == -1 ? strlen(string) : len);
    
    set_size( n+1 );

    for ( uint i = 0; i < n; i++ )
        (*this)[i] = string[i];
    
    (*this)[n] = '\0';
}

//
// access substrings
//
TString
TString::substr ( uint pos, uint len ) const
{
    TString   temp;
    uint      max = (pos + len) < size() ? pos + len : size();

    temp.set_size( len+1 );
    
    for ( uint j = pos, k = 0; j < max; j++, k++ )
        temp[k] = (*this)[j];

    temp[len] = '\0';
    
    return temp;
}

TString 
TString::index ( uint pos ) const
{
    return substr( pos, length()-pos );
}

//
// locate substrings or char.
//
TString 
TString::strchr ( char c ) const
{
    uint  len = length();

    for ( uint i = 0; i < len; i++ )
    {
        if ( (*this)[i] == c )
            return index( i );
    }// for

    return TString();
}

TString
TString::strstr ( const TString & needle ) const
{
    bool     found;
    TString  string( *this );
    uint     n = needle.length();

    if (n == 0)
        return string;
    
    while ((string = string.strchr( needle[0] )).length() > 0)
    {
        found = true;
        
        for ( uint i = 1; i < n; i++ )
        {
            if ( string[i] != needle[i] )
            {
                found = false;
                break;
            }// if
        }// for

        if ( found )
            return string;
    }// while

    return string;
}

//
// split string into substrings seperated by delim
//
int
TString::split ( const TString & delim, TArray< TString > & substr )
{
    uint  pos       = 0;
    uint  len       = length();
    uint  len_delim = delim.length();
    char  sub[256];
    uint  n_sub     = 0;

    substr.set_size( 0 );
    
    while ( pos < len )
    {
        //
        // skip anything in delim
        //

        bool skip = true;
        
        while ( skip && ( pos < len ))
        {
            skip = false;
            
            for ( uint i = 0; i < len_delim; i++ )
            {
                if ( (*this)[pos] == delim[i] )
                    skip = true;
            }// for

            if ( skip )
                pos++;
        }// while

        //
        // read substr until next delimiter occurs
        //

        uint  j = 0;
        bool  cont = true;
        
        while ( cont && ( pos < len ) )
        {
            for ( uint i = 0; i < len_delim; i++ )
            {
                if ( (*this)[pos] == delim[i] )
                {
                    cont = false;
                    break;
                }// if
            }// for

            if ( cont )
                sub[j++] = (*this)[pos++];
        }// while

        if ( j > 0 )
        {
            sub[j] = '\0';
            substr.set_size( n_sub+1 );
            substr[n_sub] = sub;
            n_sub++;
        }// if
    }// while

    return n_sub;
}
    
//
// various operators for conversion and manipulation
//

TString &
TString::operator = ( const TString & string )
{
    set_size( string.size() );

    for ( uint i = 0; i < size(); i++ )
        (*this)[i] = string[i];

    return *this;
}

TString &
TString::operator = ( char c )
{
    if ( size() < 2 )
        set_size( 2 );

    (*this)[0] = c;
    (*this)[1] = '\0';

    return *this;
}

TString & 
TString::operator += ( const TString & string )
{
    uint  eos = length();
    uint  n   = eos + string.length();

    if ( n+1 > size() )
        set_size( n+1 );
    
    for ( uint i = eos; i < n; i++ )
        (*this)[i] = string[ i - eos ];

    (*this)[n] = '\0';
    
    return *this;
}

TString & 
TString::operator += ( char c )
{
    uint n = length();

    if ( c == '\0' )
    {
        if ( n+1 > size() )
            set_size( n + 1 );
    }// if
    else
    {
        if ( n+2 > size() )
            set_size( n + 2 );
    }// else
    
    (*this)[ n ] = c;

    if ( c != '\0' )
        (*this)[ n+1 ] = '\0';

    return *this;
}

bool
TString::operator == ( const TString & string ) const
{
    uint  n = length();

    if ( n != string.length() )
        return false;

    for ( uint i = 0; i < n; i++ )
    {
        if ( (*this)[i] != string[i] )
            return false;
    }// for

    return true;
}

//
// convert characters in string to lower/upper
//
TString &
TString::to_lower ()
{
    uint  n = length();

    for ( uint i = 0; i < n; i++ )
        (*this)[i] = tolower( (*this)[i] );

    return *this;
}

TString &
TString::to_upper ()
{
    uint  n = length();

    for ( uint i = 0; i < n; i++ )
        (*this)[i] = toupper( (*this)[i] );

    return *this;
}
    
//
// misc. methods
//

//
// remove 0's at end
//
void
TString::shorten ()
{
    uint  n = length();

    if ( n >= size()-1 )
        return;

    set_size( n+1 );
    (*this)[n] = '\0';
}

}// namespace
