//
// Project : lib3D
// File    : TGraphObject.cc
// Purpose : class representing a cube
// Author  : Ronald Kriemann
//
// arch-tag: 449289fd-663b-484d-93da-b43682ae1394
//

#include "TCube.hh"

namespace RGL
{

//////////////////////////////////////////////////
//
// raytracing methods
//

//
// compute intersection with ray
// - return true if intersected, else false
//
bool
TCube::intersect ( const TRay & r, real & dist ) const
{
    real  t_min = 0.0, t, a, b;
    bool  first = true, intersected = false;
    TRay  ray;

    // compute ray in object-coordinates
    trans_ray_to_obj( r, ray );

    //
    // check against all faces of the cube
    //

    //
    // left/right faces
    //
    
    if ( ray.direction().x() != 0.0 )
    {
        // left
        t = - ( 1.0 + ray.origin().x() ) / ray.direction().x();

        if (( t >= EPSILON ) && ( first || ( t < t_min )))
        {
            a = ray.origin().y() + t * ray.direction().y();
            b = ray.origin().z() + t * ray.direction().z();

            if (( std::abs( a ) < 1.0 ) && ( std::abs( b ) < 1.0 ))
            {
                t_min       = t;
                intersected = true;
                first       = false;
            }// if
        }// if

        // right
        t = ( 1.0 - ray.origin().x() ) / ray.direction().x();

        if (( t >= EPSILON ) && ( first || ( t < t_min )))
        {
            a = ray.origin().y() + t * ray.direction().y();
            b = ray.origin().z() + t * ray.direction().z();

            if (( std::abs( a ) < 1.0 ) && ( std::abs( b ) < 1.0 ))
            {
                t_min       = t;
                intersected = true;
                first       = false;
            }// if
        }// if
    }// if

    //
    // down/up faces
    //
    
    if ( ray.direction().y() != 0.0 )
    {
        // down
        t = - ( 1.0 + ray.origin().y() ) / ray.direction().y();

        if (( t >= EPSILON ) && ( first || ( t < t_min )))
        {
            a = ray.origin().x() + t * ray.direction().x();
            b = ray.origin().z() + t * ray.direction().z();

            if (( std::abs( a ) < 1.0 ) && ( std::abs( b ) < 1.0 ))
            {
                t_min       = t;
                intersected = true;
                first       = false;
            }// if
        }// if

        // up
        t = ( 1.0 - ray.origin().y() ) / ray.direction().y();

        if (( t >= EPSILON ) && ( first || ( t < t_min )))
        {
            a = ray.origin().x() + t * ray.direction().x();
            b = ray.origin().z() + t * ray.direction().z();

            if (( std::abs( a ) < 1.0 ) && ( std::abs( b ) < 1.0 ))
            {
                t_min       = t;
                intersected = true;
                first       = false;
            }// if
        }// if
    }// if

    //
    // front/back faces
    //
    
    if ( ray.direction().z() != 0.0 )
    {
        // left
        t = - ( 1.0 + ray.origin().z() ) / ray.direction().z();

        if (( t >= EPSILON ) && ( first || ( t < t_min )))
        {
            a = ray.origin().x() + t * ray.direction().x();
            b = ray.origin().y() + t * ray.direction().y();

            if (( std::abs( a ) < 1.0 ) && ( std::abs( b ) < 1.0 ))
            {
                t_min       = t;
                intersected = true;
                first       = false;
            }// if
        }// if

        // right
        t = ( 1.0 - ray.origin().z() ) / ray.direction().z();

        if (( t >= EPSILON ) && ( first || ( t < t_min )))
        {
            a = ray.origin().x() + t * ray.direction().x();
            b = ray.origin().y() + t * ray.direction().y();

            if (( std::abs( a ) < 1.0 ) && ( std::abs( b ) < 1.0 ))
            {
                t_min       = t;
                intersected = true;
                first       = false;
            }// if
        }// if
    }// if

    dist = t_min;
    
    return intersected;
}

//
// compute normal at point
//
bool
TCube::calc_normal ( const TVector & point,
                     TVector       & normal ) const
{
    TVector  v;

    trans_to_obj( point, v );

    //
    // look, on which face the point is
    //

    if      ( std::abs( -1.0 - v.x() ) < EPSILON ) v.set( -1,  0,  0 );
    else if ( std::abs(  1.0 - v.x() ) < EPSILON ) v.set(  1,  0,  0 );
    else if ( std::abs( -1.0 - v.y() ) < EPSILON ) v.set(  0, -1,  0 );
    else if ( std::abs(  1.0 - v.y() ) < EPSILON ) v.set(  0,  1,  0 );
    else if ( std::abs( -1.0 - v.z() ) < EPSILON ) v.set(  0,  0, -1 );
    else if ( std::abs(  1.0 - v.z() ) < EPSILON ) v.set(  0,  0,  1 );
    else
        return false;

    trans_normal( v, normal );
    
    return true;
}

//////////////////////////////////////////////////
//
// misc. methods
//

//
// store bounding box defined by xy, xz and yz planes
// in given vectors
// - return false of bb is not defined (e.g. infinite object)
//
bool
TCube::bounding_box ( TVector & min_bb, TVector & max_bb ) const
{
    TVector  v[8], u[8];
    TMatrix  T( _trans );

    T.invert();

    v[0].set( -1, -1, -1 );
    v[1].set(  1, -1, -1 );
    v[2].set(  1,  1, -1 );
    v[3].set( -1,  1, -1 );
    v[4].set( -1, -1,  1 );
    v[5].set(  1, -1,  1 );
    v[6].set(  1,  1,  1 );
    v[7].set( -1,  1,  1 );

    for ( uint i = 0; i < 8; i++ )
        T.mul_vec( v[i], u[i] );

    min_bb.set(  1e100,  1e100,  1e100 );
    max_bb.set( -1e100, -1e100, -1e100 );
    
    for ( uint i = 0; i < 8; i++ )
    {
        min_bb.x() = min( min_bb.x(), u[i].x() );
        min_bb.y() = min( min_bb.y(), u[i].y() );
        min_bb.z() = min( min_bb.z(), u[i].z() );

        max_bb.x() = max( max_bb.x(), u[i].x() );
        max_bb.y() = max( max_bb.y(), u[i].y() );
        max_bb.z() = max( max_bb.z(), u[i].z() );
    }// 
    
    return true;
}

//
// print object
//
void
TCube::print () const
{
    std::cout << "cube" << std::endl;
}
    
}// namespace
