//
// Project : lib3D
// File    : TCylinder.cc
// Purpose : class representing a cylinder
// Author  : Ronald Kriemann
//
// arch-tag: 76253536-f474-4b36-84ae-f51ab0e56b50
//

#include "TCylinder.hh"

namespace RGL
{

//////////////////////////////////////////////
//
// constuctor and destructor
//

TCylinder::TCylinder ( bool closed )
{
    _closed = closed;
}

TCylinder::~TCylinder ()
{
}

//////////////////////////////////////////////////
//
// raytracing methods
//

//
// compute intersection with ray
// - return true if intersected, else false
//
bool
TCylinder::intersect ( const TRay & r, real & dist ) const
{
    TRay  ray;

    trans_ray_to_obj( r, ray );

    //
    // test against cylinder equation:
    //   x^2 + z^2 = 1.0
    //

    real d = (ray.direction().x() * ray.direction().x() +
              ray.direction().z() * ray.direction().z());

    bool cyl_is1  = false;
    bool cyl_is2  = false;
    bool dist_set = false;

    // check if ray is parallel to cylinder
    if ( d != 0.0 )
    {
        real cyl_t1, cyl_t2;
        
        //
        // ray, is not parallel => solve quadratic
        //   t_1,2 = - p/2 +- sqrt( p^2 / 4 - q )
        //

        real p = 2.0 * ( ray.direction().x() * ray.origin().x() +
                         ray.direction().z() * ray.origin().z() ) / d;
        real q = ( ray.origin().x() * ray.origin().x() +
                   ray.origin().z() * ray.origin().z() -
                   1.0 ) / d;

        q = p*p / 4.0 - q;

        // square-root imaginary => no intersection
        if ( q < 0.0 )
            return false;

        q = std::sqrt( q );
        p = -p/2.0;
        
        cyl_t1 = p + q;
        cyl_t2 = p - q;

        //
        // look if intersection is within height of cylinder
        //

        if ( cyl_t1 > 0.0 ) cyl_is1 = true;
        if ( cyl_t2 > 0.0 ) cyl_is2 = true;

        if ( cyl_is1 )
        {
            real y1 = ray.origin().y() + cyl_t1 * ray.direction().y();
            if (( y1 < -EPSILON ) || ( y1 > 1.0+EPSILON )) cyl_is1 = false;
        }// if

        if ( cyl_is2 )
        {
            real y2 = ray.origin().y() + cyl_t2 * ray.direction().y();
            if (( y2 < -EPSILON ) || ( y2 > 1.0+EPSILON )) cyl_is2 = false;
        }// if

        if ( cyl_is1 )
        {
            if ( ! dist_set ) dist = cyl_t1;
            else              dist = min( dist, cyl_t1 );

            dist_set = true;
        }// if
        
        if ( cyl_is2 )
        {
            if ( ! dist_set ) dist = cyl_t2;
            else              dist = min( dist, cyl_t2 );

            dist_set = true;
        }// if
    }// if

    //
    // test discs at both ends if cylinder is closed
    //

    bool  disc_is1 = false;
    bool  disc_is2 = false;

    if ( _closed && (ray.direction().y() != 0.0))
    {
        real  t1 = - ray.origin().y() / ray.direction().y();
        real  t2 = (1.0 - ray.origin().y()) / ray.direction().y();
        real  a, b;
        
        // check lower disc
        a = ray.origin().x() + t1 * ray.direction().x();
        b = ray.origin().z() + t1 * ray.direction().z();

        if ( (a*a + b*b) <= 1.0 + EPSILON )
        {
            disc_is1 = true;

            if ( ! dist_set ) dist = t1;
            else              dist = min( dist, t1 );

            dist_set = true;
        }// if
        
        // check lower disc
        a = ray.origin().x() + t2 * ray.direction().x();
        b = ray.origin().z() + t2 * ray.direction().z();

        if ( (a*a + b*b) <= 1.0 + EPSILON )
        {
            disc_is2 = true;

            if ( ! dist_set ) dist = t2;
            else              dist = min( dist, t2 );

            dist_set = true;
        }// if
    }// if

    // return intersection status
    return (cyl_is1 || cyl_is2 || disc_is1 || disc_is2);
}

//
// compute normal at point
//
bool
TCylinder::calc_normal ( const TVector & point,
                         TVector       & normal ) const
{
    TVector  v;
    real     d;
    
    trans_to_obj( point, v );

    d = v.x() * v.x() + v.z() * v.z();

    if ( _closed )
    {
        // check lower and upper discs
        if ((std::abs( v.y() ) <= EPSILON) && (d < 1.0))
        {
            // lower disc
            v.set( 0, -1, 0 );
            trans_normal( v, normal );
            return true;
        }// if
        else if ((std::abs( 1.0 - v.y() ) <= EPSILON) && (d < 1.0))
        {
            // upper disc
            v.set( 0, 1, 0 );
            trans_normal( v, normal );
            return true;
        }// if
    }// if

    // check cylinder
    if (( v.y() >= -EPSILON ) && ( v.y() <= 1.0+EPSILON ))
    {
        v.y() = 0.0;
        trans_normal( v, normal );
        return true;
    }// if
    else
        return false;
}

//////////////////////////////////////////////////
//
// misc. methods
//

//
// store bounding box defined by xy, xz and yz planes
// in given vectors
//
bool
TCylinder::bounding_box ( TVector & min_bb, TVector & max_bb ) const
{
    TVector  v[8], u[8];
    TMatrix  T( _trans );

    T.invert();

    v[0].set( -1, 0, -1 );
    v[1].set(  1, 0, -1 );
    v[2].set(  1, 0,  1 );
    v[3].set( -1, 0,  1 );
    v[4].set( -1, 1, -1 );
    v[5].set(  1, 1, -1 );
    v[6].set(  1, 1,  1 );
    v[7].set( -1, 1,  1 );

    for ( uint i = 0; i < 8; i++ )
        T.mul_vec( u[i], v[i] );

    min_bb.set( u[0].x(), u[0].y(), u[0].z() );
    max_bb.set( u[0].x(), u[0].y(), u[0].z() );
    
    for ( uint i = 1; i < 8; i++ )
    {
        min_bb.x() = min( min_bb.x(), u[i].x() );
        min_bb.y() = min( min_bb.y(), u[i].y() );
        min_bb.z() = min( min_bb.z(), u[i].z() );

        max_bb.x() = max( max_bb.x(), u[i].x() );
        max_bb.y() = max( max_bb.y(), u[i].y() );
        max_bb.z() = max( max_bb.z(), u[i].z() );
    }// 
    
    return true;
}

//////////////////////////////////////////////////
//
// virtual constructors
//

TCylinder *
TCylinder::copy  () const
{
    TCylinder * t = (TCylinder*) TGraphObject::copy();

    t->_closed = _closed;
    
    return t;
}

} // namespace
