//
// Project : lib3D
// File    : TSphere.cc
// Purpose : class representing a sphere
// Author  : Ronald Kriemann
//
// arch-tag: 2dc1ef0b-5da7-41d6-aa01-aee5159dbaeb
//

#include <cmath>

#include "defines.hh"

#include "TSphere.hh"

namespace RGL
{

// define to handle spheres by transformations instead
// of center and radius
#define USE_TRANS

//////////////////////////////////////////////
//
// constuctor and destructor
//

TSphere::TSphere ()
{
    _center = TVector( 0, 0, 0 );
    _radius = 1;
}

TSphere::TSphere ( const TVector & c, real r )
        : _center( c )
{
    _radius = std::abs( r );
    scale( _radius );
    move( c );
}

TSphere::TSphere ( const TVector & c, const TVector & r )
        : _center( c )
{
    scale( r );
    move( c );
    _radius = r.x();
}

TSphere::~TSphere ()
{
}

//////////////////////////////////////////////////
//
// raytracing methods
//

//
// calculate intersection with ray
//
bool
TSphere::intersect ( const TRay & r, real & dist ) const
{
#ifdef USE_TRANS
    TRay  ray;
    real  a, p, q;

    // compute ray in object-coordinates
    trans_ray_to_obj( r, ray );
    
    a = dot( ray.direction(), ray.direction() );

    p = (2.0 * dot( ray.origin(), ray.direction())) / a;
    q = (dot( ray.origin(), ray.origin() ) - 1.0) / a;

    real s = (p*p*0.25) - q;
    
    if ( s < 0 )
        return false;

    p *= -0.5;
    
    real  x1 = p + std::sqrt( s );
    real  x2 = p - std::sqrt( s );

    if      ( x2 > EPSILON ) dist = x2;
    else if ( x1 > EPSILON ) dist = x1;
    else
        return false;

    return true;
#else
    TRay  ray( r );
    real  a, p, q;

    a = dot( ray.direction(), ray.direction() );

    p = (2.0 * (dot( ray.origin(), ray.direction() ) -
                dot( ray.direction(), center() ))) / a;
    q = (dot( ray.origin(), ray.origin() ) +
         dot( center(), center() ) -
         2.0 * dot( ray.origin(), center() ) -
         (radius() * radius())) / a;

    real s = (p*p*0.25) - q;
    
    if ( s < 0.0 )
        return false;

    p *= -0.5;
    
    real  x1 = p + std::sqrt( s );
    real  x2 = p - std::sqrt( s );

    if      ( x2 > EPSILON ) dist = x2;
    else if ( x1 > EPSILON ) dist = x1;
    else
        return false;

    return true;
#endif
}

//
// compute normal at point
//
bool
TSphere::calc_normal ( const TVector & point,
                       TVector       & normal ) const
{
#ifdef USE_TRANS
    TVector  p;

    trans_to_obj( point, p );
    
    // check if really on surface
    if ( std::abs( 1.0 - p.norm2() ) < EPSILON )
    {
        trans_normal( p, normal );
        return true;
    }// if
    else
        return false;
    
#else
    
    normal = (point - center()) / radius();
    // normal.diff( point, center() ).scale( 1.0 / radius() );

    // check if really on surface
    if (std::abs( 1.0 - normal.norm() ) < EPSILON )
        return true;
    else
        return false;
#endif
}

//////////////////////////////////////////////////
//
// misc. methods
//

//
// store bounding box defined by xy, xz and yz planes
// in given vectors
//
bool
TSphere::bounding_box ( TVector & bbmin, TVector & bbmax ) const
{
#ifdef USE_TRANS
    TVector  v[8], u[8];
    TMatrix  T( _trans );

    T.invert();

    //
    // transform bounding box of original sphere into
    // coordinate system of transformed sphere and
    // determine maximal extents
    //
    
    v[0].set( -1, -1, -1 );
    v[1].set(  1, -1, -1 );
    v[2].set(  1,  1, -1 );
    v[3].set( -1,  1, -1 );
    v[4].set( -1, -1,  1 );
    v[5].set(  1, -1,  1 );
    v[6].set(  1,  1,  1 );
    v[7].set( -1,  1,  1 );

    for ( uint i = 0; i < 8; i++ )
        T.mul_vec( v[i], u[i] );

    bbmin.set( 1.0, u[0] );
    bbmax.set( 1.0, u[0] );
    
    for ( uint i = 1; i < 8; i++ )
    {
        bbmin.x() = min( bbmin.x(), u[i].x() );
        bbmin.y() = min( bbmin.y(), u[i].y() );
        bbmin.z() = min( bbmin.z(), u[i].z() );

        bbmax.x() = max( bbmax.x(), u[i].x() );
        bbmax.y() = max( bbmax.y(), u[i].y() );
        bbmax.z() = max( bbmax.z(), u[i].z() );
    }// for

    return true;
#else
    TVector  e( 1, 1, 1 );

    bbmin = _center - (_radius * e);
    bbmax = _center + (_radius * e);

    return true;
#endif
}

//
// print object
//
void
TSphere::print () const
{
    std::cout << "sphere: " << _center << ", " << _radius << std::endl;
}
    
//////////////////////////////////////////////////
//
// virtual constructors
//

TSphere *
TSphere::copy  () const
{
    TSphere * s = (TSphere*) TGraphObject::copy();

    s->_center = _center;
    s->_radius = _radius;
    
    return s;
}
    
} // namespace

