//
// Project : lib3D
// File    : TParser.cc
// Purpose : class for reading world-descriptions
// Author  : Ronald Kriemann
//
// arch-tag: 01c8d193-522f-4048-97a7-818da3727bf5
//

#include <iostream>

#include "TVector.hh"
#include "TMatrix.hh"
#include "TSphere.hh"

#include "TParser.hh"

// parser used in the yacc-file
RGL::TParser * parser;

// external parse-function
extern int parse_file ( RGL::TParser * parser );

// string-comparison
#define STREQ( s1, s2 )  (strcmp( s1, s2 ) == 0)

namespace RGL
{

///////////////////////////////////////////////////////
///////////////////////////////////////////////////////
//
// Error handling
//
///////////////////////////////////////////////////////
///////////////////////////////////////////////////////

void
parse_warn ( const TString           & msg,
             const TParser::TTextPos & pos )
{
    std::cout << "(TParser) WARNING : " << msg << " at " << pos << std::endl;
//    exit( -1 );
}

void
parse_error ( const TString           & msg,
              const TParser::TTextPos & pos )
{
    std::cout << "(TParser) ERROR : " << msg << " at " << pos << std::endl;
    exit( -1 );
}

void
parse_error ( const TString & msg )
{
    std::cout << "(TParser) ERROR : " << msg << std::endl;
    exit( -1 );
}

///////////////////////////////////////////////////////
///////////////////////////////////////////////////////
//
// external node-creation/-handling
//
///////////////////////////////////////////////////////
///////////////////////////////////////////////////////

// objects
TParser::TNode * new_number ( real f );
TParser::TNode * new_bool ( bool b );
TParser::TNode * new_string ( const TString & s );
TParser::TNode * new_vector ( TParser::TNode * n0, TParser::TNode * n1,
                              TParser::TNode * n2 );
TParser::TNode * new_sphere ( TParser::TNode * n0, TParser::TNode * n1 );
TParser::TNode * new_plane ( TParser::TNode * n0, TParser::TNode * n1,
                             TParser::TNode * n2 );
TParser::TNode * new_plane ( TParser::TNode * n0, TParser::TNode * n1 );
TParser::TNode * new_triangle ( TParser::TNode * n0, TParser::TNode * n1,
                                TParser::TNode * n2 );
TParser::TNode * new_cube ();
TParser::TNode * new_cylinder ( TParser::TNode * n );
TParser::TNode * new_addworld ( TParser::TNode * n, TWorld * world );
TParser::TNode * new_apply ( TParser::TNode * n0, TParser::TNode * n1 );
TParser::TNode * new_camera ( TParser * parser, TParser::TNode * orig,
                              TParser::TNode * dir, TParser::TNode * up );


// functions
TParser::TNode * new_negate ( TParser::TNode * n );
TParser::TNode * new_incdec ( TParser::TNode * n, bool dec, bool post );
TParser::TNode * new_mathfunc ( TParser::TNode * n, uint type );
TParser::TNode * new_add ( TParser::TNode * n0, TParser::TNode * n1 );
TParser::TNode * new_sub ( TParser::TNode * n0, TParser::TNode * n1 );
TParser::TNode * new_mul ( TParser::TNode * n0, TParser::TNode * n1 );
TParser::TNode * new_div ( TParser::TNode * n0, TParser::TNode * n1 );
TParser::TNode * new_assign ( TParser::TNode * n0, TParser::TNode * n1 );
TParser::TNode * new_addsubassign ( TParser::TNode * n0, TParser::TNode * n1, bool sub );
TParser::TNode * new_compare ( TParser::TNode * n0, TParser::TNode * n1, uint type );
TParser::TNode * new_andor ( TParser::TNode * n0, TParser::TNode * n1, bool b );
TParser::TNode * new_not ( TParser::TNode * n );

// transformations
TParser::TNode * new_move ( TParser::TNode * n );
TParser::TNode * new_scale ( TParser::TNode * n );
TParser::TNode * new_rotate ( TParser::TNode * n0, TParser::TNode * n1 );
TParser::TNode * new_transform ();

void add_transform ( TParser::TNode * trans, TParser::TNode * n );

// texture definitions
TParser::TNode * new_rgb_colour ( TParser::TNode * n0,
                                  TParser::TNode * n1,
                                  TParser::TNode * n2 );
TParser::TNode * new_hsv_colour ( TParser::TNode * n0,
                                  TParser::TNode * n1,
                                  TParser::TNode * n2 );
TParser::TNode * new_colour ( TParser::TNode * n );
TParser::TNode * new_ambient ( TParser::TNode * n );
TParser::TNode * new_diffuse ( TParser::TNode * n );
TParser::TNode * new_specular ( TParser::TNode * n0, TParser::TNode * n1 );
TParser::TNode * new_transmiss ( TParser::TNode * n );
TParser::TNode * new_ior ( TParser::TNode * n );
TParser::TNode * new_texture ();
void add_texture ( TParser::TNode * n0, TParser::TNode * n1 );

// lights
TParser::TNode * new_ambient_light ( TParser::TNode * n );
TParser::TNode * new_point_light ( TParser::TNode * n0,
                                   TParser::TNode * n1 );
TParser::TNode * new_spot_light ( TParser::TNode * n0,
                                  TParser::TNode * n1,
                                  TParser::TNode * n2,
                                  TParser::TNode * n3 );
TParser::TNode * new_addlight ( TParser::TNode * n,
                                TWorld         * world );

// program statements
TParser::TNode * new_while ( TParser::TNode * stop, TParser::TNode * body );
TParser::TNode * new_if   ( TParser::TNode * cond, TParser::TNode * block1,
                            TParser::TNode * block2 );
TParser::TNode * new_stmt ( TParser::TNode * n );
TParser::TNode * add_stmt ( TParser::TNode * node1, TParser::TNode * node2 );
TParser::TNode * add_stmt_to_block ( TParser::TNode * node1, TParser::TNode * node2 );
TParser::TNode * new_block ( TParser::TNode * n );
TParser::TNode * new_block ();
TParser::TNode * add_arglist ( TParser::TNode * node1, TParser::TNode * node2 );
TSLL< TParser::TNode * > * get_arglist ( TParser::TNode * n );

///////////////////////////////////////////////////////
///////////////////////////////////////////////////////
//
// Node-Data Classes
//
///////////////////////////////////////////////////////
///////////////////////////////////////////////////////

uint
TNDNumber::type () const
{
    return TYPE_NUMBER;
}

bool
TNDNumber::is_type  ( uint t ) const
{
    return ((t == TYPE_NUMBER) || TNodeData::is_type(t));
}
    
TParser::TNodeData *
TNDNumber::copy () const
{
    return new TNDNumber(_number);
}


uint
TNDBool::type () const
{
    return TYPE_BOOL;
}

bool
TNDBool::is_type  ( uint t ) const
{
    return ((t == TYPE_BOOL) || TNodeData::is_type(t));
}
    
TParser::TNodeData *
TNDBool::copy () const
{
    return new TNDBool(_boolean);
}


uint
TNDString::type () const
{
    return TYPE_STRING;
}

bool
TNDString::is_type  ( uint t ) const
{
    return ((t == TYPE_STRING) || TNodeData::is_type(t));
}
    
TParser::TNodeData *
TNDString::copy () const
{
    return new TNDString(_string);
}


uint
TNDVector::type () const
{
    return TYPE_VECTOR;
}

bool
TNDVector::is_type  ( uint t ) const
{
    return ((t == TYPE_VECTOR) || TNodeData::is_type(t));
}
    
TParser::TNodeData *
TNDVector::copy () const
{
    return new TNDVector(_vec);
}


uint
TNDGraphObject::type () const
{
    return TYPE_GRAPH_OBJECT;
}

bool
TNDGraphObject::is_type  ( uint t ) const
{
    return ((t == TYPE_GRAPH_OBJECT) || TNodeData::is_type(t));
}

void
TNDGraphObject::reset_object ()
{
    delete _obj;
    _obj = NULL;
}

TParser::TNodeData *
TNDGraphObject::copy () const
{
    if ( _obj != NULL ) return new TNDGraphObject( _obj->copy() );
    else                return new TNDGraphObject( NULL );
}

///////////////////////////////////////////////////////
///////////////////////////////////////////////////////
//
// Variables
//
///////////////////////////////////////////////////////
///////////////////////////////////////////////////////

void
TParser::TVariable::set_data ( TParser::TNodeData       * data,
                               const TParser::TTextPos  & pos )
{
    if ( _constant )
    {
        if ( _data == NULL )
            _data = data;
        else
            parse_error( "can not assign to a constant", pos );
    }// if
    else
    {
        delete _data;
        _data = data;
    }// else
}

void
TParser::TVariable::print ( const TTextPos & pos ) const
{
    std::cout << name() << " : ";

    if ( data() != NULL )
    {
        if ( data()->type() == TYPE_UNDEF )
            std::cout << "undef";
        else if ( data()->type() == TYPE_NUMBER )
            std::cout << ((TNDNumber *) data())->number();
        else if ( data()->type() == TYPE_BOOL )
        {
            if ( ((TNDBool*) data())->boolean() )
                std::cout << "true";
            else
                std::cout << "false";
        }// if
        else if ( data()->type() == TYPE_VECTOR )
            std::cout << ((TNDVector *) data())->vector();
        else if ( data()->type() == TYPE_GRAPH_OBJECT )
            std::cout << "graphical object";
        else if ( data()->type() == TYPE_TRANSFORM )
            std::cout << "transform";
        else if ( data()->type() == TYPE_COLOUR )
            std::cout << "colour";
        
        std::cout << " [ at " << pos << " ]" << std::endl;
    }// if
    else
        std::cout << "untyped" << std::endl;
}

///////////////////////////////////////////////////////
///////////////////////////////////////////////////////
//
// Node Classes
//
///////////////////////////////////////////////////////
///////////////////////////////////////////////////////

////////////////////////////////////////////////
//
// TVarNode
//

TParser::TNodeData *
TParser::TVarNode::eval () const
{
    if ( _var != NULL )
    {
        if ( _var->data() != NULL )
            return _var->data()->copy();
        else
            parse_error( "variable \"" + _var->name() + "\" not initialised", pos() );
    }// if
    else
        parse_error( "no variable associated", pos() );

    return NULL;
}

///////////////////////////////////////////////////////
///////////////////////////////////////////////////////
//
// Parser Class
//
///////////////////////////////////////////////////////
///////////////////////////////////////////////////////

/////////////////////////////////////////////
//
// constructor and destructor
//

TParser::TParser ( TWorld * w )
{
    _world = w;

    _cam_up = VEC_Y;
}

TParser::~TParser ()
{}

//
// access data
//

void
TParser::apply_camera ( TCamera * camera ) const
{
    assert( camera != NULL );
    camera->set_origin_dir( _cam_orig, _cam_dir );
}
    
void
TParser::set_camera ( const TVector & orig,
                      const TVector & dir  )
{
    _cam_orig = orig;
    _cam_dir  = dir;
}

void
TParser::set_camera_up ( const TVector & up )
{
    _cam_up = up;
}

/////////////////////////////////////////////
//
// read/parse file
//

void
TParser::parse ( TIO * input )
{
    if ( input == NULL )
        return;
    
    _input = input;
    
    parse_file( this );

    _input = NULL;
}

//
// read from input (for scanner)
//
int
TParser::readbuf ( char * buf, int max_size )
{
    if ( _input != NULL )
    {
        TString  str( max_size );
        
        if ( _input->gets( str ) != 0 )
            return 0;
        else
        {
            str.copy( buf, max_size );
            return str.length();
        }// else
    }// if

    return 0;
}
    
/////////////////////////////////////////////
//
// build parse-tree
//

//
// convert name-string into variable
//
TParser::TNode *
TParser::name_to_var ( const char * name )
{
    //
    // scan list of IDs for given name
    //

    TSLL< TVariable * >::TIterator  iter = _var_list.first();

    while ( ! iter.eol() )
    {
        if ( iter()->name() == name )
            return new TVarNode( iter() );
        
        ++iter;
    }// while

    //
    // no variable found, create one
    //

//      std::cout << "(TParser) name_to_var : created new variable \""
//                << name << "\"" << std::endl;

    TVariable  * var = new TVariable( name );

    _var_list.append( var );

    return new TVarNode( var );
}

//
// build new node in parse-tree
//
TParser::TNode *
TParser::new_node ( uint node_type, TParser::TNode * n1 )
{
//      std::cout << "(TParser) new_node : new unary node (" << node_type << ")"
//                << std::endl;
    
    switch ( node_type )
    {
        case OP_NEG     : return new_negate( n1 );
        case OP_NOT     : return new_not( n1 );
        case OP_PREINC  : return new_incdec( n1, false, false );
        case OP_POSTINC : return new_incdec( n1, false, true );
        case OP_PREDEC  : return new_incdec( n1, true, false );
        case OP_POSTDEC : return new_incdec( n1, true, true );
        case OP_STMT    : return new_stmt( n1 );
        default :
            std::cout << "(TParser) new_node : unkown node-type" << std::endl;
    }// switch

    return NULL;
}

TParser::TNode *
TParser::new_node ( uint node_type, TParser::TNode * n1, TParser::TNode * n2 )
{
//      std::cout << "(TParser) new_node : new binary node (" << node_type << ")"
//                << std::endl;

    switch ( node_type )
    {
        case OP_ADD    : return new_add( n1, n2 );
        case OP_SUB    : return new_sub( n1, n2 );
        case OP_MUL    : return new_mul( n1, n2 );
        case OP_DIV    : return new_div( n1, n2 );
        case OP_ASSIGN    : return new_assign( n1, n2 );
        case OP_ADDASSIGN : return new_addsubassign( n1, n2, false );
        case OP_SUBASSIGN : return new_addsubassign( n1, n2, true );
        case OP_EQ     : 
        case OP_NE     : 
        case OP_LT     : 
        case OP_GT     : 
        case OP_LE     : 
        case OP_GE     : return new_compare( n1, n2, node_type );
        case OP_AND    : return new_andor( n1, n2, false );
        case OP_OR     : return new_andor( n1, n2, true );
        case OP_WHILE  : return new_while( n1, n2 );
        default :
            std::cout << "(TParser) new_node : unkown node-type" << std::endl;
    }// switch

    return NULL;
}

TParser::TNode *
TParser::new_funccall ( const char * name, TParser::TNode * args ) const
{
//      std::cout << "(TParser) new_funccall : " << name << std::endl;
    
    TParser::TTextPos  pos( text_row, text_col );

    if ( args != NULL )
        pos = args->pos();
    
    TSLL< TParser::TNode * > * arglist = get_arglist( args );

    if ( arglist == NULL )
    {
        parse_error( "no argument-list available", pos );
        return NULL;
    }// if
    
    uint                                 no_args = arglist->size();
    TSLL< TParser::TNode * >::TIterator  iter    = arglist->first();
    
    // decide upon function-name
    if ( STREQ( "sphere", name ) )
    {
        TParser::TNode *  arg1 = NULL, * arg2 = NULL;
        
        if ( no_args >= 1 ) { arg1 = iter(); ++iter; }
        if ( no_args >= 2 ) { arg2 = iter(); ++iter; }

        if ( no_args >  2 )
            parse_error( "too many arguments for \"sphere\"", pos );
        else
            return new_sphere( arg1, arg2 );
    }// if
    else if ( STREQ( "plane", name ) )
    {
        if (( no_args != 2 ) && ( no_args != 3 ))
            parse_error( "wrong number of arguments for \"plane\"", pos );
        else
        {
            TParser::TNode * arg1, * arg2, * arg3 = NULL;

            arg1 = iter(); ++iter;
            arg2 = iter(); ++iter;
            if ( no_args >= 3 ) { arg3 = iter(); ++iter; }
            
            if ( no_args == 2 ) return new_plane( arg1, arg2 );
            else                return new_plane( arg1, arg2, arg3 );
        }// else
    }// if
    else if ( STREQ( "triangle", name ) )
    {
        if ( no_args != 3 )
            parse_error( "wrong number of arguments for \"triangle\"", pos );
        else
        {
            TParser::TNode * arg1, * arg2, * arg3 = NULL;

            arg1 = iter(); ++iter;
            arg2 = iter(); ++iter;
            arg3 = iter(); ++iter;
            
            return new_triangle( arg1, arg2, arg3 );
        }// else
    }// if
    else if ( STREQ( "cube", name ) )
    {
        if ( no_args == 0 )
            return new_cube();
        else
            parse_error( "wrong number of arguments for \"cube\"", pos );
    }// if
    else if ( STREQ( "cylinder", name ) )
    {
        if ( no_args == 1 )
            return new_cylinder( iter() );
        else
            parse_error( "wrong number of arguments for \"cylinder\"", pos );
    }// if
    else if ( STREQ( "sqrt", name ) )
    {
        if ( no_args == 1 )
            return new_mathfunc( iter(), OP_SQRT );
        else
            parse_error( "incorrect number of arguments in \"sqrt\"", pos );
    }// if
    else if ( STREQ( "sin", name ) )
    {
        if ( no_args == 1 )
            return new_mathfunc( iter(), OP_SIN );
        else
            parse_error( "incorrect number of arguments in \"sin\"", pos );
    }// if
    else if ( STREQ( "cos", name ) )
    {
        if ( no_args == 1 )
            return new_mathfunc( iter(), OP_COS );
        else
            parse_error( "incorrect number of arguments in \"cos\"", pos );
    }// if
    else if ( STREQ( "tan", name ) )
    {
        if ( no_args == 1 )
            return new_mathfunc( iter(), OP_TAN );
        else
            parse_error( "incorrect number of arguments in \"tan\"", pos );
    }// if
    else if ( STREQ( "asin", name ) )
    {
        if ( no_args == 1 )
            return new_mathfunc( iter(), OP_ASIN );
        else
            parse_error( "incorrect number of arguments in \"asin\"", pos );
    }// if
    else if ( STREQ( "acos", name ) )
    {
        if ( no_args == 1 )
            return new_mathfunc( iter(), OP_ACOS );
        else
            parse_error( "incorrect number of arguments in \"acos\"", pos );
    }// if
    else if ( STREQ( "atan", name ) )
    {
        if ( no_args == 1 )
            return new_mathfunc( iter(), OP_ATAN );
        else
            parse_error( "incorrect number of arguments in \"atan\"", pos );
    }// if
    else if ( STREQ( "norm2", name ) )
    {
        if ( no_args == 1 )
            return new_mathfunc( iter(), OP_NORM2 );
        else
            parse_error( "incorrect number of arguments in \"norm2\"", pos );
    }// if
    else if ( STREQ( "add_to_world", name ) )
    {
        if ( no_args == 1 )
            return new_addworld( iter(), _world );
        else
            parse_error( "incorrect number of arguments in \"add_to_world\"", pos );
    }// if
    else if ( STREQ( "move", name ) )
    {
        if ( no_args == 1 )
            return new_move( iter() );
        else
            parse_error( "incorrect number of arguments in \"move\"", pos );
    }// if
    else if ( STREQ( "scale", name ) )
    {
        if ( no_args == 1 )
            return new_scale( iter() );
        else
            parse_error( "incorrect number of arguments in \"scale\"", pos );
    }// if
    else if ( STREQ( "rotate", name ) )
    {
        if ( no_args != 2 )
            parse_error( "incorrect number of arguments in \"rotate\"", pos );
        else
        {
            TParser::TNode *  arg1 = NULL, * arg2 = NULL;

            arg1 = iter(); ++iter;
            arg2 = iter(); ++iter;

            return new_rotate( arg1, arg2 );
        }// else
    }// if
    else if ( STREQ( "transform", name ) )
    {
        TParser::TNode * n = new_transform();

        while ( ! iter.eol() )
        {
            add_transform( n, iter() );
            ++iter;
        }// while

        return n;
    }// if
    else if ( STREQ( "apply", name ) )
    {
        if ( no_args != 2 )
            parse_error( "incorrect number of arguments in \"apply\"", pos );
        else
        {
            TParser::TNode *  arg1 = NULL, * arg2 = NULL;

            arg1 = iter(); ++iter;
            arg2 = iter(); ++iter;

            return new_apply( arg1, arg2 );
        }// else
    }// if
    else if ( STREQ( "set_camera", name ) )
    {
        if (( no_args == 2 ) || ( no_args == 3 ))
        {
            TParser::TNode *  arg1 = NULL, * arg2 = NULL, * arg3 = NULL;

            arg1 = iter(); ++iter;
            arg2 = iter(); ++iter;

            if ( no_args == 3 )
                arg3 = iter(); ++iter;

            return new_camera( (TParser*) this, arg1, arg2, arg3 );
        }// if
        else
            parse_error( "incorrect number of arguments in \"set_camera\"", pos );
    }// if
    else if ( STREQ( "rgb", name ) )
    {
        if ( no_args != 3 )
            parse_error( "incorrect number of arguments in \"rgb\"", pos );
        else
        {
            TParser::TNode *  arg1 = NULL, * arg2 = NULL, * arg3 = NULL;

            arg1 = iter(); ++iter;
            arg2 = iter(); ++iter;
            arg3 = iter(); ++iter;

            return new_rgb_colour( arg1, arg2, arg3 );
        }// else
    }// if
    else if ( STREQ( "hsv", name ) )
    {
        if ( no_args != 3 )
            parse_error( "incorrect number of arguments in \"hsv\"", pos );
        else
        {
            TParser::TNode *  arg1 = NULL, * arg2 = NULL, * arg3 = NULL;

            arg1 = iter(); ++iter;
            arg2 = iter(); ++iter;
            arg3 = iter(); ++iter;

            return new_hsv_colour( arg1, arg2, arg3 );
        }// else
    }// if
    else if ( STREQ( "colour", name ) )
    {
        if ( no_args != 1 )
            parse_error( "incorrect number of arguments in \"colour\"", pos );
        else
            return new_colour( iter() );
    }// if
    else if ( STREQ( "ambient", name ) )
    {
        if ( no_args != 1 )
            parse_error( "incorrect number of arguments in \"ambient\"", pos );
        else
            return new_ambient( iter() );
    }// if
    else if ( STREQ( "diffuse", name ) )
    {
        if ( no_args != 1 )
            parse_error( "incorrect number of arguments in \"diffuse\"", pos );
        else
            return new_diffuse( iter() );
    }// if
    else if ( STREQ( "specular", name ) )
    {
        if ( no_args != 2 )
            parse_error( "incorrect number of arguments in \"specular\"", pos );
        else
        {
            TParser::TNode *  arg1 = NULL, * arg2 = NULL;

            arg1 = iter(); ++iter;
            arg2 = iter(); ++iter;

            return new_specular( arg1, arg2 );
        }// else
    }// if
    else if ( STREQ( "transmiss", name ) )
    {
        if ( no_args != 1 )
            parse_error( "incorrect number of arguments in \"transmiss\"", pos );
        else
            return new_transmiss( iter() );
    }// if
    else if ( STREQ( "ior", name ) )
    {
        if ( no_args != 1 )
            parse_error( "incorrect number of arguments in \"ior\"", pos );
        else
            return new_ior( iter() );
    }// if
    else if ( STREQ( "texture", name ) )
    {
        TParser::TNode * n = new_texture();

        while ( ! iter.eol() )
        {
            add_texture( n, iter() );
            ++iter;
        }// while

        return n;
    }// if
    else if ( STREQ( "ambientlight", name ) )
    {
        if ( no_args != 1 )
            parse_error( "incorrect number of arguments in \"ambientlight\"", pos );
        else
            return new_ambient_light( iter() );
    }// if
    else if ( STREQ( "pointlight", name ) )
    {
        if ( no_args != 2 )
            parse_error( "incorrect number of arguments in \"pointlight\"", pos );
        else
        {
            TParser::TNode *  arg1 = NULL, * arg2 = NULL;

            arg1 = iter(); ++iter;
            arg2 = iter(); ++iter;

            return new_point_light( arg1, arg2 );
        }// else
    }// if
    else if ( STREQ( "spotlight", name ) )
    {
        if ( no_args != 4 )
            parse_error( "incorrect number of arguments in \"spotlight\"", pos );
        else
        {
            TParser::TNode *  arg[4] = { NULL, NULL, NULL, NULL };

            arg[0] = iter(); ++iter;
            arg[1] = iter(); ++iter;
            arg[2] = iter(); ++iter;
            arg[3] = iter(); ++iter;

            return new_spot_light( arg[0], arg[1], arg[2], arg[3] );
        }// else
    }// if
    else if ( STREQ( "add_light", name ) )
    {
        if ( no_args != 1 )
            parse_error( "incorrect number of arguments in \"add_light\"", pos );
        else
            return new_addlight( iter(), _world );
    }// if
    else
        parse_error( "unkown function \"" + TString( name ) + "\"", pos );

    // TODO : remove arg-list
    
    return NULL;
}

TParser::TNode *
TParser::new_datanode ( uint dtype, void * val )
{
//    std::cout << "(TParser) new_datanode : type = " << dtype << std::endl;
    
    TNode * node = NULL;

    switch ( dtype )
    {
        case TYPE_NUMBER : return new_number( * ((double*) val) );
        case TYPE_BOOL   : return new_bool( * ((bool*) val) );
        case TYPE_STRING : return new_string( (char *) val );
        default:
            std::cout << "(TParser) new_datanode : unsupported type" << std::endl;
    }// switch

    return node;
}

TParser::TNode *
TParser::new_vector ( TParser::TNode * n0,
                      TParser::TNode * n1,
                      TParser::TNode * n2 ) const
{
//    std::cout << "(TParser) new_vector" << std::endl;
    return RGL::new_vector( n0, n1, n2 );
}

TParser::TNode *
TParser::new_stmt ( TNode * n ) const
{
    return RGL::new_stmt( n );
}

TParser::TNode *
TParser::new_block ( TParser::TNode * n ) const
{
    return RGL::new_block( n );
}

TParser::TNode *
TParser::new_while ( TNode * stop, TNode * block ) const
{
    return new_stmt( RGL::new_while( stop, block ) );
}

TParser::TNode *
TParser::new_if ( TNode * cond, TNode * block1, TNode * block2 ) const
{
    return new_stmt( RGL::new_if( cond, block1, block2 ) );
}

//
// set variable as constant
//
void
TParser::set_constant ( TNode * node ) const
{
    if ( node == NULL )
        parse_error( "have NULL expression as constant",
                     TTextPos( text_row, text_col ) );
    
    TVarNode * var = dynamic_cast< TVarNode * >( node );

    if ( var == NULL )
        parse_error( "unsupported constant type", var->pos() );
    else if ( var->variable() == NULL )
        parse_error( "variable not initialised", var->pos() );
    else
        var->variable()->set_constant();
}

//
// append node2 to node1 (assuming node1 is a statement)
//
TParser::TNode *
TParser::append_stmt ( TNode * node1, TNode * node2 ) const
{
//    std::cout << "(TParser) append_stmt" << std::endl;
    return add_stmt( node1, node2 );
}

//
// append node2 to node1 (assuming node1 is a argument-list)
//
TParser::TNode *
TParser::append_arg ( TNode * node1, TNode * node2 ) const
{
//    std::cout << "(TParser) append_arg" << std::endl;
    return add_arglist( node1, node2 );
}

//
// eval given parse-tree and remove it
//
void
TParser::eval_tree ( TNode * n ) const
{
//    std::cout << "(TParser) eval_tree" << std::endl;
    
    if ( n == NULL )
        parse_error( "NULL program to eval" );
    
    TNodeData * dat = n->eval();

    delete dat;
    delete n;
}

}// namespace

