//
// Project : lib3D
// File    : func_nodes.hh
// Purpose : functional parse-nodes
// Author  : Ronald Kriemann
//
// arch-tag: 08367372-6c2e-4d46-bd8c-4a7731c37a45
//

#include "TParser.hh"

namespace RGL
{

//
// negates right hand side
//

class TNegNode : public TParser::TUnaryNode
{
public:
    TNegNode ( TParser::TNode * son ) : TUnaryNode( son )
    {}

    virtual uint type () const { return (_son != NULL ? _son->type() : TYPE_UNDEF); }
    
    virtual TParser::TNodeData * eval () const
    {
        TParser::TNodeData * data = _son->eval();
    
        if ( data == NULL )
        {
            parse_warn( "NULL data in eval", _son->pos() );
            return NULL;
        }// if
        
        if ( data->is_type( TYPE_NUMBER ) )
            ((TNDNumber*) data)->number() *= -1.0;
        else if ( data->is_type( TYPE_VECTOR ) )
            ((TNDVector*) data)->vector().scale( -1.0 );
        else
        {
            parse_error( "negation not defined for datatype", _son->pos() );
            delete data;
            return NULL;
        }// if

        return data;
    }
};

//
// increment / decrement number (variable)
//

class TIncDecNode : public TParser::TUnaryNode
{
protected:
    // inc/dec, pre/post
    bool  _dec, _post;
    
public:
    TIncDecNode ( TParser::TNode * son, bool dec, bool post )
            : TUnaryNode( son ), _dec(dec), _post(post)
    {}

    virtual uint type () const { return (_son != NULL ? _son->type() : TYPE_UNDEF); }
    
    virtual TParser::TNodeData * eval () const
    {
        if ( _son == NULL )
        {
            parse_error( "NULL node during inc/dec", pos() );
            return NULL;
        }// if
        
        TParser::TVarNode * var = dynamic_cast< TParser::TVarNode * >( _son );
        
        if ( var != NULL )
        {
            if ( var->variable() == NULL )
            {
                parse_error( "no variable associated", var->pos() );
                return NULL;
            }// if
        
            if ( var->variable()->is_const() )
                parse_error( "can not assign to a constant", var->pos() );
            
            if ( var->variable()->data() == NULL )
            {
                parse_error( "variable not initialised", var->pos() );
                return NULL;
            }// if
        
            if ( var->variable()->data()->is_type( TYPE_NUMBER ) )
            {
                TNDNumber * data = NULL;
                
                if ( _post )
                    data = (TNDNumber*) var->variable()->data()->copy();
                else
                    data = (TNDNumber*) var->variable()->data();

                if ( _dec )
                    ((TNDNumber*) var->variable()->data())->number()--;
                else
                    ((TNDNumber*) var->variable()->data())->number()++;

                var->variable()->print( var->pos() );
                return data;
            }// if
            else
            {
                parse_error( "inc/dec not defined for given datatype", _son->pos() );
                return NULL;
            }// if
        }// if
        else
            parse_error( "inc/dec only defined for variables", _son->pos() );
        
        return NULL;
    }
};

//
// unary mathematical functions (sin, cos, etc)
//

class TMathFuncNode : public TParser::TUnaryNode
{
protected:
    // kind of math. function
    uint  _func_type;
    
public:
    TMathFuncNode ( TParser::TNode * son, uint ft )
            : TUnaryNode( son ), _func_type(ft)
    {}

    virtual uint type () const
    {
        if ((_son != NULL) &&
            (_son->is_type( TYPE_NUMBER ) ||
             (_son->is_type( TYPE_VECTOR ) && (_func_type == OP_NORM2))))
            return TYPE_NUMBER;
        
        return TYPE_UNDEF;
    }
    
    virtual TParser::TNodeData * eval () const
    {
        if ( _son == NULL )
            return NULL;
        
        TParser::TNodeData * data = _son->eval();
    
        if ( data == NULL )
        {
            parse_warn( "NULL data in eval", _son->pos() );
            return NULL;
        }// if
        
        if ( data->is_type( TYPE_NUMBER ) )
        {
            real f = ((TNDNumber*) data)->number();

            switch ( _func_type )
            {
                case OP_SQRT :
                    if ( f >= 0.0 )
                        f = std::sqrt( f );
                    else
                        parse_warn( "sqrt not defined for x < 0", _son->pos() );
                    break;
                case OP_SIN  : f = std::sin( f ); break;
                case OP_COS  : f = std::cos( f ); break;
                case OP_TAN  : f = std::sqrt( f ); break;
                case OP_ASIN :
                    if ( std::abs( f ) <= 1.0 )
                        f = std::asin( f );
                    else
                        parse_warn( "asin not defined for |x| > 1", _son->pos() );
                    break;
                case OP_ACOS : 
                    if ( std::abs( f ) <= 1.0 )
                        f = std::acos( f );
                    else
                        parse_warn( "acos not defined for |x| > 1", _son->pos() );
                    break;
                case OP_ATAN : 
                    if ( std::abs( f ) <= 1.0 )
                        f = std::atan( f );
                    else
                        parse_warn( "atan not defined for |x| > 1", _son->pos() );
                    break;
                case OP_NORM2 :
                    f = std::abs( f );
                default :
                    parse_error( "unkown function-type", pos() );
            }// switch

            ((TNDNumber*) data)->number() = f;
            return data;
        }// if
        else if ( data->is_type( TYPE_VECTOR ) )
        {
            if ( _func_type == OP_NORM2 )
            {
                real f = ((TNDVector*) data)->vector().norm2();

                delete data;
                return new TNDNumber( f );
            }// if
            else
            {
                parse_warn( "unknown function-type", pos() );
                delete data;
                return NULL;
            }// else
        }// if
        else
        {
            parse_error( "function only defined for numbers (or vectors)", _son->pos() );
            delete data;
            return NULL;
        }// if

        return NULL;
    }
};

//
// add two nodes together
//

class TAddNode : public TParser::TBinaryNode
{
public:
    TAddNode ( TParser::TNode * son0, TParser::TNode * son1 )
            : TParser::TBinaryNode( son0, son1 )
    {}

    virtual uint type () const
    {
        if ( _son0->type() == _son1->type() )
            return _son0->type();
        else
            return TYPE_UNDEF;
    }
    
    TParser::TNodeData * eval () const
    {
        TParser::TNodeData * data0 = _son0->eval();
        TParser::TNodeData * data1 = _son1->eval();
        
        if ( data0 == NULL )
        {
            parse_warn( "NULL data in eval", _son0->pos() );
            return NULL;
        }// if
        
        if ( data1 == NULL )
        {
            parse_warn( "NULL data in eval", _son1->pos() );
            return NULL;
        }// if
        
        if ( data0->type() != data1->type() )
        {
            parse_warn( "different types in data", _son1->pos() );
            return NULL;
        }// if
        
        TParser::TNodeData * res = NULL;
        
        if ( data0->is_type( TYPE_NUMBER ) )
        {
            res = data0->copy();
            ((TNDNumber *) res)->number() += ((TNDNumber *) data1)->number();
        }// if
        else if ( data0->is_type( TYPE_VECTOR ) )
        {
            res = data0->copy();
            ((TNDVector *) res)->vector().add( 1.0, ((TNDVector *) data1)->vector() );
        }// if
        else
            parse_error( "addition not defined for datatype", _son0->pos() );

        delete data0;
        delete data1;
        
        return res;
    }
};

//
// subtract two nodes
//

class TSubNode : public TParser::TBinaryNode
{
public:
    TSubNode ( TParser::TNode * son0, TParser::TNode * son1 )
            : TParser::TBinaryNode( son0, son1 )
    {}

    virtual uint type () const
    {
        if ( _son0->type() == _son1->type() )
            return _son0->type();
        else
            return TYPE_UNDEF;
    }
    
    TParser::TNodeData * eval () const
    {
        TParser::TNodeData * data0 = _son0->eval();
        TParser::TNodeData * data1 = _son1->eval();
        
        if ( data0 == NULL )
        {
            parse_warn( "NULL data in eval", _son0->pos() );
            return NULL;
        }// if
        
        if ( data1 == NULL )
        {
            parse_warn( "NULL data in eval", _son1->pos() );
            return NULL;
        }// if
        
        if ( data0->type() != data1->type() )
        {
            parse_warn( "different types in data", _son1->pos() );
            return NULL;
        }// if
        
        TParser::TNodeData * res = NULL;
        
        if ( data0->is_type( TYPE_NUMBER ) )
        {
            res = data0->copy();
            ((TNDNumber *) res)->number() -= ((TNDNumber *) data1)->number();
        }// if
        else if ( data0->is_type( TYPE_VECTOR ) )
        {
            res = data0->copy();
            ((TNDVector *) res)->vector().add( -1.0, ((TNDVector *) data1)->vector() );
        }// if
        else
            parse_error( "subtraction not defined for datatype", _son0->pos() );

        delete data0;
        delete data1;
        
        return res;
    }
};

//
// multiply two nodes
//

class TMulNode : public TParser::TBinaryNode
{
public:
    TMulNode ( TParser::TNode * son0, TParser::TNode * son1 )
            : TParser::TBinaryNode( son0, son1 )
    {}

    virtual uint type () const
    {
        if ( _son0->is_type( TYPE_VECTOR ) )
        {
            if ( _son1->is_type( TYPE_NUMBER ) )
                 return TYPE_VECTOR;
        }// if
        else if ( _son0->is_type( TYPE_NUMBER ) )
        {
            if ( _son1->is_type( TYPE_NUMBER ) )
                return TYPE_NUMBER;
            else if ( _son1->is_type( TYPE_VECTOR ) )
                return TYPE_VECTOR;
        }// if

        return TYPE_UNDEF;
    }
    
    TParser::TNodeData * eval () const
    {
        TParser::TNodeData * data0 = _son0->eval();
        TParser::TNodeData * data1 = _son1->eval();
        
        if ( data0 == NULL )
        {
            parse_warn( "NULL data in eval", _son0->pos() );
            return NULL;
        }// if
        
        if ( data1 == NULL )
        {
            parse_warn( "NULL data in eval", _son1->pos() );
            return NULL;
        }// if
        
        TParser::TNodeData * res = NULL;
        
        if ( data0->is_type( TYPE_NUMBER ) )
        {
            if ( data1->is_type( TYPE_NUMBER ) )
            {
                res = data0->copy();
                ((TNDNumber *) res)->number() *= ((TNDNumber *) data1)->number();
            }// if
            else if ( data1->is_type( TYPE_VECTOR ) )
            {
                res = data1->copy();
                ((TNDVector *) res)->vector().scale( ((TNDNumber *) data0)->number() );
            }// if
            else
                parse_error( "multiplication not defined for datatype", _son1->pos() );
        }// if
        else if ( data0->is_type( TYPE_VECTOR ) )
        {
            if ( data1->is_type( TYPE_NUMBER ) )
            {
                res = data0->copy();
                ((TNDVector *) res)->vector().scale( ((TNDNumber *) data1)->number() );
            }// if
            else
                parse_error( "multiplication not defined for datatype", _son1->pos() );
        }// if
        else
            parse_error( "multiplication not defined for datatype", _son0->pos() );

        delete data0;
        delete data1;
        
        return res;
    }
};

//
// divide two nodes
//

class TDivNode : public TParser::TBinaryNode
{
public:
    TDivNode ( TParser::TNode * son0, TParser::TNode * son1 )
            : TParser::TBinaryNode( son0, son1 )
    {}

    virtual uint type () const
    {
        if ( _son0->type() == _son1->type() )
            return _son0->type();
        else
            return TYPE_UNDEF;
    }
    
    TParser::TNodeData * eval () const
    {
        TParser::TNodeData * data0 = _son0->eval();
        TParser::TNodeData * data1 = _son1->eval();
        
        if ( data0 == NULL )
        {
            parse_warn( "NULL data in eval", _son0->pos() );
            return NULL;
        }// if
        
        if ( data1 == NULL )
        {
            parse_warn( "NULL data in eval", _son1->pos() );
            return NULL;
        }// if
        
        if ( data0->type() != data1->type() )
        {
            parse_warn( "different types in data", _son1->pos() );
            return NULL;
        }// if
        
        TParser::TNodeData * res = NULL;
        
        if ( data0->is_type( TYPE_NUMBER ) )
        {
            if ( ((TNDNumber *) data1)->number() == 0.0 )
                parse_warn( "division by zero", _son1->pos() );
            else
            {
                res = data0->copy();
                ((TNDNumber *) res)->number() /= ((TNDNumber *) data1)->number();
            }// else
        }// if
        else
            parse_error( "division not defined for datatype", _son0->pos() );

        delete data0;
        delete data1;
        
        return res;
    }
};

//
// assign right node to left (which must be a variable)
//

class TAssignNode : public TParser::TBinaryNode
{
public:
    TAssignNode ( TParser::TNode * son0, TParser::TNode * son1 )
            : TParser::TBinaryNode( son0, son1 )
    {}

    virtual uint type () const
    {
        return (_son1 != NULL ? _son1->type() : TYPE_UNDEF);
    }
    
    TParser::TNodeData * eval () const
    {
        if ( _son0 == NULL )
        {
            parse_warn( "NULL node0 in eval", pos() );
            return NULL;
        }// if
        
        if ( _son1 == NULL )
        {
            parse_warn( "NULL node1 in eval", pos() );
            return NULL;
        }// if
        
        //
        // check if first node is a variable
        //
        
        TParser::TVarNode * var = dynamic_cast< TParser::TVarNode * >( _son0 );
        
        if ( var != NULL )
        {
            if ( var->variable() == NULL )
                parse_error( "variable not initialised", var->pos() );

            var->variable()->set_data( _son1->eval(), var->pos() );
            
            // var->variable()->print( var->pos() );
            return var->eval();
        }// if
        else
        {
            parse_error( "can't assign to constant", _son0->pos() );
            return NULL;
        }// else
    }
};

//
// add/subtract right-hand-side from left value
//
class TAddSubAssignNode : public TParser::TBinaryNode
{
protected:
    // subtract instead of add
    bool  _sub;
    
public:
    TAddSubAssignNode ( TParser::TNode * son0, TParser::TNode * son1,
                        bool sub )
            : TParser::TBinaryNode( son0, son1 ), _sub(sub)
    {}

    virtual uint type () const
    {
        return _son0->type();
    }
    
    TParser::TNodeData * eval () const
    {
        if ( _son0 == NULL )
        {
            parse_warn( "NULL node0 in eval", pos() );
            return NULL;
        }// if
        
        if ( _son1 == NULL )
        {
            parse_warn( "NULL node1 in eval", pos() );
            return NULL;
        }// if
        
        //
        // check if first node is a variable
        //
        
        TParser::TVarNode * var = dynamic_cast< TParser::TVarNode * >( _son0 );
        
        if ( var != NULL )
        {
            if ( var->variable()->is_const() )
                parse_error( "can not assign to a constant", var->pos() );
            
            if ( var->variable()->data() == NULL )
            {
                parse_error( "variable \"" + var->variable()->name() + "\" not initialised",
                             var->pos() );
                return NULL;
            }// if

            if ( var->variable()->data()->is_type( TYPE_NUMBER ) )
            {
                TParser::TNodeData * dat = _son1->eval();
                
                if ( dat->is_type( TYPE_NUMBER ) )
                {
                    if ( _sub )
                        ((TNDNumber*) var->variable()->data())->number() -= ((TNDNumber*) dat)->number();
                    else 
                        ((TNDNumber*) var->variable()->data())->number() += ((TNDNumber*) dat)->number();
                }// if
                else
                    parse_error( "expected number in +=/-=", _son1->pos() );

                delete dat;
            }// if
            else if ( var->variable()->data()->is_type( TYPE_VECTOR ) )
            {
                TParser::TNodeData * dat = _son1->eval();
                
                if ( dat->is_type( TYPE_VECTOR ) )
                {
                    if ( _sub )
                        ((TNDVector*) var->variable()->data())->vector().add( -1.0, ((TNDVector*) dat)->vector() );
                    else 
                        ((TNDVector*) var->variable()->data())->vector().add( 1.0, ((TNDVector*) dat)->vector() );
                }// if
                else
                    parse_error( "expected vector in +=/-=", _son1->pos() );

                delete dat;
            }// if
            else
                parse_error( "unsupported type for +=/-=", var->pos() );
            
            // var->variable()->print( var->pos() );
            return var->eval();
        }// if
        else
        {
            parse_error( "can't assign to constant", _son0->pos() );
            return NULL;
        }// else
    }
};

//
// size-compare two nodes
//
class TCompareNode : public TParser::TBinaryNode
{
protected:
    // compare-op
    uint  _comp_op;
    
public:
    TCompareNode ( TParser::TNode * son0, TParser::TNode * son1,
                   uint comp_op )
            : TParser::TBinaryNode( son0, son1 ), _comp_op(comp_op)
    {}

    virtual uint type () const
    {
        if ((_son0 == NULL) || (_son1 == NULL))
            return TYPE_UNDEF;

        if (( _comp_op == OP_EQ ) || ( _comp_op == OP_NE ))
        {
            if (( _son0->is_type( TYPE_NUMBER ) && _son1->is_type( TYPE_NUMBER )) ||
                ( _son0->is_type( TYPE_VECTOR ) && _son1->is_type( TYPE_VECTOR )))
                return TYPE_BOOL;
        }// if
        else
        {
            if ( _son0->is_type( TYPE_NUMBER ) && _son1->is_type( TYPE_NUMBER ))
                return TYPE_BOOL;
        }// else

        return TYPE_UNDEF;
    }
    
    TParser::TNodeData * eval () const
    {
        if ( _son0 == NULL )
        {
            parse_warn( "NULL node0 in eval", pos() );
            return NULL;
        }// if
        
        if ( _son1 == NULL )
        {
            parse_warn( "NULL node1 in eval", pos() );
            return NULL;
        }// if
        
        TParser::TNodeData * dat0 = _son0->eval();
        TParser::TNodeData * dat1 = _son1->eval();
        TParser::TNodeData * res  = NULL;

        if ( dat0 == NULL )
        {
            parse_warn( "NULL data in eval", _son0->pos() );
            return NULL;
        }// if
        
        if ( dat1 == NULL )
        {
            parse_warn( "NULL data in eval", _son1->pos() );
            return NULL;
        }// if
        
        if ( dat0->is_type( TYPE_NUMBER ) && dat1->is_type( TYPE_NUMBER ) )
        {
            switch ( _comp_op )
            {
                case OP_EQ :
                    res = new TNDBool( ((TNDNumber*)dat0)->number() == ((TNDNumber*)dat1)->number() );
                    break;
                case OP_NE :
                    res = new TNDBool( ((TNDNumber*)dat0)->number() != ((TNDNumber*)dat1)->number() );
                    break;
                case OP_LT :
                    res = new TNDBool( ((TNDNumber*)dat0)->number() < ((TNDNumber*)dat1)->number() );
                    break;
                case OP_GT :
                    res = new TNDBool( ((TNDNumber*)dat0)->number() > ((TNDNumber*)dat1)->number() );
                    break;
                case OP_LE :
                    res = new TNDBool( ((TNDNumber*)dat0)->number() <= ((TNDNumber*)dat1)->number() );
                    break;
                case OP_GE :
                    res = new TNDBool( ((TNDNumber*)dat0)->number() >= ((TNDNumber*)dat1)->number() );
                    break;
                default :
                    parse_error( "unkown comparison type", _son0->pos() );
            }// switch
        }// if
        else if ( dat0->is_type( TYPE_VECTOR ) && dat1->is_type( TYPE_VECTOR ) )
        {
            TVector  d;

            d.set( 1.0, ((TNDVector*)dat0)->vector(),
                   -1.0, ((TNDVector*)dat1)->vector() );
            
            switch ( _comp_op )
            {
                case OP_EQ :
                    std::cout << "TODO" << std::endl;
                    res = new TNDBool( d.norm2() == 0.0 );
                    break;
                case OP_NE :
                    std::cout << "TODO" << std::endl;
                    res = new TNDBool( d.norm2() != 0.0 );
                    break;
                default :
                    parse_error( "unkown comparison type", _son0->pos() );
            }// switch
        }// if
        else
        {
            parse_error( "comparison not defined for datatype", _son0->pos() );
        }// else

        delete dat0;
        delete dat1;
        
        return res;
    }
};

//
// logical operators ("and" and "or")
//

class TAndOrNode : public TParser::TBinaryNode
{
protected:
    // use "or" instead of default "and"
    bool  _use_or;

public:
    TAndOrNode ( TParser::TNode * son0, TParser::TNode * son1,
                 bool use_or )
            : TParser::TBinaryNode( son0, son1 ), _use_or(use_or)
    {}

    virtual uint type () const
    {
        if ((_son0 == NULL) || (_son1 == NULL))
            return TYPE_UNDEF;
        
        if ((_son0->is_type( TYPE_BOOL ) || _son0->is_type( TYPE_NUMBER )) &&
            (_son1->is_type( TYPE_BOOL ) || _son1->is_type( TYPE_NUMBER )))
            return TYPE_BOOL;
        else
            return TYPE_UNDEF;
    }
    
    TParser::TNodeData * eval () const
    {
        if ((_son0 == NULL) || (_son1 == NULL))
            return NULL;
        
        TParser::TNodeData * data0, * data1;
        bool                 b0, b1;

        data0 = _son0->eval();
        if ( data0 == NULL )
        {
            parse_warn( "NULL data in eval", _son0->pos() );
            return NULL;
        }// if

        b0 = to_bool( data0, _son0->pos() );

        delete data0;

        // or-clause and left is true
        if ( _use_or && b0)
            return new TNDBool( true );

        // and-clause and right is true
        if ( ! _use_or && ! b0 )
            return new TNDBool( false );
        
        data1 = _son1->eval();
        if ( data1 == NULL )
        {
            parse_warn( "NULL data in eval", _son1->pos() );
            return NULL;
        }// if

        b1 = to_bool( data1, _son1->pos() );

        delete data1;

        if ( _use_or )
        {
            if ( b0 || b1 ) return new TNDBool( true );
        }// if
        else
        {
            if ( b0 && b1 ) return new TNDBool( true );
        }// else

        return new TNDBool( false );
    }

    bool to_bool ( TParser::TNodeData * nd, const TParser::TTextPos & pos ) const
    {
        if ( nd->is_type( TYPE_BOOL ) )
            return ((TNDBool*) nd)->boolean();
        else if ( nd->is_type( TYPE_NUMBER ) )
            return ((TNDNumber*) nd)->number() == 0;
        else
        {
            parse_warn( "can not convert type to boolean", pos );
            return false;
        }// else
    }
};

//
// logical operator (not)
//

class TNotNode : public TParser::TUnaryNode
{
public:
    TNotNode ( TParser::TNode * son )
            : TParser::TUnaryNode( son )
    {}

    virtual uint type () const
    {
        if (_son == NULL)
            return TYPE_UNDEF;
        
        if ( _son->is_type( TYPE_BOOL ) || _son->is_type( TYPE_NUMBER ) )
            return TYPE_BOOL;
        else
            return TYPE_UNDEF;
    }
    
    TParser::TNodeData * eval () const
    {
        if (_son == NULL)
            return NULL;
        
        TParser::TNodeData * data;
        bool                 b;

        data = _son->eval();
        if ( data == NULL )
        {
            parse_warn( "NULL data in eval", _son->pos() );
            return NULL;
        }// if

        b = to_bool( data, _son->pos() );

        delete data;

        return new TNDBool( ! b );
    }

    bool to_bool ( TParser::TNodeData * nd, const TParser::TTextPos & pos ) const
    {
        if ( nd->is_type( TYPE_BOOL ) )
            return ((TNDBool*) nd)->boolean();
        else if ( nd->is_type( TYPE_NUMBER ) )
            return ((TNDNumber*) nd)->number() == 0;
        else
        {
            parse_warn( "can not convert type to boolean", pos );
            return false;
        }// else
    }
};

///////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////
//
// return local node types to outer world
//
///////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////

TParser::TNode * new_negate ( TParser::TNode * n )
{
    return new TNegNode( n );
}

TParser::TNode * new_incdec ( TParser::TNode * n, bool dec, bool post )
{
    return new TIncDecNode( n, dec, post );
}

TParser::TNode * new_mathfunc ( TParser::TNode * n, uint type )
{
    return new TMathFuncNode( n, type );
}

TParser::TNode * new_add ( TParser::TNode * n0, TParser::TNode * n1 )
{
    return new TAddNode( n0, n1 );
}

TParser::TNode * new_sub ( TParser::TNode * n0, TParser::TNode * n1 )
{
    return new TSubNode( n0, n1 );
}

TParser::TNode * new_mul ( TParser::TNode * n0, TParser::TNode * n1 )
{
    return new TMulNode( n0, n1 );
}

TParser::TNode * new_div ( TParser::TNode * n0, TParser::TNode * n1 )
{
    return new TDivNode( n0, n1 );
}

TParser::TNode * new_assign ( TParser::TNode * n0, TParser::TNode * n1 )
{
    return new TAssignNode( n0, n1 );
}

TParser::TNode * new_addsubassign ( TParser::TNode * n0, TParser::TNode * n1, bool sub )
{
    return new TAddSubAssignNode( n0, n1, sub );
}

TParser::TNode * new_compare ( TParser::TNode * n0, TParser::TNode * n1, uint type )
{
    return new TCompareNode( n0, n1, type );
}

TParser::TNode * new_andor ( TParser::TNode * n0, TParser::TNode * n1, bool b )
{
    return new TAndOrNode( n0, n1, b );
}

TParser::TNode * new_not ( TParser::TNode * n )
{
    return new TNotNode( n );
}

}// namespace
