//
// Project : lib3D
// File    : prog_nodes.hh
// Purpose : programming parse-nodes
// Author  : Ronald Kriemann
//
// arch-tag: 658941fd-02cb-4821-9cb4-4128671ac387
//

#include "TParser.hh"

namespace RGL
{

/////////////////////////////////////////////////////
//
// while-loop
//
/////////////////////////////////////////////////////

class TWhileNode : public TParser::TNode
{
protected:
    // stop-criterion and body of loop
    TParser::TNode  * _stop, * _body;
    
public:
    TWhileNode ( TParser::TNode * stop, TParser::TNode * body )
            : _stop(stop), _body(body)
    {
        if ( _stop == NULL )
            parse_warn( "NULL stop-criterion in while-loop", pos() );
        
        if ( _body == NULL )
            parse_warn( "NULL body in while-loop", pos() );
    }

    virtual uint type () const
    {
        return TYPE_UNDEF;
    }
    
    TParser::TNodeData * eval () const
    {
        if ( _stop == NULL )
            return NULL;

        if ( _body == NULL )
            return NULL;
        
        TParser::TNodeData  * dat0, * dat1 = NULL;

        dat0 = _stop->eval();
            
        while ( to_bool( dat0, _stop->pos() ) )
        {
            delete dat0;
            delete dat1;

            dat1 = _body->eval();
            dat0 = _stop->eval();
        }// while

        delete dat0;
        
        return dat1;
    }

    bool to_bool ( TParser::TNodeData * nd, const TParser::TTextPos & pos ) const
    {
        if ( nd->is_type( TYPE_BOOL ) )
            return ((TNDBool*) nd)->boolean();
        else if ( nd->is_type( TYPE_NUMBER ) )
            return ((TNDNumber*) nd)->number() == 0;
        else
        {
            parse_warn( "can not convert type to boolean", pos );
            return false;
        }// else
    }
};

/////////////////////////////////////////////////////
//
// if-statement
//
/////////////////////////////////////////////////////

class TIfNode : public TParser::TNode
{
protected:
    // condition and both blocks
    TParser::TNode  * _cond, * _block_if, * _block_else;
    
public:
    TIfNode ( TParser::TNode * cond,
              TParser::TNode * block_if,
              TParser::TNode * block_else )
            : _cond(cond), _block_if(block_if), _block_else(block_else)
    {
        if ( _cond == NULL )
            parse_warn( "NULL condition in if-statement", pos() );
        
        if ( _block_if == NULL )
            parse_warn( "NULL if-block in if-statement", pos() );
    }

    virtual uint type () const
    {
        return TYPE_UNDEF;
    }
    
    TParser::TNodeData * eval () const
    {
        if ( _cond == NULL )
            parse_error( "condition is NULL in if-statement", pos() );

        if ( _block_if == NULL )
            parse_error( "if-block is NULL in if-statement", pos() );
        
        TParser::TNodeData  * cond = _cond->eval();
        bool                  res  = to_bool( cond, _cond->pos() );

        delete cond;
        
        if ( res )
            return _block_if->eval();
        else if ( _block_else != NULL )
            return _block_else->eval();

        return NULL;
    }

    bool to_bool ( TParser::TNodeData * nd, const TParser::TTextPos & pos ) const
    {
        if ( nd->is_type( TYPE_BOOL ) )
            return ((TNDBool*) nd)->boolean();
        else if ( nd->is_type( TYPE_NUMBER ) )
            return ((TNDNumber*) nd)->number() == 0;
        else
        {
            parse_warn( "can not convert type to boolean", pos );
            return false;
        }// else
    }
};

/////////////////////////////////////////////////////
//
// node representing one element in a node-list
//
/////////////////////////////////////////////////////

class TStmtNode : public TParser::TNode
{
protected:
    // associated statement
    TNode     * _stmt;
    
    // next node in statement-list
    TStmtNode * _next;
    
public:
    // constructor
    TStmtNode ( TParser::TNode * n ) : _stmt(n) {}
    
    ~TStmtNode () { delete _stmt; delete _next; }
    
    // return next statement
    TStmtNode * next () { return _next; }
    
    // append node to stmt-list
    void append ( TStmtNode * n ) { _next = n; }
    
    // eval node and return result
    virtual TParser::TNodeData * eval () const
    {
        TParser::TNodeData * res = NULL;
        
        if ( _stmt != NULL )
            res = _stmt->eval();
        
        if ( _next != NULL )
        {
            delete res;
            res = _next->eval();
        }// if
        
        return res;
    }
};

/////////////////////////////////////////////////////
//
// node representing a block of statements
//
/////////////////////////////////////////////////////

class TBlockNode : public TParser::TNode
{
protected:
    // first statement in block
    TParser::TNode * _stmts;

public:
    TBlockNode ( TParser::TNode * stmts )
            : _stmts(stmts)
    {}

    ~TBlockNode () { delete _stmts; }

    // eval node and return result
    virtual TParser::TNodeData * eval () const
    {
        if ( _stmts != NULL )
            return _stmts->eval();
        else
            return NULL;
    }
};

/////////////////////////////////////////////////////
//
// argument lists
//
/////////////////////////////////////////////////////

class TArgListNode : public TParser::TNode
{
protected:
    // list of arguments
    TSLL< TParser::TNode * >  _list;
    
public:
    TArgListNode () {}

    virtual uint type () const { return TYPE_UNDEF; }

    // append node to argument-list
    virtual void append ( TParser::TNode * n )
    {
        if ( n != NULL )
            _list.append( n );
        else
            parse_warn( "appending NULL node", TParser::TTextPos( text_row, text_col ) );
    }

    // return number of arguments
    uint size () const { return _list.size(); }

    // return argument-list
    TSLL< TParser::TNode * > & arguments () { return _list; }

    // no evaluation of argument-list needed
    TParser::TNodeData * eval () const
    {
        return NULL;
    }
};

///////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////
//
// return local node types to outer world
//
///////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////

TParser::TNode * new_while ( TParser::TNode * stop, TParser::TNode * body )
{
    return new TWhileNode( stop, body );
}

TParser::TNode * new_if ( TParser::TNode * cond,
                          TParser::TNode * block1,
                          TParser::TNode * block2 )
{
    return new TIfNode( cond, block1, block2 );
}

TParser::TNode * new_stmt ( TParser::TNode * n )
{
    return new TStmtNode( n );
}

TParser::TNode * new_block ( TParser::TNode * n )
{
    return new TBlockNode( n );
}

TParser::TNode * add_stmt ( TParser::TNode * node1, TParser::TNode * node2 )
{
    // std::cout << "append_stmt : " << node1 << " - " << node2 << std::endl;
    
    if ( node1 == NULL )
        return node2;

    if ( node2 == NULL )
        return node1;
    
    TStmtNode  * stmt1 = dynamic_cast< TStmtNode * >( node1 );

    if ( stmt1 == NULL )
        parse_error( "can not append to non-statement-node", node1->pos() );
    else
    {
        if ( node2 != NULL )
        {
            TStmtNode  * stmt2 = dynamic_cast< TStmtNode * >( node2 );

            if ( stmt2 == NULL )
                parse_error( "can not append non-statement-node", node2->pos() );
            else
            {
                TStmtNode  * last = stmt1;

                while ( last->next() != NULL )
                    last = last->next();
                
                last->append( stmt2 );
            }// else
        }// if
        else
            parse_warn( "try to append NULL node", node2->pos() );
    }// else

    return node1;
}

TParser::TNode * add_arglist ( TParser::TNode * node1, TParser::TNode * node2 )
{
    //std::cout << "append_arg : " << node1 << " - " << node2 << std::endl;

    TArgListNode * args;
    
    if ( node1 == NULL )
        args = new TArgListNode();
    else
        args = dynamic_cast< TArgListNode * >( node1 );

    if ( args == NULL )
    {
        parse_error( "no argument-list to append to",
                     TParser::TTextPos( text_row, text_col ) );
        return NULL;
    }// if

    if ( node2 == NULL )
        return args;

    args->append( node2 );

    return args;
}

TSLL< TParser::TNode * > * get_arglist ( TParser::TNode * n )
{
    assert( n != NULL );
    
    TArgListNode * arg = dynamic_cast< TArgListNode * >( n );
    
    if ( arg == NULL )
    {
        parse_error( "cant get argument-list from non-arglist-node",
                     n->pos() );
        return NULL;
    }// if

    return & arg->arguments();
}

}// namespace
