//
// Project : lib3D
// File    : trans_nodes.hh
// Purpose : transformation parse-nodes
// Author  : Ronald Kriemann
//
// arch-tag: 927273fd-1eda-4eec-b6bf-444ccf0010f2
//

#include "TParser.hh"

namespace RGL
{

///////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////
//
// local nodedata-types
//
///////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////

class TNDTransform : public TParser::TNodeData
{
protected:
    TMatrix  _mat;

public:
    TNDTransform () {}
    TNDTransform ( const TMatrix & mat ) : _mat(mat) {}

    TMatrix &       matrix ()       { return _mat; }
    const TMatrix & matrix () const { return _mat; }
    
    virtual uint type () const { return TYPE_TRANSFORM; }
    virtual bool is_type  ( uint t ) const
    { return ((t == TYPE_TRANSFORM) || TNodeData::is_type(t)); }
    
    virtual TNodeData * copy () const { return new TNDTransform(_mat); }
};

///////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////
//
// local node-types
//
///////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////

//
// translation
//
class TMoveNode : public TParser::TNode
{
protected:
    TParser::TNode * _vec;
    
public:
    TMoveNode ( TParser::TNode * vec ) : _vec(vec)
    {}

    virtual uint type () const
    {
        if (( _vec == NULL ) || _vec->is_type( TYPE_VECTOR ))
            return TYPE_TRANSFORM;
        else
            return TYPE_UNDEF;
    }

    TParser::TNodeData * eval () const
    {
        if ((_vec != NULL) && ! _vec->is_type( TYPE_VECTOR ))
        {
            parse_error( "wrong argument, expected vector", _vec->pos() );
            return NULL;
        }// if

        TNDTransform * res = new TNDTransform;

        if ( _vec != NULL )
        {
            TParser::TNodeData * dat = _vec->eval();

            if ( dat == NULL )
            {
                parse_error( "got NULL data in eval", _vec->pos() );
                delete dat;
                delete res;
                return NULL;
            }// if
            
            if ( dat->is_type( TYPE_VECTOR ) )
            {
                res->matrix()(0,3) = ((TNDVector*) dat)->vector()[0];
                res->matrix()(1,3) = ((TNDVector*) dat)->vector()[1];
                res->matrix()(2,3) = ((TNDVector*) dat)->vector()[2];
            }// if
            else
            {
                parse_error( "wrong data, expected vector", _vec->pos() );
                delete dat;
                delete res;
                return NULL;
            }// if

            delete dat;
        }// if

        return res;
    }
};

//
// scaling
//
class TScaleNode : public TParser::TNode
{
protected:
    TParser::TNode * _scale;
    
public:
    TScaleNode ( TParser::TNode * scale ) : _scale(scale)
    {}

    virtual uint type () const
    {
        if (( _scale == NULL ) || _scale->is_type( TYPE_NUMBER ) || _scale->is_type( TYPE_VECTOR ))
            return TYPE_TRANSFORM;
        else
            return TYPE_UNDEF;
    }

    TParser::TNodeData * eval () const
    {
        if ((_scale != NULL) &&
            ! (_scale->is_type( TYPE_NUMBER ) || _scale->is_type( TYPE_VECTOR )))
        {
            parse_error( "wrong argument, expected number or vector", _scale->pos() );
            return NULL;
        }// if

        TNDTransform * res = new TNDTransform;

        if ( _scale != NULL )
        {
            TParser::TNodeData * dat = _scale->eval();

            if ( dat == NULL )
            {
                parse_error( "got NULL data in eval", _scale->pos() );
                delete dat;
                delete res;
                return NULL;
            }// if
            
            if ( dat->is_type( TYPE_NUMBER ) )
            {
                res->matrix()(0,0) = ((TNDNumber*) dat)->number();
                res->matrix()(1,1) = ((TNDNumber*) dat)->number();
                res->matrix()(2,2) = ((TNDNumber*) dat)->number();
            }// if
            else if ( dat->is_type( TYPE_VECTOR ) )
            {
                res->matrix()(0,0) = ((TNDVector*) dat)->vector()[0];
                res->matrix()(1,1) = ((TNDVector*) dat)->vector()[1];
                res->matrix()(2,2) = ((TNDVector*) dat)->vector()[2];
            }// if
            else
            {
                parse_error( "wrong data, expected number or vector", _scale->pos() );
                delete dat;
                delete res;
                return NULL;
            }// if

            delete dat;
        }// if

        return res;
    }
};

//
// rotation
//
class TRotateNode : public TParser::TNode
{
protected:
    TParser::TNode * _axis;
    TParser::TNode * _angle;
    
public:
    TRotateNode ( TParser::TNode * axis, TParser::TNode * angle )
            : _axis(axis), _angle(angle)
    {}

    virtual uint type () const
    {
        if ((_axis != NULL) && (_angle != NULL) &&
            _axis->is_type( TYPE_VECTOR ) && _angle->is_type( TYPE_NUMBER ))
            return TYPE_TRANSFORM;
        else
            return TYPE_UNDEF;
    }

    TParser::TNodeData * eval () const
    {
        if ( _axis == NULL )
        {
            parse_error( "NULL rotation-axis given", pos() );
            return NULL;
        }// if

        if ( _angle == NULL )
        {
            parse_error( "NULL rotation-angle given", pos() );
            return NULL;
        }// if

        TParser::TNodeData * dat0 = _axis->eval();

        if ( dat0 == NULL )
        {
            parse_error( "got NULL result in eval", _axis->pos() );
            return NULL;
        }// if

        if ( ! dat0->is_type( TYPE_VECTOR ) )
        {
            parse_error( "wrong type in first argument, expected vector",
                         _axis->pos() );
            delete dat0;
            return NULL;
        }// if
        
        TParser::TNodeData * dat1 = _angle->eval();

        if ( dat1 == NULL )
        {
            parse_error( "got NULL result in eval", _angle->pos() );
            delete dat0;
            return NULL;
        }// if

        if ( ! dat1->is_type( TYPE_NUMBER ) )
        {
            parse_error( "wrong type in second argument, expected number",
                         _angle->pos() );
            delete dat0;
            delete dat1;
            return NULL;
        }// if

        TNDTransform  * res = new TNDTransform;
        TVector         u( ((TNDVector*) dat0)->vector() );
        real            angle = ((TNDNumber*) dat1)->number();
        
        real  c = std::cos( angle );
        real  s = std::sin( angle );
        real  xx = u.x() * u.x();
        real  yy = u.y() * u.y();
        real  zz = u.z() * u.z();
        real  emcc = 1.0 - c;

        res->matrix()(0,0) = xx + c * ( 1.0 - xx );
        res->matrix()(0,1) = u.x() * u.y() * emcc - u.z() * s;
        res->matrix()(0,2) = u.z() * u.x() * emcc + u.y() * s;
        
        res->matrix()(1,0) = u.x() * u.y() * emcc + u.z() * s;
        res->matrix()(1,1) = yy + c * ( 1.0 - yy );
        res->matrix()(1,2) = u.y() * u.z() * emcc - u.x() * s;
        
        res->matrix()(2,0) = u.z() * u.x() * emcc - u.y() * s;
        res->matrix()(2,1) = u.y() * u.z() * emcc + u.x() * s;
        res->matrix()(2,2) = zz + c * ( 1.0 - zz );

        delete dat0;
        delete dat1;

        return res;
    }
};

//
// composite transformation
//
class TTransformNode : public TParser::TNode
{
protected:
    // a transformation is a list of sub-transforms (move,scale,rotate)
    TSLL< TParser::TNode * >  _trans_list;

public:
    TTransformNode ()
    {}

    ~TTransformNode ()
    {
        while ( _trans_list.size() > 0 )
            delete _trans_list.behead();
    }
    
    virtual uint type () const { return TYPE_TRANSFORM; }

    // append transformation-node to list
    virtual void append ( TParser::TNode * n )
    {
        if ( n == NULL )
        {
            parse_warn( "appending NULL node to transformation-list",
                        TParser::TTextPos( text_row, text_col ) );
            return;
        }// if

        _trans_list.append( n );
    }
    
    TParser::TNodeData * eval () const
    {
        TNDTransform  * res = new TNDTransform;

        TSLL< TParser::TNode * >::TIterator  iter = _trans_list.first();

        while ( ! iter.eol() )
        {
            if ( ! iter()->is_type( TYPE_TRANSFORM ) )
            {
                parse_error( "have non-transformation-node in list", iter()->pos() );
                delete res;
                return NULL;
            }// if
            
            TParser::TNodeData * dat = iter()->eval();

            if ( dat == NULL )
            {
                parse_warn( "got NULL data in eval", iter()->pos() );
                ++iter;
                continue;
            }// if

            if ( ! dat->is_type( TYPE_TRANSFORM ) )
            {
                parse_warn( "got non-transformation-data in eval", iter()->pos() );
                ++iter;
                continue;
            }// if

            res->matrix().mul_left( ((TNDTransform*) dat)->matrix() );

            delete dat;
            ++iter;
        }// while

        return res;
    }
};

///////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////
//
// return local node types to outer world
//
///////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////

TParser::TNode * new_move ( TParser::TNode * n )
{
    return new TMoveNode( n );
}

TParser::TNode * new_scale ( TParser::TNode * n )
{
    return new TScaleNode( n );
}

TParser::TNode * new_rotate ( TParser::TNode * n0, TParser::TNode * n1 )
{
    return new TRotateNode( n0, n1 );
}

TParser::TNode * new_transform ()
{
    return new TTransformNode();
}

void add_transform ( TParser::TNode * n0, TParser::TNode * n1 )
{
    assert( n0 != NULL );
    
    TTransformNode * trans = dynamic_cast< TTransformNode * >( n0 );

    if ( trans != NULL )
        trans->append( n1 );
    else
        parse_error( "no transformation to add to", n0->pos() );
}

void apply_trans ( TGraphObject * obj, TParser::TNodeData * nd )
{
    TNDTransform * trans = dynamic_cast< TNDTransform * >( nd );

    if ( nd != NULL )
        obj->transform( trans->matrix() );
    else
        parse_error( "no transformation to apply" );
}

}// namespace
