//
// Project : lib3D
// File    : TImageIO.cc
// Purpose : classes for image file-io
// Author  : Ronald Kriemann
//
// arch-tag: bb167582-f20e-4f88-ae13-73de3858a22c
//

#include <stdio.h>

#include <cmath>

#include "ufo.hh"

#include "TImageIO.hh"

#ifdef USE_PNG
#include <png.h>
#endif

#ifdef USE_JPEG
#include <jpeglib.h>
#endif

namespace RGL
{

//////////////////////////////////////////////////////////////
//
// local functions
//
//////////////////////////////////////////////////////////////

//
// corrects the colour by given gamma value
//
static inline real
gamma_correct ( real f, real gamma )
{
    return std::pow( f, real(1.0) / gamma );
}

//////////////////////////////////////////////////////////////
//
// TImgIOHandler
//
// handler for transparent image loading/saving
//
//////////////////////////////////////////////////////////////

//
// basic functionality: load and save
//

int
TImgIOHandler::load ( const TString & fname, TImage & img ) const
{
    //
    // determine type of image by looking at suffix
    //

    TString   suffix;
    int       len   = fname.length();
    int       pos   = len-1;
    bool      found = false;
    
    while ( pos >= 0 )
    {
        if ( fname[pos] == '.' )
        {
            suffix = fname.index( pos + 1 );
            found = true;
            break;
        }// if

        --pos;
    }// while

    suffix.to_lower();
    
    //
    // compare suffix with known extensions
    //
    
    if ( suffix == "png" )
    {
        TPNGImgIO  io( gamma() );

        return io.load( fname, img );
    }

    if (( suffix == "jpeg" ) || ( suffix == "jpg" ))
    {
        TJPEGImgIO  io( gamma() );

        return io.load( fname, img );
    }

    if ( suffix == "pnm" )
    {
        TPNMImgIO  io( gamma() );

        return io.load( fname, img );
    }

    std::cout << "(TImgIOHandler) load : unsupported file type \""
              << suffix << "\"" << std::endl;
    
    return -1;
}

int
TImgIOHandler::save ( const TString & fname, const TImage & img ) const
{
    //
    // determine type of image by looking at suffix
    //

    TString   suffix;
    int       len   = fname.length();
    int       pos   = len-1;
    bool      found = false;
    
    while ( pos >= 0 )
    {
        if ( fname[pos] == '.' )
        {
            suffix = fname.index( pos + 1 );
            found = true;
            break;
        }// if

        --pos;
    }// while

    suffix.to_lower();
    
    //
    // compare suffix with known extensions
    //
    
    if ( suffix == "png" )
    {
        TPNGImgIO  io( gamma() );

        return io.save( fname, img );
    }

    if (( suffix == "jpeg" ) || ( suffix == "jpg" ))
    {
        TJPEGImgIO  io( gamma() );

        return io.save( fname, img );
    }

    if ( suffix == "pnm" )
    {
        TPNMImgIO  io( gamma() );

        return io.save( fname, img );
    }

    std::cout << "(TImgIOHandler) load : unsupported file type "
              << suffix << "\"" << std::endl;
    
    return -1;
}

/////////////////////////////////////////////////////////
//
// class for PNG-files
//
/////////////////////////////////////////////////////////

int
TPNGImgIO::load ( const TString & fname, TImage & img ) const
{
#ifdef USE_PNG
    FILE  * file;

    if (( file = fopen( fname.c_str(), "r" ) ) == NULL)
    {
        std::cout << "(TPNGImgIO) load : could not open file (" << fname << ")" << std::endl;
        return -1;
    }// if

    //
    // check if file is a png-file
    //

#define BYTES_TO_CHECK  4
    
    uchar  header[ BYTES_TO_CHECK ];

    // read first BYTES_TO_CHECK characters
    if ( fread( header, 1, BYTES_TO_CHECK, file ) < BYTES_TO_CHECK )
    {
        // no BYTES_TO_CHECK char => no png
        std::cout << "(TPNGImgIO) load : file too short for a PNG-file" << std::endl;
        fclose( file );
        return -1;
    }// if

    if ( png_sig_cmp( header, 0, BYTES_TO_CHECK ) != 0 )
    {
        std::cout << "(TPNGImgIO) load : wrong header for a PNG-file" << std::endl;
        fclose( file );
        return -1;
    }// if 

    //
    // init png-structs
    //

    png_structp  png_ptr;
    png_infop    info_ptr;
    png_infop    end_ptr;

    png_ptr = png_create_read_struct( PNG_LIBPNG_VER_STRING, NULL, NULL, NULL );
    if ( ! png_ptr )
    {
        std::cout << "(TPNGImgIO) load : could not allocate png_struct" << std::endl;
        fclose( file );
        return -1;
    }// if

    if ((info_ptr = png_create_info_struct( png_ptr )) == NULL)
    {
        png_destroy_read_struct( & png_ptr, NULL, NULL);
        fclose( file );
        return -1;
    }// if

    if ((end_ptr = png_create_info_struct( png_ptr )) == NULL )
    {
        png_destroy_read_struct( &png_ptr, &info_ptr, NULL );
        fclose( file );
        return -1;
    }// if

    if ( setjmp(png_ptr->jmpbuf) )
    {
        png_destroy_read_struct( & png_ptr, & info_ptr, & end_ptr );
        fclose(file);
        return -1;
    }// if

    //
    // actually read the file
    //

    png_uint_32  width, height;
    int          depth, col_type, interlace;

    png_init_io( png_ptr, file );

    png_set_sig_bytes( png_ptr, BYTES_TO_CHECK );
    
    // get info about file
    png_read_info( png_ptr, info_ptr );
    png_get_IHDR( png_ptr, info_ptr,
                  & width, & height, & depth,
                  & col_type, & interlace, NULL, NULL );

    // adjust depth with number of channels
    depth *= png_get_channels( png_ptr, info_ptr );
    
    // adjust own definition
    img.set_geometry( width, height );
        
    // alloc data to store image
    int          bpl = png_get_rowbytes( png_ptr, info_ptr );
    png_bytep  * row_data = new png_bytep[ height ];

    for ( uint j = 0; j < height; j++ )
        row_data[j] = new png_byte[ bpl ];

    png_read_image( png_ptr, row_data );

    for ( uint j = 0; j < height; j++ )
    {
        for ( uint i = 0; i < width; i++ )
        {
            TColour  col;

            if ( depth == 32 )
            {
                col.set_rgb32( (uint) row_data[i][j] );
            }// if

            // memcpy( & img.data()[ addr ], row_data[i], bpl );
            // addr += img.bytes_per_line();
        }// for
    }// for
    
    //
    // finish
    //

    for ( uint i = 0; i < height; i++ )
        delete[] row_data[i];

    delete[] row_data;
    
    png_destroy_read_struct( & png_ptr, & info_ptr, & end_ptr ) ;

    fclose( file );
#else
    std::cout << "(TPNGImgIO) load : PNG not supported in this configuration" << std::endl;
#endif

    return 0;
}

int
TPNGImgIO::save ( const TString & fname, const TImage & img ) const
{
#ifdef USE_PNG
    FILE        * file;
    png_structp   png_ptr;
    png_infop     info_ptr;

    // open file
    if ((file = fopen( fname.c_str(), "w" )) == NULL )
    {
        std::cout << "(TPNGImgIO) save : could not open file ("
                  << fname << ")" << std::endl;
        return -1;
    }// if

    // init. png-struct without error-handler
    png_ptr = png_create_write_struct( PNG_LIBPNG_VER_STRING, NULL, NULL, NULL );
    if ( png_ptr == NULL )
    {
        std::cout << "(TPNGImgIO) save : error in png_create_write_struct" << std::endl;
        fclose( file );
        return -1;
    }// if

    // initialize image information data
    info_ptr = png_create_info_struct( png_ptr );
    if ( info_ptr == NULL )
    {
        std::cout << "(TPNGImgIO) save : error in png_create_info_struct" << std::endl;
        png_destroy_write_struct( & png_ptr, (png_infopp) NULL );
        fclose( file );
        return -1;
    }// if
   
    // setup error-return-handling
    if ( setjmp( png_ptr->jmpbuf ) )
    {
        std::cout << "(TPNGImgIO) save : error in setjmp" << std::endl;
        png_destroy_write_struct( & png_ptr, (png_infopp) NULL );
        fclose( file );
        return -1;
    }// if
        
    // setup default C-streams
    png_init_io( png_ptr, file );

    //
    // setup the image info data
    //

    uint   png_depth = 16;
    uint   factor    = (2 << (png_depth-1)) - 1;
    double gammacor  = gamma();

    png_set_IHDR( png_ptr, info_ptr,
                  img.width(), img.height(), png_depth,
                  PNG_COLOR_TYPE_RGB,
                  PNG_INTERLACE_NONE,
                  PNG_COMPRESSION_TYPE_BASE,
                  PNG_FILTER_TYPE_BASE );

    // set bit-depth for each channel
    png_color_8  sig_bit;

    sig_bit.red   = png_depth;
    sig_bit.green = png_depth;
    sig_bit.blue  = png_depth;

#if defined(PNG_gAMA_SUPPORTED) && defined(PNG_FLOATING_POINT_SUPPORTED)
    png_set_gAMA( png_ptr, info_ptr, gammacor );
    // reset, such that values are not changed later
//   gammacor = 1.0;
#endif
   
#if defined(PNG_sBIT_SUPPORTED)   
    png_set_sBIT( png_ptr, info_ptr, & sig_bit );
#endif

    // write file header
    png_write_info( png_ptr, info_ptr );

    // setup image-data for writing
    uint       png_stride = png_depth * 3 / 8;
    png_byte * png_row    = new png_byte[ img.width() * png_stride ];
   
    for ( uint j = 0; j < img.height(); j++ )
    {
        for ( uint i = 0, pos = 0; i < img.width(); i++, pos += png_stride )
        {
            TColour col( img.get_colour( i, j ) );
            ulong   r = ulong( gamma_correct( col.red(),   gammacor ) * real(factor) );
            ulong   g = ulong( gamma_correct( col.green(), gammacor ) * real(factor) );
            ulong   b = ulong( gamma_correct( col.blue(),  gammacor ) * real(factor) );

            if ( r > factor ) r = factor;
            if ( g > factor ) g = factor;
            if ( b > factor ) b = factor;
               
            if ( png_depth == 16 )
            {
                png_row[pos]   = png_byte(r >> 8);
                png_row[pos+1] = png_byte(r & 0xff);
               
                png_row[pos+2] = png_byte(g >> 8);
                png_row[pos+3] = png_byte(g & 0xff);
               
                png_row[pos+4] = png_byte(b >> 8);
                png_row[pos+5] = png_byte(b & 0xff);
            }// if
            else if ( png_depth == 8 )
            {
                png_row[pos]   = png_byte(r);
                png_row[pos+1] = png_byte(g);
                png_row[pos+2] = png_byte(b);
            }// else
        }// for

        // write single line
        png_write_row( png_ptr, png_row );
    }// for

    // finish writing
    png_write_end( png_ptr, info_ptr );

    delete[] png_row;

    // clean up
    png_destroy_write_struct( & png_ptr, & info_ptr );
    fclose( file );
#endif
   
    return 0;
}

/////////////////////////////////////////////////////////
//
// class for JPEG-files
//
/////////////////////////////////////////////////////////

int
TJPEGImgIO::load ( const TString & fname, TImage & img ) const
{
#ifdef USE_JPEG
    struct jpeg_decompress_struct   cinfo;
    struct jpeg_error_mgr           jerr;
    FILE                          * file;
    JSAMPROW                        row_buf;
    int                             row_stride;

    if ((file = fopen( fname.c_str(), "r" )) == NULL)
    {
        std::cout << "(TJPEGImgIO) load : could not open file (" << fname << ")" << std::endl;
        return -1;
    }// if

    // init. jpeg-structs
    cinfo.err = jpeg_std_error( & jerr );
    
    jpeg_create_decompress( & cinfo );

    // setup datasource
    jpeg_stdio_src( & cinfo, file );

    //
    // read header of jpeg-file
    // and apply characteristics to local image
    //
    jpeg_read_header( & cinfo, TRUE );

    uint  width, height, depth = 1;
    bool  grayscale = false;
    
    width  = cinfo.image_width;
    height = cinfo.image_height;
    
    if ( cinfo.out_color_space == JCS_RGB )
    {
        grayscale = false;
        
        if ( cinfo.num_components == 3 )
            depth = 24;
        else
            std::cout << "(TJPEGImgIO) load : unsupported number of components" << std::endl;
    }// if
    else if ( cinfo.out_color_space == JCS_GRAYSCALE )
    {
        grayscale = true;
        
        if ( cinfo.num_components == 1 )
            depth = 8;
        else
            std::cout << "(TJPEGImgIO) load : unsupported number of components" << std::endl;
    }// if
    else
    {
        std::cout << "(TJPEGImgIO) load : unsupported colorspace" << std::endl;
    }// else

    // allocate image-memory
    img.set_geometry( width, height );
    
    // start decompressing loop
    jpeg_start_decompress( & cinfo );

    // make a one-row-high sample array that will go away when done with image
    row_stride = cinfo.output_width * cinfo.output_components;
    row_buf = new JSAMPLE[ row_stride ];
    // (*cinfo.mem->alloc_sarray)( (j_common_ptr) & cinfo, JPOOL_IMAGE, row_stride, 1 );

    //
    // read image and copy into local memory
    //
    
    int  i = 0;
    
    while (cinfo.output_scanline < cinfo.output_height)
    {
        jpeg_read_scanlines( & cinfo, & row_buf, 1 );
        // memcpy( & img.data()[ i * row_stride ], & row_buf[0], row_stride );

        i++;
    }// while

    // finish process
    jpeg_finish_decompress( & cinfo );

    // remove jpeg-structures
    jpeg_destroy_decompress( & cinfo );

    delete row_buf;
    
    // finish
    fclose( file );
#else
    std::cout << "(TJPEGImgIO) load : JPEG not supported in this configuration" << std::endl;
#endif

    return 0;
}

int
TJPEGImgIO::save ( const TString & fname, const TImage & img ) const
{
#ifdef USE_JPEG
    struct jpeg_compress_struct  cinfo;
    struct jpeg_error_mgr        jerr;
    FILE                       * file;
    JSAMPROW                     row_pointer[1];
    int                          row_stride;

    //
    // open file for writing
    //
    
    if ((file = fopen( fname.c_str(), "w" )) == NULL)
    {
        std::cout << "(TJPEGImgIO) save : could not open file (" << fname << ")" << std::endl;
        return -1;
    }// if

    //
    // init jpeg-data
    //
    
    cinfo.err = jpeg_std_error( & jerr );

    jpeg_create_compress( & cinfo );
    jpeg_stdio_dest( & cinfo, file );

    cinfo.image_width  = img.width();
    cinfo.image_height = img.height();

    cinfo.input_components = 3;
    cinfo.in_color_space   = JCS_RGB;

    //
    // set parameters for compression
    //

    jpeg_set_defaults( & cinfo );
    jpeg_set_quality( & cinfo, 100, TRUE );
    
    //
    // compress
    //

    char * row_data;
    real   gammacor = gamma();
    
    jpeg_start_compress( & cinfo, TRUE );

    row_stride = img.width() * 3;
    row_data   = new char[ row_stride ];
    
    while ( cinfo.next_scanline < cinfo.image_height )
    {
        // assign data to write
        for ( uint i = 0; i < img.width(); i++ )
        {
            TColour  col( img.get_colour( i, cinfo.next_scanline ) );
            ulong    r = ulong( floor( gamma_correct( col.red(),   gammacor ) * 255.0));
            ulong    g = ulong( floor( gamma_correct( col.green(), gammacor ) * 255.0));
            ulong    b = ulong( floor( gamma_correct( col.blue(),  gammacor ) * 255.0));
            
            if ( r > 255 ) r = 255;
            if ( g > 255 ) g = 255;
            if ( b > 255 ) b = 255;
               
            row_data[3*i]   = (uchar) r;
            row_data[3*i+1] = (uchar) g;
            row_data[3*i+2] = (uchar) b;
        }// for
        
        row_pointer[0] = (JSAMPROW) row_data;
            
        jpeg_write_scanlines( & cinfo, row_pointer, 1 );
    }// while

    //
    // finish
    //

    jpeg_finish_compress( & cinfo );

    fclose( file );

    jpeg_destroy_compress( & cinfo );
#endif
    
    return 0;
}

/////////////////////////////////////////////////////////
//
// class for PNM-files
//
/////////////////////////////////////////////////////////

int
TPNMImgIO::load ( const TString & fname, TImage & img ) const
{
    return 0;
}

int
TPNMImgIO::save ( const TString & fname, const TImage & img ) const
{
    TStdFile  file;

    if ( file.open( fname, "w" ) != 0 )
        return -1;
    
    return save_ppm( file, img );

    return 0;
}

//
// special functions for individual formats
//

int
TPNMImgIO::save_ppm ( TFile & file, const TImage & img ) const
{
    uint  w        = img.width();
    uint  h        = img.height();
    uint  max_comp = 0;
    real  gammacor = gamma();

    //
    // determine maximal colour component in image
    //

    for ( uint i = 0; i < w; i++ )
    {
        for ( uint j = 0; j < h; j++ )
        {
            TColour  col( img.get_colour( i, j ) );
            uint     r = uint( gamma_correct( col.red(),   gammacor ) * 255.0 );
            uint     g = uint( gamma_correct( col.green(), gammacor ) * 255.0 );
            uint     b = uint( gamma_correct( col.blue(),  gammacor ) * 255.0 );

            if ( r > 255 ) r = 255;
            if ( g > 255 ) g = 255;
            if ( b > 255 ) b = 255;
               
            max_comp = max( max_comp, r );
            max_comp = max( max_comp, g );
            max_comp = max( max_comp, b );
        }// for
    }// for

    //
    // write image
    //

    file.puts( "P3\n" );
    file.puts( to_str( "%d\n", w ) + to_str( "%d\n", h ) );
    file.puts( to_str( "%d\n", max_comp ) );
    
    for ( uint j = 0; j < h; j++ )
    {
        for ( uint i = 0; i < w; i++ )
        {
            TColour  col( img.get_colour( i, j ) );
            uint     r = uint( gamma_correct( col.red(),   gammacor ) * 255.0 );
            uint     g = uint( gamma_correct( col.green(), gammacor ) * 255.0 );
            uint     b = uint( gamma_correct( col.blue(),  gammacor ) * 255.0 );

            if ( r > 255 ) r = 255;
            if ( g > 255 ) g = 255;
            if ( b > 255 ) b = 255;
               
            file.puts( to_str( "%d ", r ) + to_str( "%d ", g ) + to_str( "%d\n", b ) );
        }// for
    }// for

    return 0;
}

}// namespace
