//
// Project : PHI
// File    : TXImage.cc
// Purpose : class for a X11-based printer
// Author  : Ronald Kriemann
//
// arch-tag: 7c40495e-09f1-4c09-bf46-7747d2bf5830
//

#include <stdlib.h>
#include <sys/types.h>
#include <sys/select.h>

#include <cmath>

#ifdef USE_X11
#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#endif

#include "ufo.hh"
#include "defines.hh"
#include "TSLL.hh"

#include "TXImage.hh"

namespace RGL
{

//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//
// local defines
//
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////

#define RED_COL    0
#define GREEN_COL  1
#define BLUE_COL   2

#define CHECK_ACTIVE   { if ( ! _active ) return; }
#define HANDLE_EVENTS
//#define HANDLE_EVENTS  { handle_events(); }

//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//
// local types
//
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////

//
// X11 related data
//
class TX11Data
{
#ifdef USE_X11
public:
    Display      * display;
    int            screen;
    Window         window;
    Pixmap         pixmap;
    int            depth;
    int            vis_class;
    GC             gc_win_pen, gc_win_brush;
    GC             gc_pix_pen, gc_pix_brush;
    XFontStruct  * font;
    Colormap       cmap;

    TX11Data ()
    {
        display = NULL;
        font    = NULL;
    }
#endif
};

//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//
// local data
//
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////

#ifdef USE_X11

// names of visual classes
static TString visual_class[] = { "StaticGray" , "GrayScale", "StaticColor",
                                  "PseudoColor", "TrueColor", "DirectColor" };

// conversion of font names
static const uint   n_fonts      = 4;
static const char * font_table[] = {
    "Times",     "-adobe-times-medium-r-normal--*",
    "Courier",   "-adobe-courier-medium-r-normal--*",
    "Helvetica", "-adobe-helvetica-medium-r-normal--*",
    "Symbol",    "-adobe-symbol-medium-r-normal--*"
};

#endif

//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//
// local functions
//
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////

#ifdef USE_X11

//
// return position of highest bit
//
static int
highest_bit ( uint v )
{
    int   i = -1;

    while ( v > 0 )
    {
        v >>= 1;
        i++;
    }// for

    return i;
}

//
// corrects the colour by given gamma value
//
static inline real
gamma_correct ( real f, real gamma )
{
    return std::pow( f, real(1.0) / gamma );
}

#endif

//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//
// TXImage
//
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////
//
// constructor and destructor
//

TXImage::TXImage ( int w, int h, real gamma )
        : TImage( max( 1, w ), max( 1, h ) )
{
    _x11data    = NULL;
    
    _active     = false;
    _finish     = false;
    _draw_win   = true;

    _old_pos[0] = _old_pos[1] = 0;
    
    _old_type   = (uint) -1;
    _old_size   = -1;

    _gamma      = gamma;
    
    init();
}

TXImage::~TXImage ()
{
    if ( _active )
        end();
}

void
TXImage::init ()
{
    if ( _active )
    {
        std::cout << "(TXImage) init : already running window" << std::endl;
        return;
    }// if

#ifdef USE_X11
    _x11data = new TX11Data;
    
    //
    // connect to display
    //
    
    if (( _x11data->display = XOpenDisplay( NULL )) == NULL )
    {
        std::cout << "(TXImage) init : could not connect to display ("
                  << XDisplayName( NULL ) << ")" << std::endl;
        return;
    }// if

    //
    // get default screen number
    //
    
    _x11data->screen = DefaultScreen( _x11data->display );

    //
    // get characteristics of display
    //
    
    _x11data->depth  = DefaultDepth( _x11data->display, _x11data->screen );

    //
    // determine available visual classes
    //
    
    XVisualInfo  vis_info;
    bool         found_vis = false;

    _x11data->vis_class = TrueColor;

    while ( _x11data->vis_class >= 0 )
    {
        if ( XMatchVisualInfo( _x11data->display, _x11data->screen,
                               _x11data->depth, _x11data->vis_class, & vis_info ) != 0 )
        {
            found_vis = true;
            break;
        }// if
        
        _x11data->vis_class--;
    }// while

    if ( ! found_vis )
    {
        std::cout << "(TXImage) init : no suitable visual class found" << std::endl;
        return;
    }// if
    
    if ( verbose(5) )
        std::cout << "(TXImage) init : found " << visual_class[_x11data->vis_class]
                  << " visual class" << std::endl;
    
    //
    // init colors for truecolor visual
    //
    
    if ( _x11data->vis_class == TrueColor )
    {
        _truecolour = true;
        
        _col_mask[ RED_COL ]    = (uint) vis_info.red_mask;
        _col_mask[ GREEN_COL ]  = (uint) vis_info.green_mask;
        _col_mask[ BLUE_COL ]   = (uint) vis_info.blue_mask;
        _col_shift[ RED_COL ]   = highest_bit( _col_mask[ RED_COL ] )   - 7;
        _col_shift[ GREEN_COL ] = highest_bit( _col_mask[ GREEN_COL ] ) - 7;
        _col_shift[ BLUE_COL ]  = highest_bit( _col_mask[ BLUE_COL ] )  - 7;
    }// if
    else
    {
        _truecolour = false;
        _x11data->cmap = DefaultColormap( _x11data->display, _x11data->screen );
    }// else

    //
    // setup font handling
    //

    setup_fonts();
    
    //
    // create window
    //

    XSetWindowAttributes   attr;
    ulong                  valuemask;
    
    attr.background_pixel  = WhitePixel( _x11data->display, _x11data->screen );
    attr.border_pixel      = BlackPixel( _x11data->display, _x11data->screen );
    attr.event_mask        = ExposureMask | FocusChangeMask |
                             KeyPressMask | ButtonPressMask |
                             StructureNotifyMask;
    attr.override_redirect = false;

    valuemask = CWBackPixel | CWBorderPixel | CWEventMask | CWOverrideRedirect;

    _x11data->window = XCreateWindow( _x11data->display,
                                      RootWindow(_x11data->display, _x11data->screen),
                                      0, 0, _width, _height,
                                      5, CopyFromParent, InputOutput, CopyFromParent,
                                      valuemask, & attr );

    //
    // set properties of window
    //
    
    XWMHints      * wm_hints;
    XClassHint    * class_hint;
    XSizeHints    * size_hints;
    bool            use_hints = true;

    // allocate memory for every hint
    if (((size_hints  = XAllocSizeHints()) == NULL) ||
        ((wm_hints    = XAllocWMHints())   == NULL) ||
        ((class_hint  = XAllocClassHint()) == NULL))
    {
        std::cout << "(TXImage) init : could not allocate hints" << std::endl;
        use_hints = false;
    }// if

    if ( use_hints )
    {
        XTextProperty   window_name;
        bool            use_win_name = true;
        char            text[]   = "libPHI";
        char          * win_name = text;
        
        //
        // set hints
        //

        size_hints->flags      = PPosition | PSize | PMinSize | PMaxSize;
        size_hints->min_width  = _width;
        size_hints->min_height = _height;
        size_hints->max_width  = _width;
        size_hints->max_height = _height;
        
        if ( XStringListToTextProperty( & win_name, 1, & window_name ) == 0 )
        {
            std::cout << "(TX11Window) init : could not allocate windowname" << std::endl;
            use_win_name = false;
        }// if
        
        wm_hints->initial_state = NormalState;
        wm_hints->input         = True;
        wm_hints->flags         = StateHint | InputHint;
        
        class_hint->res_name  = win_name;
        class_hint->res_class = win_name;
        
        //
        // apply hints
        //
        
        XSetWMProperties( _x11data->display, _x11data->window, & window_name, NULL, NULL, 0,
                          size_hints, wm_hints, class_hint );
    }// if
    
    XStoreName( _x11data->display, _x11data->window, "prender" );

    XMapRaised( _x11data->display, _x11data->window );
    XFlush( _x11data->display );

    //
    // create pixmap for intermediate drawing
    //

    _x11data->pixmap = XCreatePixmap( _x11data->display,
                                      DefaultRootWindow( _x11data->display ),
                                      _width, _height, _x11data->depth );
    
    //
    // create graphics context (fg,bg,fonts)
    //

    XGCValues  gc_values;

    // for the pen
    gc_values.foreground = BlackPixel( _x11data->display, _x11data->screen );
    gc_values.background = WhitePixel( _x11data->display, _x11data->screen );
    
    _x11data->gc_pix_pen = XCreateGC( _x11data->display, _x11data->pixmap,
                                      GCForeground | GCBackground, & gc_values );
    XSetLineAttributes( _x11data->display, _x11data->gc_pix_pen,
                        1, LineSolid, CapButt, JoinRound );
    
    _x11data->gc_win_pen = XCreateGC( _x11data->display, _x11data->window,
                                      GCForeground | GCBackground, & gc_values );
    XSetLineAttributes( _x11data->display, _x11data->gc_win_pen,
                        1, LineSolid, CapButt, JoinRound );

    // for the brush
    gc_values.foreground = BlackPixel( _x11data->display, _x11data->screen );
    gc_values.background = WhitePixel( _x11data->display, _x11data->screen );
    gc_values.fill_style = FillSolid;
    _x11data->gc_pix_brush = XCreateGC( _x11data->display, _x11data->pixmap,
                                        GCForeground | GCBackground | GCFillStyle, & gc_values );
    XSetLineAttributes( _x11data->display, _x11data->gc_pix_brush,
                        0, LineSolid, CapButt, JoinRound );
    
    _x11data->gc_win_brush = XCreateGC( _x11data->display, _x11data->window,
                                        GCForeground | GCBackground | GCFillStyle, & gc_values );
    XSetLineAttributes( _x11data->display, _x11data->gc_win_brush,
                        0, LineSolid, CapButt, JoinRound );

    //
    // mark as active, running window
    //
    
    _active = true;
    _finish = false;

    //
    // setup additional details
    //
    
    set_font( "Courier", 12 );
    set_line_width( 1 );

    // clear window
    XSetForeground( _x11data->display, _x11data->gc_pix_brush,
                    WhitePixel( _x11data->display, _x11data->screen ) );
    XSetForeground( _x11data->display, _x11data->gc_win_brush,
                    WhitePixel( _x11data->display, _x11data->screen ) );
    XFillRectangle( _x11data->display, _x11data->pixmap,
                    _x11data->gc_pix_brush, 0, 0, _width, _height );
    XFillRectangle( _x11data->display, _x11data->window,
                    _x11data->gc_win_brush, 0, 0, _width, _height );
    XSetForeground( _x11data->display, _x11data->gc_pix_brush,
                    BlackPixel( _x11data->display, _x11data->screen ) );
    XSetForeground( _x11data->display, _x11data->gc_win_brush,
                    BlackPixel( _x11data->display, _x11data->screen ) );

    //
    // fill all with black
    //

    fill( TColour( 0, 0, 0 ) );
    
#else
    std::cout << "(TXImage) init : no X11 support chosen during configuration" << std::endl;
#endif
}

//
// determine available fonts
//
void
TXImage::setup_fonts ()
{
#ifdef USE_X11
    //
    // for each local font, loop through available X11 fonts
    //

    // remove old table
    _fonts.set_size( 0 );
    _fonts.set_size( n_fonts );
    
    for ( uint i = 0; i < n_fonts; i++ )
    {
        char  ** fontnames;
        int      n;

        // look for a suitable font
        fontnames = XListFonts( _x11data->display, font_table[2*i+1], 100, & n );

        // return immediately if no fonts are available
        if ( fontnames == NULL )
            break;

        TSLL< TString >  fontlist;
        int              max_size = 0;
        
        // parse through fontlist and look for closests
        for ( int j = 0; j < n; j++ )
        {
            char * size_info = strstr( fontnames[j], "--" );

            if ( size_info != NULL )
            {
                int  s = atoi( size_info + 2 );

                if ( s != 0 )
                {
                    max_size = max( max_size, s );
                    fontlist.append( fontnames[j] );
                }// if
            }// if
        }// for

        XFreeFontNames( fontnames );

        _fonts[i].set_size( max_size + 1 );

        while ( fontlist.size() > 0 )
        {
            TString       name      = fontlist.behead();
            const char  * size_info = strstr( name.c_str(), "--" );

            if ( size_info != NULL )
            {
                int  s = atoi( size_info + 2 );

                if (( s != 0 ) && ( s <= max_size ) && ( _fonts[i][s] == "" ))
                    _fonts[i][s] = name;
            }// if
        }// while
    }// for

//     for ( uint i = 0; i < n_fonts; i++ )
//     {
//         for ( uint j = 0; j < _fonts[i].size(); j++ )
//             if ( _fonts[i][j] != "" )
//                 std::cout << j << " : " << _fonts[i][j] << std::endl;
//     }// for
#endif
}
    
//
// handle pending events
//
void
TXImage::handle_events ()
{
#ifdef USE_X11
    //
    // eat up all events for own window
    //
    
    XEvent          event;
    KeySym          keysym;
    int             length, max_bytes;
    XComposeStatus  composestatus;
    char            keyb_buf[20];
    
    while ( XPending( _x11data->display ) )
    {
        XNextEvent( _x11data->display, & event );

        switch( event.type )
        {
        case GraphicsExpose:
        case Expose :
            draw();
            break;
            
        case ConfigureNotify :
//             _width  = event.xconfigure.width;
//             _height = event.xconfigure.height;
            draw();
            break;
            
        case KeyPress :
            max_bytes = 1;
            length = XLookupString( & event.xkey, keyb_buf, max_bytes, & keysym,
                                    & composestatus );
            
            if ( keyb_buf[0] == 'q' )
                _finish = true;
        }// switch
    }// while
#endif
}

//
// internal drawing method
//
void
TXImage::draw ()
{
#ifdef USE_X11
    XCopyArea( _x11data->display, _x11data->pixmap, _x11data->window,
               _x11data->gc_win_pen, 0, 0, _width, _height, 0, 0 );
#endif
}

//////////////////////////////////////////////////
//
// methods for image drawing
//

//
// draw a point
//
void
TXImage::draw_point ( uint x, uint y, const TColour & c )
{
    TImage::draw_point( x, y, c );

#ifdef USE_X11
    //
    // draw point into X11 image
    //
    
    int  r = int( 255.0 * gamma_correct( c.red(),   _gamma ) );
    int  g = int( 255.0 * gamma_correct( c.green(), _gamma ) );
    int  b = int( 255.0 * gamma_correct( c.blue(),  _gamma ) );

    set_rgb( r, g, b );
    draw_point( x, y );
#endif
}

//
// fill image with colour
//
void
TXImage::fill ( const TColour & c )
{
    TImage::fill( c );

#ifdef USE_X11
    //
    // fill X11 image with colour
    //
    
    int  r = int( 255.0 * gamma_correct( c.red(),   _gamma ) );
    int  g = int( 255.0 * gamma_correct( c.green(), _gamma ) );
    int  b = int( 255.0 * gamma_correct( c.blue(),  _gamma ) );

    set_rgb( r, g, b );
    fill_rect( 0, 0, _width-1, _height-1 );
#endif
}

//
// copy given image at given position
//
void
TXImage::bitblt ( uint x, uint y, const TImage & img )
{
    TImage::bitblt( x, y, img );
    
#ifdef USE_X11
    //
    // copy image into X11 window
    //
    
    uint  w = img.width();
    uint  h = img.height();
    
    for ( uint j = 0; j < h; j++ )
        for ( uint i = 0; i < w; i++ )
        {
            TColour  c( img.get_colour( i, j ) );
            int      r = int( 255.0 * gamma_correct( c.red(),   _gamma ) );
            int      g = int( 255.0 * gamma_correct( c.green(), _gamma ) );
            int      b = int( 255.0 * gamma_correct( c.blue(),  _gamma ) );

            set_rgb( r, g, b );
            draw_point( i+x, j+y );
        }// for
#endif
}

//////////////////////////////////////////////////
//
// methods for X11 drawing
//

//
// begin and end drawing
//
void
TXImage::begin ()
{
    init();
}

void
TXImage::end ()
{
#ifdef USE_X11
    CHECK_ACTIVE;

    int               fd_X = XConnectionNumber( _x11data->display );
    fd_set            fds;
    struct timeval  * timeout;

    // no longer draw to window directly
    _draw_win = false;
    
    while ( true )
    {
        handle_events();

        // check request to finish
        if ( _finish )
            break;
        
        // wait for X11 activity
        FD_SET( fd_X, & fds );
        timeout = NULL;
        select( fd_X + 1, & fds, 0, 0, timeout );
    }// while
    
    //
    // free X11
    //

    if ( _x11data->font != NULL )
    {
        XFreeFont( _x11data->display, _x11data->font );
        _x11data->font = NULL;
    }// if
    
    XFreeGC( _x11data->display, _x11data->gc_pix_pen );
    XFreeGC( _x11data->display, _x11data->gc_win_pen );
    XFreeGC( _x11data->display, _x11data->gc_pix_brush );
    XFreeGC( _x11data->display, _x11data->gc_win_brush );
    XFreePixmap( _x11data->display, _x11data->pixmap );
    XDestroyWindow( _x11data->display, _x11data->window );
    XCloseDisplay( _x11data->display );

    _active = false;

    delete _x11data;
    _x11data = NULL;
#endif
}

//
// 2D - drawing
//

void
TXImage::draw_point ( int x, int y )
{
#ifdef USE_X11
    CHECK_ACTIVE;

    XDrawPoint( _x11data->display, _x11data->pixmap, _x11data->gc_pix_pen, x, y );

    if ( _draw_win )
        XDrawPoint( _x11data->display, _x11data->window, _x11data->gc_win_pen, x, y );
        
    HANDLE_EVENTS;
#endif
}

void
TXImage::draw_line ( int x1, int y1, int x2, int y2 )
{
#ifdef USE_X11
    CHECK_ACTIVE;

    XDrawLine( _x11data->display, _x11data->pixmap, _x11data->gc_pix_pen,
               x1, y1, x2, y2 );

    if ( _draw_win )
        XDrawLine( _x11data->display, _x11data->window, _x11data->gc_win_pen,
                   x1, y1, x2, y2 );
    
    HANDLE_EVENTS;
#endif
}


void
TXImage::draw_rect ( int x1, int y1, int x2, int y2 )
{
#ifdef USE_X11
    CHECK_ACTIVE;

    if ( x1 < x2 )
        if ( y1 < y2 )
        {
            XDrawRectangle( _x11data->display, _x11data->pixmap, _x11data->gc_pix_pen,
                            x1, y1, x2-x1, y2-y1 );
            if ( _draw_win )
                XDrawRectangle( _x11data->display, _x11data->window, _x11data->gc_win_pen,
                                x1, y1, x2-x1, y2-y1 );
        }// if
        else
        {
            XDrawRectangle( _x11data->display, _x11data->pixmap, _x11data->gc_pix_pen,
                            x1, y2, x2-x1, y1-y2 );
            if ( _draw_win )
                XDrawRectangle( _x11data->display, _x11data->window, _x11data->gc_win_pen,
                                x1, y2, x2-x1, y1-y2 );
        }// else
    else
        if ( y1 < y2 )
        {
            XDrawRectangle( _x11data->display, _x11data->pixmap, _x11data->gc_pix_pen,
                            x1, y1, x1-x2, y2-y1 );
            if ( _draw_win )
                XDrawRectangle( _x11data->display, _x11data->window, _x11data->gc_win_pen,
                                x1, y1, x1-x2, y2-y1 );
        }// if
        else
        {
            XDrawRectangle( _x11data->display, _x11data->pixmap, _x11data->gc_pix_pen,
                            x1, y2, x1-x2, y1-y2 );
            if ( _draw_win )
                XDrawRectangle( _x11data->display, _x11data->window, _x11data->gc_win_pen,
                                x1, y2, x1-x2, y1-y2 );
        }// else

    HANDLE_EVENTS;
#endif
}

void
TXImage::fill_rect ( int x1, int y1, int x2, int y2 )
{
#ifdef USE_X11
    CHECK_ACTIVE;
    
    if ( x1 < x2 )
        if ( y1 < y2 )
        {
            XFillRectangle( _x11data->display, _x11data->pixmap, _x11data->gc_pix_brush,
                            x1, y1, x2-x1+1, y2-y1+1 );
            if ( _draw_win )
                XFillRectangle( _x11data->display, _x11data->window, _x11data->gc_win_brush,
                                x1, y1, x2-x1+1, y2-y1+1 );
        }// if
        else
        {
            XFillRectangle( _x11data->display, _x11data->pixmap, _x11data->gc_pix_brush,
                            x1, y2, x2-x1+1, y1-y2+1 );
            if ( _draw_win )
                XFillRectangle( _x11data->display, _x11data->window, _x11data->gc_win_brush,
                                x1, y2, x2-x1+1, y1-y2+1 );
        }// else
    else
        if ( y1 < y2 )
        {
            XFillRectangle( _x11data->display, _x11data->pixmap, _x11data->gc_pix_brush,
                            x1, y1, x1-x2+1, y2-y1+1 );
            if ( _draw_win )
                XFillRectangle( _x11data->display, _x11data->window, _x11data->gc_win_brush,
                                x1, y1, x1-x2+1, y2-y1+1 );
        }// if
        else
        {
            XFillRectangle( _x11data->display, _x11data->pixmap, _x11data->gc_pix_brush,
                            x1, y2, x1-x2+1, y1-y2+1 );
            if ( _draw_win )
                XFillRectangle( _x11data->display, _x11data->window, _x11data->gc_win_brush,
                                x1, y2, x1-x2+1, y1-y2+1 );
        }// else

    HANDLE_EVENTS;
#endif
}

void
TXImage::draw_circle ( int x, int y, int radius )
{
#ifdef USE_X11
    CHECK_ACTIVE;
    
    XDrawArc( _x11data->display, _x11data->pixmap, _x11data->gc_pix_pen,
              x, y, radius, radius, 0, 360 );
              
    if ( _draw_win )
        XDrawArc( _x11data->display, _x11data->window, _x11data->gc_win_pen,
                  x, y, radius, radius, 0, 360 );
        
    HANDLE_EVENTS;
#endif
}

void
TXImage::fill_circle ( int x, int y, int radius )
{
#ifdef USE_X11
    CHECK_ACTIVE;
    
    XFillArc( _x11data->display, _x11data->pixmap, _x11data->gc_pix_brush,
              x, y, radius, radius, 0, 360 );
 
    if ( _draw_win )
        XFillArc( _x11data->display, _x11data->window, _x11data->gc_win_brush,
                  x, y, radius, radius, 0, 360 );
    
    HANDLE_EVENTS;
#endif
}

void
TXImage::move_to ( int x, int y )
{
#ifdef USE_X11
    CHECK_ACTIVE;
    
    _old_pos[0] = x;
    _old_pos[1] = y;
    
    HANDLE_EVENTS;
#endif
}

void
TXImage::line_to ( int x, int y )
{
#ifdef USE_X11
    CHECK_ACTIVE;
    
    draw_line( _old_pos[0], _old_pos[1], x, y );
    _old_pos[0] = x;
    _old_pos[1] = y;
    
    HANDLE_EVENTS;
#endif
}

void
TXImage::draw_text( int x, int y, const char * text )
{
#ifdef USE_X11
    CHECK_ACTIVE;

    if ( _x11data->font != NULL )
    {
        XDrawString( _x11data->display, _x11data->pixmap, _x11data->gc_pix_pen,
                     x, y, text, strlen( text ));

        if ( _draw_win )
            XDrawString( _x11data->display, _x11data->window, _x11data->gc_win_pen,
                         x, y, text, strlen( text ));
    }// if
    
    HANDLE_EVENTS;
#endif
}

//
// set styles
//

void
TXImage::set_rgb  ( int r, int g, int b )
{
#ifdef USE_X11
    CHECK_ACTIVE;
    
    ulong  pix;

    if ( _truecolour )
    {
        if ( _col_shift[ RED_COL   ] > 0 ) r <<=  _col_shift[ RED_COL   ];
        else                               r >>= -_col_shift[ RED_COL   ];
        
        if ( _col_shift[ GREEN_COL ] > 0 ) g <<=  _col_shift[ GREEN_COL ];
        else                               g >>= -_col_shift[ GREEN_COL ];
        
        if ( _col_shift[ BLUE_COL  ] > 0 ) b <<=  _col_shift[ BLUE_COL  ];
        else                               b >>= -_col_shift[ BLUE_COL  ];

        pix = ((r & _col_mask[ RED_COL ]) |
               (g & _col_mask[ GREEN_COL ]) |
               (b & _col_mask[ BLUE_COL ]));
    }// if
    else
    {
        //
        // allocate colour in local colourmap
        //
        
        XColor    col;
        
        col.red   = r << 8;
        col.green = g << 8;
        col.blue  = b << 8;
        
        if ( XAllocColor( _x11data->display, _x11data->cmap, & col ) )
            pix = col.pixel;
        else
        {
            std::cout << "(TXImage) set_rgb : can't allocate colour" << std::endl;
            HANDLE_EVENTS;
            return;
        }// else
    }// else
    
    XSetForeground( _x11data->display, _x11data->gc_pix_pen,   pix );
    XSetForeground( _x11data->display, _x11data->gc_pix_brush, pix );
    
    if ( _draw_win )
    {
        XSetForeground( _x11data->display, _x11data->gc_win_pen,   pix );
        XSetForeground( _x11data->display, _x11data->gc_win_brush, pix );
    }// if

    HANDLE_EVENTS;
#endif
}

void
TXImage::set_pen ( uint type )
{
#ifdef USE_X11
    CHECK_ACTIVE;
    

    HANDLE_EVENTS;
#endif
}

void
TXImage::set_line_width ( int width )
{
#ifdef USE_X11
    CHECK_ACTIVE;
    
    XSetLineAttributes( _x11data->display, _x11data->gc_pix_pen,
                        uint(width), LineSolid, CapButt, JoinMiter );
    if ( _draw_win )
        XSetLineAttributes( _x11data->display, _x11data->gc_win_pen,
                            uint(width), LineSolid, CapButt, JoinMiter );
    
    HANDLE_EVENTS;
#endif
}
    
void
TXImage::set_font ( const char * font, int size )
{
#ifdef USE_X11
    CHECK_ACTIVE;

    //
    // loop through font names and get X11 font name
    //

    int  isize = int(size+0.5);

    for ( uint i = 0; i < n_fonts; i++ )
    {
        if ( strcmp( font, font_table[2*i] ) == 0 )
        {
            // compare with old font and return if it is the same
            if (( _old_type == i ) && ( _old_size == isize ))
                return;

            _old_type = i;
            _old_size = isize;
            
            // remove old font
            if ( _x11data->font != NULL )
            {
                XFreeFont( _x11data->display, _x11data->font );
                _x11data->font = NULL;
            }// if
            
            // check if some fonts are available
            if ( _fonts[i].size() == 0 )
                return;
            
            const char  * fontname = NULL;
            int           min_diff = 0;
            
            if (( int(_fonts[i].size()) > isize ) && ( _fonts[i][isize] != "" ))
                fontname = _fonts[i][isize].c_str();
            else
            {
                min_diff = 10000;
                
                for ( uint j = 0; j < _fonts[i].size(); j++ )
                {
                    if ( _fonts[i][j] == "" )
                        continue;

                    if ( std::abs(int(isize - j)) < min_diff )
                    {
                        fontname = _fonts[i][j].c_str();
                        min_diff = std::abs(int(isize - j));
                    }// if
                    else
                        break;
                }// for
            }// else
            
            // no or no suitable font found, return
            if ((( isize < 10 ) && ( min_diff > 2 )) ||
                ( fontname == NULL ) || ( fontname[0] == '\0' ))
                return;
            
            if ((_x11data->font = XLoadQueryFont( _x11data->display, fontname )) == NULL )
                std::cout << "(TXImage) set_font : could not load font"
                          << " \"" << fontname << "\"" << std::endl;
            else
            {
                XSetFont( _x11data->display, _x11data->gc_pix_pen, _x11data->font->fid );
                if ( _draw_win )
                    XSetFont( _x11data->display, _x11data->gc_win_pen, _x11data->font->fid );
            }// else
            
            return;
        }// if
    }// for
    
    // no font found : remove old font
    if ( _x11data->font != NULL )
    {
        XFreeFont( _x11data->display, _x11data->font );
        _x11data->font = NULL;
    }// if
    
    HANDLE_EVENTS;
#endif
}

}// namespace
