//
// Project : lib3D
// File    : TParallelRT.cc
// Purpose : class for parallel raytracing
// Author  : Ronald Kriemann
//
// arch-tag: 7aa83634-1673-45d9-8a99-03da1c3dc9af
//

#include <sys/types.h>
#include <poll.h>
#include <unistd.h>
#include <signal.h>
#include <errno.h>

#include "TArray.hh"

#include "TParallelRT.hh"

namespace RGL
{

//
// check return value of read/write
//

#define CHECK_READ  { if ( n <  0 ) { perror( "(TParallelRT) render_world : read" ); continue; } \
                      if ( n == 0 ) { std::cout << "(TParallelRT) render_world : zero-length msg" \
                                                << std::endl; continue; } }

#define CHECK_WRITE  { if ( n < 0 ) { perror( "(TParallelRT) render_tiles : write" ); \
                                      continue; } }

//
// message types
//

#define MSG_TILE  1
#define MSG_STAT  2
#define MSG_HALT  3

//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////
//
// class TParallelRT
//
//////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////

///////////////////////////////////////////////
//
// constructor and destructor
//

TParallelRT::TParallelRT ( TRaytracer * rt, uint nprocs, uint tile_size )
        : _raytrace(rt), _nprocs(nprocs), _tile_size(tile_size)
{
    // check if we have a raytracer
    if ( _raytrace == NULL )
    {
        std::cout << "(TParallelRT) TParallelRT : no valid raytracer supplied" << std::endl;
        exit( 1 );
    }// if
}

TParallelRT::~TParallelRT ()
{
}

///////////////////////////////////////////////
//
// rendering
//

//
// render world to printer
//
void
TParallelRT::render_world ( TImage & img, TProgressBar * bar ) const
{
    // ignore child-signals
    signal( SIGCHLD, SIG_IGN );
    
    //
    // in case of one thread, call the single-cpu-method
    //
    
    if ( _nprocs == 1 )
    {
        _raytrace->render_world( img, bar );
        return;
    }// if
    
    //
    // reset statistics
    //

    reset_stat();
    
    //
    // multi-thread-algorithm:
    //   decompose image into tiles and let threads render these tiles
    //
    
    uint  width  = img.width();
    uint  height = img.height();

    //
    // create tile-sets for each task
    //

    uint                         x, y;
    uint                         t = 0;
    uint                         no_of_tiles = 0;
    TArray< TSLL< TTileDesc > >  tile_sets( _nprocs );

    y = 0;
    while ( y < height )
    {
        uint  h = _tile_size;

        if ( y + h > height )
            h = height - y;

        if ( h > 0 )
        {
            x = 0;
            while ( x < width )
            {
                uint  w = _tile_size;
                
                if ( x + w > width  )
                    w = width  - x;

                if ( w > 0 )
                {
                    TTileDesc  desc = { x, y, w, h, width, height };
                    
                    tile_sets[t].append( desc );
                    t++;
                    t %= _nprocs;

                    no_of_tiles++;
                    
                    x += w;
                }// if
            }// while

            y += h;
        }// if
    }// while

    //
    // start tasks
    //

    TArray< int >   com_pd( _nprocs );

    if ( bar != NULL )
        bar->init( 0, no_of_tiles, 0 );
    
    for ( uint i = 0; i < _nprocs; i++ )
    {
        int    pd[2];
        pid_t  pid;

        if ( pipe( pd ) != 0 )
            std::cout << "(TParallelRT) render_world : error in \"pipe\" ("
                      << std::strerror( errno ) << ")" << std::endl;
        else
        {
            // std::cout << "created pipe " << pd[0] << "/" << pd[1] << std::endl;
        
            if (( pid = fork() ) < 0 )
            {
                perror( "(TParallelRT) render_world : fork" );
                exit( 1 );
            }// if
            
            if ( pid == 0 )
            {
                // we dont read in the client
                if ( close( pd[0] ) < 0 )
                    std::cout << "(TParallelRT) render_world : error in \"close\" ("
                              << std::strerror( errno ) << ")" << std::endl;
                
                // std::cout << "write to pipe " << pd[1] << std::endl;
                
                render_tiles( pd[1], img, tile_sets[i] );
                
                // finish writing
                if ( close( pd[1] ) < 0 )
                    std::cout << "(TParallelRT) render_world : error in \"close\" ("
                              << std::strerror( errno ) << ")" << std::endl;
                
                exit(0);
            }// if
            else
            {
                // we dont write in the server
                if ( close( pd[1] ) < 0 )
                    std::cout << "(TParallelRT) render_world : error in \"close\" ("
                              << std::strerror( errno ) << ")" << std::endl;
                
                // std::cout << "reading from pipe " << pd[0] << std::endl;

                com_pd[i] = pd[0];
            }// else
        }// else
    }// for

    //
    // read data from clients until images was rendered
    //

    TArray< struct pollfd >  pfd( _nprocs );
    TImage                   tile( 0, 0 );
    TParallelRT::TTileDesc   desc;
    uint                     total, sum, max_lvl, j, n_pd;
    int                      n, type, n_rays;

    n_pd = _nprocs;
    
    while ( n_pd > 0 )
    {
        //
        // check for data in pipes
        //
        
        for ( uint i = 0; i < n_pd; i++ )
        {
            pfd[i].fd     = com_pd[i];
            pfd[i].events = POLLIN | POLLPRI; 
        }// for

        n = poll( pfd.c_array(), n_pd, -1 );

        if ( n == -1 )
            std::cout << "(TParallelRT) render_world : error in \"poll\" ("
                      << std::strerror( errno ) << ")" << std::endl;
        else if ( n > 0 )
        {
            //
            // go through all pipes and look for data
            //
            
            for ( uint i = 0; i < n_pd; i++ )
            {
                if ( pfd[i].revents & POLLIN )
                {
                    //
                    // read all data from pipe
                    //

                    // read first byte to decide upon message type
                    n = read( com_pd[i], & type, sizeof(type) );
                    CHECK_READ;
                    
                    // decide upon message type
                    if ( type == MSG_TILE )
                    {
                        //
                        // read desc of tile
                        //
                        
                        n = read( com_pd[i], & desc, sizeof(desc) );
                        CHECK_READ;
                        
                        //
                        // read image and copy tile to global image
                        //
                        
                        tile.set_geometry( desc.width, desc.height );
                        
                        sum   = 0;
                        total = tile.byte_size();
                        while ( sum < total )
                        {
                            n = read( com_pd[i], tile.pix_array().c_array() + sum, total - sum );
                            CHECK_READ;
                            
                            if ( n >= 0 )
                                sum += n;
                        }// while
                        
                        img.bitblt( desc.x_ofs, desc.y_ofs, tile );
                        
                        if ( verbose(1) && ( bar != NULL ))
                            bar->advance( 1 );
                    }// if
                    else if ( type == MSG_STAT )
                    {
                        //
                        // read statistics of child
                        //
                        
                        n = read( com_pd[i], & n_rays, sizeof(n_rays) );
                        CHECK_READ;

                        n = read( com_pd[i], & max_lvl, sizeof(max_lvl) );
                        CHECK_READ;
                        
                        _stat.no_of_rays   += n_rays;
                        _stat.deepest_level = max( _stat.deepest_level, max_lvl );
                    }// if
                    else if ( type == MSG_HALT )
                    {
                        //
                        // client has finished
                        //

                        if ( close( com_pd[i] ) < 0 )
                            std::cout << "(TParallelRT) render_world : error in \"close\" ("
                                      << std::strerror( errno ) << ")" << std::endl;

                        com_pd[i] = -1;
                    }// if
                    else
                    {
                        std::cout << "(TPipeIO) read_pipe : unkown message ("
                                  << type << ")" << std::endl;
                    }// else
                }// if
            }// for
        }// else

        //
        // sort active pd's in array
        //

        j = 0;
        for ( uint i = 0; i < _nprocs; i++ )
        {
            if ( com_pd[i] != -1 )
                com_pd[j++] = com_pd[i];
        }// for

        n_pd = j;

        // deactivate all other pd's
        for ( uint i = n_pd; i < _nprocs; i++ )
            com_pd[i] = -1;
    }// while
    
    if ( verbose(1) && ( bar != NULL ))
        bar->finish();
}

//
// render part of image
//
void
TParallelRT::render_tile ( const TTileDesc & tile,
                           TImage          & img,
                           TProgressBar    * bar ) const
{
    _raytrace->render_tile( tile, img, NULL );
}

//
// child-method: render list of tiles and send image to server
//
void
TParallelRT::render_tiles ( int                 pd,
                            TImage            & image,
                            TSLL< TTileDesc > & tile_list ) const
{
    //
    // render each tile and copy it back to global image
    //

    TImage  tile( 0, 0 );
    int     n, type;

    while ( tile_list.size() > 0 )
    {
        TTileDesc  desc( tile_list.behead() );

        if ( verbose(3) )
            std::cout << desc.width << "x" << desc.height
                      << "+" << desc.x_ofs << "+" << desc.y_ofs << std::endl;
        
        tile.set_geometry( desc.width, desc.height );

        render_tile( desc, tile, NULL );
        
        //
        // compose message and send tile to server
        //

        type = MSG_TILE;
        n = write( pd, & type, sizeof(type) );
        CHECK_WRITE;
        
        n = write( pd, & desc, sizeof(desc) );
        CHECK_WRITE;
        
        n = write( pd, tile.pix_array().c_array(), tile.byte_size() );
        CHECK_WRITE;
    }// while

    // adjust statistics
    _stat.no_of_rays    = _raytrace->no_of_rays();
    _stat.deepest_level = _raytrace->deepest_level();
    
    // send statistics
    type = MSG_STAT;
    if ((n = write( pd, & type, sizeof(type) )) < 0 )
        perror( "(TParallelRT) render_tiles : write" );

    if ((n = write( pd, & _stat.no_of_rays, sizeof(_stat.no_of_rays))) < 0 )
        perror( "(TParallelRT) render_tiles : write" );
    
    if ((n = write( pd, & _stat.deepest_level, sizeof(_stat.deepest_level) )) < 0 )
        perror( "(TParallelRT) render_tiles : write" );

    // send "halt" message
    type = MSG_HALT;
    if ((n = write( pd, & type, sizeof(type))) < 0 )
        perror( "(TParallelRT) render_tiles : write" );
}

} // namespace
